//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//	IrcWnd.cpp : implementation file
//

#include "stdafx.h"
#include "IrcWnd.h"
#include "emule.h"
#include "otherfunctions.h"
#include "opcodes.h"
#include "ColourPopup.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif

static const COLORREF _acrColorPal[16] =
{
	RGB(255, 255, 255), RGB(0, 0, 0), RGB(0, 0, 127), RGB(0, 127, 0),
	RGB(255, 0, 0), RGB(127, 0, 0), RGB(156, 0, 156), RGB(252, 127, 0),
	RGB(255, 255, 0), RGB(0, 252, 0), RGB(0, 147, 147), RGB(0, 255, 255),
	RGB(0, 0, 252), RGB(255, 0, 255), RGB(127, 127, 127), RGB(210, 210, 210)
};

/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
//	CIrcWnd dialog
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
IMPLEMENT_DYNAMIC(CIrcWnd, CDialog)
CIrcWnd::CIrcWnd(CWnd* pParent /*=NULL*/)
		: CResizableDialog(CIrcWnd::IDD, pParent)
{
	m_pIrcMain = NULL;
	m_bConnected = false;
	m_bLoggedIn = false;
	m_pCurrentChannel = NULL;
	memset(&m_bSortAscendingNickList, true, sizeof(m_bSortAscendingNickList));
	memset(&m_bSortAscendingChanList, true, sizeof(m_bSortAscendingChanList));
}

CIrcWnd::~CIrcWnd()
{
	if (m_bConnected)
		m_pIrcMain->Disconnect(true);

	POSITION	pos1, pos2;

	for (pos1 = channelLPtrList.GetHeadPosition(); (pos2 = pos1) != NULL;)
	{
		ChannelList* cur_channel = (ChannelList*)channelLPtrList.GetNext(pos1);
		channelLPtrList.RemoveAt(pos2);
		delete cur_channel;
	}
	DeleteAllChannel();
	delete m_pIrcMain;
}

void CIrcWnd::UpdateNickCount()
{
	CHeaderCtrl		*pHeaderCtrl = nickList.GetHeaderCtrl();
	HDITEM			hdi;
	CString			strRes;

	hdi.mask = HDI_TEXT;
	if (nickList.GetItemCount() > 0)
		strRes.Format(_T("%s[%i]"), GetResString(IDS_IRC_NICK), nickList.GetItemCount());
	else
		GetResString(&strRes, IDS_IRC_NICK);
	hdi.pszText = (LPSTR)strRes.GetString();
	pHeaderCtrl->SetItem(0, &hdi);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CIrcWnd::Localize()
{
	static const uint16 s_auResTbl[] =
	{
		IDS_IRC_NAME,
		IDS_UUSERS,
		IDS_DESCRIPTION
	};

	SetDlgItemText(IDC_BN_IRCCONNECT, GetResString((m_bConnected) ? IDS_IRC_DISCONNECT : IDS_IRC_CONNECT));
	SetDlgItemText(IDC_CHATSEND, GetResString(IDS_IRC_SEND));
	SetDlgItemText(IDC_CLOSECHAT, GetResString(IDS_CW_CLOSE));

	TCITEM		tcitem;
	Channel		*pCurChannel;

	for (int i = 0; i < channelselect.GetItemCount();i++)
	{
		tcitem.mask = TCIF_PARAM;
		tcitem.lParam = -1;
		channelselect.GetItem(i, &tcitem);
		pCurChannel = (Channel*)tcitem.lParam;
		if (pCurChannel != NULL)
		{
			if (pCurChannel->type == 1)
			{
				GetResString(&pCurChannel->title, IDS_STATUS);
				tcitem.mask = TCIF_TEXT;
				tcitem.pszText = (LPSTR)pCurChannel->title.GetString();
				channelselect.SetItem(i, &tcitem);
			}
			else if (pCurChannel->type == 2)
			{
				GetResString(&pCurChannel->title, IDS_IRC_CHANNELLIST);
				tcitem.mask = TCIF_TEXT;
				tcitem.pszText = (LPSTR)pCurChannel->title.GetString();
				channelselect.SetItem(i, &tcitem);
			}
		}
	}
	if (m_pCurrentChannel != NULL)
	{
		titleWindow.Reset();
		if (m_pCurrentChannel->type == 1)
			titleWindow.AppendText(GetResString(IDS_STATUS), RGB(0, 0, 128));
		else if(m_pCurrentChannel->type == 2)
			titleWindow.AppendText(GetResString(IDS_IRC_CHANNELLIST), RGB(0, 0, 128));
	}

	CHeaderCtrl	*pHeaderCtrl = nickList.GetHeaderCtrl();
	CString		strRes = GetResString(IDS_STATUS);
	HDITEM		hdi;

	hdi.mask = HDI_TEXT;
	hdi.pszText = (LPSTR)strRes.GetString();
	pHeaderCtrl->SetItem(1, &hdi);

	UpdateNickCount();

	pHeaderCtrl = serverChannelList.GetHeaderCtrl();

	for (unsigned ui = 0; ui < ELEMENT_COUNT(s_auResTbl); ui++)
	{
		::GetResString(&strRes, static_cast<UINT>(s_auResTbl[ui]));
		hdi.pszText = (LPSTR)strRes.GetString();
		pHeaderCtrl->SetItem(static_cast<int>(ui), &hdi);
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
BOOL CIrcWnd::OnInitDialog()
{
	static const uint16 s_auColHdr[][2] =
	{
		{ LVCFMT_LEFT, 90 },	// IRC1COL_NICK
		{ LVCFMT_LEFT, 70 }		// IRC1COL_STATUS
	};
	static const uint16 s_auColHdr2[][2] =
	{
		{ LVCFMT_LEFT, 203 },	// IRC2COL_NAME
		{ LVCFMT_RIGHT, 50 },	// IRC2COL_USERS
		{ LVCFMT_LEFT, 350 }	// IRC2COL_DESCRIPTION
	};

	CResizableDialog::OnInitDialog();

	CRect			rcRect;

	for (unsigned ui = 0; ui < ELEMENT_COUNT(s_auColHdr); ui++)
		nickList.InsertColumn(ui, _T(""), static_cast<int>(s_auColHdr[ui][0]), static_cast<int>(s_auColHdr[ui][1]));

	for (unsigned ui = 0; ui < ELEMENT_COUNT(s_auColHdr2); ui++)
		serverChannelList.InsertColumn(ui, _T(""), static_cast<int>(s_auColHdr2[ui][0]), static_cast<int>(s_auColHdr2[ui][1]));
	serverChannelList.LoadSettings(CPreferences::TABLE_IRC);

	m_bConnected = false;
	m_bLoggedIn = false;
	Localize();
	m_pIrcMain = new CIrcMain();
	m_pIrcMain->SetIRCWnd(this);

	NewChannel(GetResString(IDS_STATUS), 1);
	NewChannel(GetResString(IDS_IRC_CHANNELLIST), 2);

	serverChannelList.GetWindowRect(&rcRect);
	ScreenToClient(&rcRect);
	rcRect.bottom = rcRect.top + 18;
	titleWindow.Create(WS_CHILD | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | ES_AUTOHSCROLL | ES_READONLY | ES_NOHIDESEL, rcRect, this, (UINT)-1);
	titleWindow.ModifyStyleEx(0, WS_EX_STATICEDGE, SWP_FRAMECHANGED);
	titleWindow.SetEventMask(titleWindow.GetEventMask() | ENM_MOUSEEVENTS);
	titleWindow.m_dwFlags = HTC_ISLIMITED;
	//titleWindow.m_crDefaultForeground = GetSysColor(COLOR_INFOTEXT);
	//titleWindow.m_crDefaultBackground = GetSysColor(COLOR_INFOBK);

	UpdateFont();
	m_pCurrentChannel = (Channel*)channelPtrList.GetTail();
	AddMessage(_T(""), _T(""), CLR_DEFAULT, GetResString(IDS_IRC_STATUSLOG));
	channelselect.SetCurSel(2);
	channelselect.SetCurFocus(2);
	OnTcnSelchangeTab2(NULL, NULL);
	channelselect.SetItemState(1, TCIS_HIGHLIGHTED, NULL);

	AddAnchor(IDC_BN_IRCCONNECT, BOTTOM_LEFT);
	AddAnchor(IDC_CLOSECHAT, BOTTOM_LEFT);
	AddAnchor(IDC_CHATSEND, BOTTOM_RIGHT);
	AddAnchor(IDC_INPUTWINDOW, BOTTOM_LEFT, BOTTOM_RIGHT);
	AddAnchor(IDC_TEXTBOLD, BOTTOM_RIGHT);
	AddAnchor(IDC_TEXTUNDERLINE, BOTTOM_RIGHT);
	AddAnchor(IDC_TEXTCOLOR, BOTTOM_RIGHT);
	AddAnchor(IDC_NICKLIST, TOP_LEFT, BOTTOM_LEFT);
	AddAnchor(IDC_SERVERCHANNELLIST, TOP_LEFT, BOTTOM_RIGHT);
	AddAnchor(IDC_TAB2, TOP_LEFT, TOP_RIGHT);

	m_bSortAscendingChanList[IRC2COL_USERS] = false;

	nickList.SetSortArrow(IRC1COL_STATUS, m_bSortAscendingNickList[IRC1COL_STATUS]);
	nickList.SortItems(SortProcNick, IRC1COL_STATUS);

	if (g_App.m_pPrefs->DoUseSort())
		SortInit(g_App.m_pPrefs->GetIrcSortCol());
	else
	{
		int		iSortCode = g_App.m_pPrefs->GetColumnSortItem(CPreferences::TABLE_IRC);

		iSortCode |= g_App.m_pPrefs->GetColumnSortAscending(CPreferences::TABLE_IRC) ? MLC_SORTASC : MLC_SORTDESC;
		SortInit(iSortCode);
	}

	GetDlgItem(IDC_CLOSECHAT)->EnableWindow(false);
	GetDlgItem(IDC_CHATSEND)->EnableWindow(false);

	m_ctrlTextBoldBtn.SetThemeHelper(&g_App.m_pMDlg->m_themeHelper);
	m_ctrlTextBoldBtn.SetIcon(IDI_TEXTBOLD);
	m_ctrlTextUnderlineBtn.SetThemeHelper(&g_App.m_pMDlg->m_themeHelper);
	m_ctrlTextUnderlineBtn.SetIcon(IDI_TEXTUNDERLINE);
	m_ctrlTextColorBtn.SetThemeHelper(&g_App.m_pMDlg->m_themeHelper);
	m_ctrlTextColorBtn.SetIcon(IDI_TEXTCOLOR);

	return true;
}

void CIrcWnd::OnDestroy()
{
	CResizableDialog::OnDestroy();
}

void CIrcWnd::UpdateFont()
{
	TCITEM		tcitem;
	Channel		*pChannel;
	int			i = 0;

	tcitem.mask = TCIF_PARAM;
	while (channelselect.GetItem(i++, &tcitem))
	{
		pChannel = (Channel*)tcitem.lParam;
		if (pChannel->log.m_hWnd != NULL)
			pChannel->log.SetFont(&g_App.m_pMDlg->m_fontDefault);
	}
	titleWindow.SetFont(&g_App.m_pMDlg->m_fontDefault);
	inputWindow.SetFont(&g_App.m_pMDlg->m_fontDefault);
}

void CIrcWnd::OnSize(UINT nType, int cx, int cy)
{
	CResizableDialog::OnSize(nType, cx, cy);

	if (titleWindow.m_hWnd)
	{
		CRect		rcRect;

		serverChannelList.GetWindowRect(&rcRect);
		ScreenToClient(&rcRect);
		titleWindow.SetWindowPos(NULL, rcRect.left, rcRect.top, rcRect.Width(), 18, SWP_NOZORDER);
	}

	if (m_pCurrentChannel != NULL && m_pCurrentChannel->log.m_hWnd)
	{
		CRect		rcRect;

		serverChannelList.GetWindowRect(&rcRect);
		ScreenToClient(&rcRect);

		if (m_pCurrentChannel->type == 4)
			rcRect.top += 18;

		m_pCurrentChannel->log.SetWindowPos(NULL, rcRect.left, rcRect.top, rcRect.Width(), rcRect.Height(), SWP_NOZORDER);
	}
}

int CIrcWnd::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	return CResizableDialog::OnCreate(lpCreateStruct);
}

void CIrcWnd::DoDataExchange(CDataExchange* pDX)
{
	CResizableDialog::DoDataExchange(pDX);
	DDX_Control(pDX, IDC_NICKLIST, nickList);
	DDX_Control(pDX, IDC_INPUTWINDOW, inputWindow);
	DDX_Control(pDX, IDC_TEXTBOLD, m_ctrlTextBoldBtn);
	DDX_Control(pDX, IDC_TEXTUNDERLINE, m_ctrlTextUnderlineBtn);
	DDX_Control(pDX, IDC_TEXTCOLOR, m_ctrlTextColorBtn);
	DDX_Control(pDX, IDC_SERVERCHANNELLIST, serverChannelList);
	DDX_Control(pDX, IDC_TAB2, channelselect);
}

BOOL CIrcWnd::OnCommand(WPARAM wParam, LPARAM lParam)
{
	int			nickItem = nickList.GetNextItem(-1, LVIS_SELECTED | LVIS_FOCUSED);
	int			chanItem = channelselect.GetCurSel();
	Nick		*pNick;
	TCITEM		tcitem;
	Channel		*pChannel;
	CString		strSend;

	switch (wParam)
	{
		case IDC_BN_IRCCONNECT:
		{
			OnBnClickedBnIrcconnect();
			return true;
		}
		case IDC_TEXTBOLD:
		{
			OnBnClickedTextBold();
			return true;
		}
		case IDC_TEXTUNDERLINE:
		{
			OnBnClickedTextUnderline();
			return true;
		}
		case IDC_TEXTCOLOR:
		{
			OnBnClickedTextColor();
			return true;
		}
		case IDC_CHATSEND:
		{
			OnBnClickedChatsend();
			return true;
		}
		case IDC_CLOSECHAT:
		{
			OnBnClickedClosechat();
			return true;
		}
		case Irc_Priv:
		{
			pNick = (Nick*)nickList.GetItemData(nickItem);
			if (pNick != NULL && FindChannelByName(pNick->nick) == NULL)
				AddMessage(pNick->nick, _T(""), RGB(0, 147, 0), GetResString(IDS_IRC_PRIVATECHANSTART));
			return true;
		}
		case Irc_WhoIs:
		{
			pNick = (Nick*)nickList.GetItemData(nickItem);
			if (pNick != NULL)
			{
				strSend.Format(_T("WHOIS %s"), pNick->nick);
				m_pIrcMain->SendString(strSend);
			}
			return true;
		}
		case Irc_Owner:
		{
			pNick = (Nick*)nickList.GetItemData(nickItem);
			tcitem.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &tcitem);
			pChannel = (Channel*)tcitem.lParam;
			if (pNick != NULL && pChannel != NULL)
			{
				strSend.Format(_T("PRIVMSG chanserv owner %s %s"), pChannel->name, pNick->nick);
				m_pIrcMain->SendString(strSend);
			}
			return true;
		}
		case Irc_DeOwner:
		{
			pNick = (Nick*)nickList.GetItemData(nickItem);
			tcitem.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &tcitem);
			pChannel = (Channel*)tcitem.lParam;
			if (pNick != NULL && pChannel != NULL)
			{
				strSend.Format(_T("PRIVMSG chanserv deowner %s %s"), pChannel->name, pNick->nick);
				m_pIrcMain->SendString(strSend);
			}
			return true;
		}
		case Irc_Op:
		{
			pNick = (Nick*)nickList.GetItemData(nickItem);
			tcitem.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &tcitem);
			pChannel = (Channel*)tcitem.lParam;
			if (pNick != NULL && pChannel != NULL)
			{
				strSend.Format(_T("MODE %s +o %s"), pChannel->name, pNick->nick);
				m_pIrcMain->SendString(strSend);
			}
			return true;
		}
		case Irc_DeOp:
		{
			pNick = (Nick*)nickList.GetItemData(nickItem);
			tcitem.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &tcitem);
			pChannel = (Channel*)tcitem.lParam;
			if (pNick != NULL && pChannel != NULL)
			{
				strSend.Format(_T("MODE %s -o %s"), pChannel->name, pNick->nick);
				m_pIrcMain->SendString(strSend);
			}
			return true;
		}
		case Irc_HalfOp:
		{
			pNick = (Nick*)nickList.GetItemData(nickItem);
			tcitem.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &tcitem);
			pChannel = (Channel*)tcitem.lParam;
			if (pNick != NULL && pChannel != NULL)
			{
				strSend.Format(_T("MODE %s +h %s"), pChannel->name, pNick->nick);
				m_pIrcMain->SendString(strSend);
			}
			return true;
		}
		case Irc_DeHalfOp:
		{
			pNick = (Nick*)nickList.GetItemData(nickItem);
			tcitem.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &tcitem);
			pChannel = (Channel*)tcitem.lParam;
			if (pNick != NULL && pChannel != NULL)
			{
				strSend.Format(_T("MODE %s -h %s"), pChannel->name, pNick->nick);
				m_pIrcMain->SendString(strSend);
			}
			return true;
		}
		case Irc_Voice:
		{
			pNick = (Nick*)nickList.GetItemData(nickItem);
			tcitem.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &tcitem);
			pChannel = (Channel*)tcitem.lParam;
			if (pNick != NULL && pChannel != NULL)
			{
				strSend.Format(_T("MODE %s +v %s"), pChannel->name, pNick->nick);
				m_pIrcMain->SendString(strSend);
			}
			return true;
		}
		case Irc_DeVoice:
		{
			pNick = (Nick*)nickList.GetItemData(nickItem);
			tcitem.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &tcitem);
			pChannel = (Channel*)tcitem.lParam;
			if (pNick != NULL && pChannel != NULL)
			{
				strSend.Format(_T("MODE %s -v %s"), pChannel->name, pNick->nick);
				m_pIrcMain->SendString(strSend);
			}
			return true;
		}
		case Irc_Protect:
		{
			pNick = (Nick*)nickList.GetItemData(nickItem);
			tcitem.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &tcitem);
			pChannel = (Channel*)tcitem.lParam;
			if (pNick != NULL && pChannel != NULL)
			{
				strSend.Format(_T("PRIVMSG chanserv protect %s %s"), pChannel->name, pNick->nick);
				m_pIrcMain->SendString(strSend);
			}
			return true;
		}
		case Irc_DeProtect:
		{
			pNick = (Nick*)nickList.GetItemData(nickItem);
			tcitem.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &tcitem);
			pChannel = (Channel*)tcitem.lParam;
			if (pNick != NULL && pChannel != NULL)
			{
				strSend.Format(_T("PRIVMSG chanserv deprotect %s %s"), pChannel->name, pNick->nick);
				m_pIrcMain->SendString(strSend);
			}
			return true;
		}
		case Irc_Kick:
		{
			pNick = (Nick*)nickList.GetItemData(nickItem);
			tcitem.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &tcitem);
			pChannel = (Channel*)tcitem.lParam;
			if (pNick != NULL && pChannel != NULL)
			{
				strSend.Format(_T("KICK %s %s"), pChannel->name, pNick->nick);
				m_pIrcMain->SendString(strSend);
			}
			return true;
		}
		case Irc_Slap:
		{
			pNick = (Nick*)nickList.GetItemData(nickItem);
			tcitem.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &tcitem);
			pChannel = (Channel*)tcitem.lParam;
			if (pNick != NULL && pChannel != NULL)
			{
				strSend.Format(GetResString(IDS_IRC_SLAPMSGSEND), pChannel->name, pNick->nick);
				AddMessage(pChannel->name, _T(""), RGB(252, 127, 0), GetResString(IDS_IRC_SLAPMSG), m_pIrcMain->GetNick(), pNick->nick);
				m_pIrcMain->SendString(strSend);
			}
			return true;
		}
		case Irc_AddFriend:
		{
			pNick = (Nick*)nickList.GetItemData(nickItem);
			tcitem.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &tcitem);
			pChannel = (Channel*)tcitem.lParam;
			if (pNick != NULL && pChannel != NULL)
			{
				strSend.Format(_T("PRIVMSG %s :\001RQSFRIEND |%i|\001"), pNick->nick, m_pIrcMain->SetVerify());
				m_pIrcMain->SendString(strSend);
			}
			return true;
		}
		case Irc_SendLink:
		{
		   	if (!GetSendFileString())
				   return true;

			pNick = (Nick*)nickList.GetItemData(nickItem);
			tcitem.mask = TCIF_PARAM;
			channelselect.GetItem(chanItem, &tcitem);
			pChannel = (Channel*)tcitem.lParam;
			if (pNick != NULL && pChannel != NULL)
			{
				strSend.Format(_T("PRIVMSG %s :\001SENDLINK|%s|%s\001"), pNick->nick, HashToString(g_App.m_pPrefs->GetUserHash()), GetSendFileString() );
				m_pIrcMain->SendString(strSend);
			}
			return true;
		}
		case Irc_Close:
		{
			OnBnClickedClosechat();
			return true;
		}
		case Irc_Join:
		{
			JoinChannels();
			return true;
		}
		case Irc_Refresh:
		{
			m_pIrcMain->SendString(_T("LIST"));
			return true;
		}
	}
	return true;
}

BOOL CIrcWnd::PreTranslateMessage(MSG* pMsg)
{
	EMULE_TRY

	if (pMsg->message == WM_KEYDOWN && (pMsg->hwnd == inputWindow.m_hWnd))
	{
		if (pMsg->wParam == VK_RETURN)
		{
			OnBnClickedChatsend();
			return TRUE;
		}

		if (pMsg->wParam == VK_UP || pMsg->wParam == VK_DOWN)
		{
			ScrollHistory(pMsg->wParam == VK_DOWN);
			return TRUE;
		}
	}

	return CResizableDialog::PreTranslateMessage(pMsg);

	EMULE_CATCH

	return false;
}

void CIrcWnd::OnBnClickedBnIrcconnect()
{
	if (!m_bConnected)
		m_pIrcMain->Connect();
	else
		m_pIrcMain->Disconnect();
	inputWindow.SetFocus();
}

void CIrcWnd::OnBnClickedClosechat(int nItem)
{
	TCITEM		tcitem;

	tcitem.mask = TCIF_PARAM;
	if (nItem == -1)
		nItem = channelselect.GetCurSel();

	if (nItem == -1)
		return;

	channelselect.GetItem(nItem, &tcitem);

	Channel		*pPartChannel = (Channel*)tcitem.lParam;

	if (pPartChannel != NULL)
	{
		if (pPartChannel->type == 4 && m_bConnected)
		{
			CString		strPart = _T("PART ");

			strPart += pPartChannel->name;
			m_pIrcMain->SendString(strPart);
			return;
		}
		else if (pPartChannel->type == 5 || pPartChannel->type == 4)
		{
			RemoveChannel(pPartChannel->name);
			return;
		}
	}
}

void CIrcWnd::OnEnLinkIrcBox(NMHDR *pNMHDR, LRESULT *pResult)
{
	*pResult = 0;

	ENLINK		*pEnLink = reinterpret_cast<ENLINK *>(pNMHDR);
	TCITEM		tcitem;

	tcitem.mask = TCIF_PARAM;

	int			iCurSel = channelselect.GetCurSel();

	if (iCurSel != -1 && pEnLink && pEnLink->msg == WM_LBUTTONDOWN && channelselect.GetItem(iCurSel, &tcitem))
	{
		Channel		*pUpdate = (Channel*)tcitem.lParam;
		CString		strUrl;

		pUpdate->log.GetTextRange(pEnLink->chrg.cpMin, pEnLink->chrg.cpMax, strUrl);
		ShellExecute(NULL, NULL, strUrl, NULL, NULL, SW_SHOWDEFAULT);
		*pResult = 1;
	}
}

void CIrcWnd::OnTcnSelchangeTab2(NMHDR *pNMHDR, LRESULT *pResult)
{
	nickList.DeleteAllItems();

	TCITEM		tcitem;

	tcitem.mask = TCIF_PARAM;

	int		iCurSel = channelselect.GetCurSel();

	if (iCurSel == -1)
		return;

	if (!channelselect.GetItem(iCurSel, &tcitem))
		return;

	Channel		*pChannel = (Channel*)tcitem.lParam;

	if (pChannel == NULL)
		return;

	m_pCurrentChannel = pChannel;

	int			i;
	bool		bWasChanged = (channelselect.GetItemState(iCurSel, TCIS_HIGHLIGHTED) & TCIS_HIGHLIGHTED);

	UpdateNickCount();
	channelselect.SetItemState(iCurSel, TCIS_HIGHLIGHTED, NULL);
	titleWindow.ShowWindow(SW_HIDE);
	if (m_pCurrentChannel->type == 1)
		GetDlgItem(IDC_CLOSECHAT)->EnableWindow(false);
	else if (m_pCurrentChannel->type == 2)
	{
		serverChannelList.ShowWindow(SW_SHOW);
		GetDlgItem(IDC_CLOSECHAT)->EnableWindow(false);
		tcitem.mask = TCIF_PARAM;
		i = 0;
		while (channelselect.GetItem(i++, &tcitem))
		{
			pChannel = (Channel*)tcitem.lParam;
			if (pChannel != m_pCurrentChannel && pChannel->log.m_hWnd != NULL)
				pChannel->log.ShowWindow(SW_HIDE);
		}
		return;
	}
	else
		GetDlgItem(IDC_CLOSECHAT)->EnableWindow(true);
	SetActivity(m_pCurrentChannel->name, false);

	CRect		rcChannel;

	serverChannelList.GetWindowRect(&rcChannel);
	ScreenToClient(&rcChannel);
	if (m_pCurrentChannel->type == 4)
	{
		rcChannel.top += 18;
		titleWindow.ShowWindow(SW_SHOW);
		SetTitle(m_pCurrentChannel->name, m_pCurrentChannel->title);
	}
	m_pCurrentChannel->log.SetWindowPos(NULL, rcChannel.left, rcChannel.top, rcChannel.Width(), rcChannel.Height(), SWP_NOZORDER);
	m_pCurrentChannel->log.SetRedraw(false);
	m_pCurrentChannel->log.ShowWindow(SW_SHOW);
	if ((m_pCurrentChannel->log.m_dwFlags & HTC_ISAUTOSCROLL) != 0 && bWasChanged)
		m_pCurrentChannel->log.ScrollToLastLine();
	m_pCurrentChannel->log.SetRedraw(true);

	tcitem.mask = TCIF_PARAM;
	i = 0;
	while (channelselect.GetItem(i++, &tcitem))
	{
		pChannel = (Channel*)tcitem.lParam;
		if (pChannel != m_pCurrentChannel && pChannel->log.m_hWnd != NULL)
			pChannel->log.ShowWindow(SW_HIDE);
	}
	serverChannelList.ShowWindow(SW_HIDE);
	RefreshNickList(m_pCurrentChannel->name);
	m_pCurrentChannel->log.SetTitle(m_pCurrentChannel->name);
	inputWindow.SetFocus();
	if (pResult != NULL)
		*pResult = 0;
}

BOOL CIrcWnd::OnNotify(WPARAM wParam, LPARAM lParam, LRESULT* pResult)
{
	bool				bResult = CDialog::OnNotify(wParam, lParam, pResult);
	MSGFILTER		   *lpMsgFilter = (MSGFILTER *)lParam;

	if (lpMsgFilter == NULL || lpMsgFilter->nmhdr.code != EN_MSGFILTER)
		return bResult;

	CHTRichEditCtrl		*pRichEditCtrl;

	if (lpMsgFilter->nmhdr.hwndFrom == m_pCurrentChannel->log.m_hWnd)
		pRichEditCtrl = &m_pCurrentChannel->log;
	else if (lpMsgFilter->nmhdr.hwndFrom == titleWindow.m_hWnd)
		pRichEditCtrl = &titleWindow;
	else
		return bResult;

	if ((pRichEditCtrl->m_dwFlags & HTC_ISDEFAULTLINKS) != 0)
		return bResult;

	pRichEditCtrl->m_dwFlags &= ~HTC_ISHANDCURSOR;

	CPoint			point;

	::GetCursorPos(&point);
	pRichEditCtrl->ScreenToClient(&point);

	int			iCharPos = pRichEditCtrl->CharFromPos(point);
	int			iLine = pRichEditCtrl->LineFromChar(iCharPos);
	int			iLineIndex = pRichEditCtrl->LineIndex(iLine);
	int			iLen = pRichEditCtrl->LineLength(iLineIndex);

	if (iLen <= 3)
		return bResult;

	TCHAR		*pBuffer = new TCHAR[iLen + 3];

	memzero(pBuffer, sizeof(TCHAR) * (iLen + 3));
	pRichEditCtrl->GetLine(iLine, pBuffer, iLen);

	CString		strLineOrginal = CString(pBuffer, iLen);

	delete pBuffer;

	CString		strLine = strLineOrginal;
	int			iIndex = 0, iScheme, iLinkEnd;

	iCharPos -= iLineIndex;
	//	Convert string to lower case only once
	strLine.MakeLower();
	while (strLineOrginal.GetLength() > iIndex)
	{
		for (iScheme = 0; iScheme < ARRSIZE(_apszSchemes); iScheme++)
		{
			iLen = _apszSchemes[iScheme].uiLen;
			iLinkEnd = iIndex + iLen;

			if ((strLine.GetLength() > iLinkEnd) &&
				(strLine.Mid(iIndex, iLen).Compare(_apszSchemes[iScheme].pszScheme) == 0))
			{
				while (strLine.GetLength() > iLinkEnd && strLine[iLinkEnd] != _T(' ')
					&& strLine[iLinkEnd] != _T('\n') && strLine[iLinkEnd] != _T('\\') &&
					strLine[iLinkEnd] != _T('\r') && strLine[iLinkEnd] != _T('\t'))
				{
					iLinkEnd++;
				}

				while ((iLinkEnd > iIndex + iLen) &&
					!((strLine[iLinkEnd - 1] >= _T('/') && strLine[iLinkEnd - 1] <= _T('9'))
					||(strLine[iLinkEnd - 1] >= _T('a') && strLine[iLinkEnd - 1] <= _T('z'))
					|| (strLine[iLinkEnd - 1] == _T('|'))))
				{
					iLinkEnd--;
				}

				if (iLinkEnd > iIndex + iLen)
				{
					if (iCharPos > iIndex && iCharPos < iLinkEnd)
					{
						pRichEditCtrl->m_strURL = strLineOrginal.Mid(iIndex, iLinkEnd - iIndex);
						pRichEditCtrl->m_dwFlags |= HTC_ISHANDCURSOR;
						return bResult;
					}
					iIndex = iLinkEnd;
				}
				else
					iIndex += iLen;

				break;
			}
		}
		if (iScheme >= ARRSIZE(_apszSchemes))
			iIndex++;
	}

	return bResult;
 }

BEGIN_MESSAGE_MAP(CIrcWnd, CDialog)
	ON_NOTIFY(EN_LINK, 125, OnEnLinkIrcBox)
	ON_NOTIFY(NM_DBLCLK, IDC_SERVERCHANNELLIST, OnNMDblclkserverChannelList)
	ON_NOTIFY(NM_DBLCLK, IDC_NICKLIST, OnNMDblclkNickList)
	ON_NOTIFY(LVN_COLUMNCLICK, IDC_NICKLIST, OnColumnClickNick)
	ON_NOTIFY(LVN_COLUMNCLICK, IDC_SERVERCHANNELLIST, OnColumnClickChanL)
	ON_NOTIFY(NM_RCLICK, IDC_NICKLIST, OnNMRclickNick)
	ON_NOTIFY(NM_RCLICK, IDC_SERVERCHANNELLIST, OnNMRclickChanL)
	ON_NOTIFY(TCN_SELCHANGE, IDC_TAB2, OnTcnSelchangeTab2)
	ON_MESSAGE(WM_CLOSETAB, OnCloseTab)
	ON_MESSAGE(CPN_SELENDOK, OnColorSelEndOK)
	ON_MESSAGE(CPN_SELENDCANCEL, OnColorSelEndCancel)
	ON_WM_SIZE()
	ON_WM_CREATE()
	ON_WM_DESTROY()
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
//	Channel List
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////

void CIrcWnd::ResetServerChannelList()
{
	POSITION		pos1, pos2;
	ChannelList		*pCurChannel;

	for (pos1 = channelLPtrList.GetHeadPosition(); (pos2 = pos1) != NULL;)
	{
		pCurChannel = (ChannelList*)channelLPtrList.GetNext(pos1);
		channelLPtrList.RemoveAt(pos2);
		delete pCurChannel;
	}
	serverChannelList.DeleteAllItems();
}

void CIrcWnd::AddChannelToList(CString name, CString user, CString description)
{
	CString		ntemp = name;
	CString		dtemp = description;
	int			iUserTest = _tstoi(user);

	if ((g_App.m_pPrefs->GetIRCChanNameFilter() || g_App.m_pPrefs->GetIRCChannelUserFilter()) && g_App.m_pPrefs->GetIRCUseChanFilter())
	{
		if (iUserTest < g_App.m_pPrefs->GetIRCChannelUserFilter())
			return;
		if (dtemp.MakeLower().Find(g_App.m_pPrefs->GetIRCChanNameFilter().MakeLower()) == -1 && ntemp.MakeLower().Find(g_App.m_pPrefs->GetIRCChanNameFilter().MakeLower()) == -1)
			return;
	}
	ChannelList		*pToAdd = new ChannelList;

	pToAdd->name = name;
	pToAdd->users = user;
	pToAdd->desc = description;
	pToAdd->desc.Replace(_T('\x04'), _T('%'));

	int			iIndex = 0, iLen, iColorIndex;
	CString		strTemp;
	TCHAR		cCurChar;

	while (pToAdd->desc.GetLength() > iIndex)
	{
		cCurChar = pToAdd->desc[iIndex];
		switch (cCurChar)
		{
			case 0x03:
			{
				pToAdd->desc.Delete(iIndex);
				if (pToAdd->desc.GetLength() > iIndex)
				{
					if (pToAdd->desc[iIndex] >= _T('0') && pToAdd->desc[iIndex] <= _T('9'))
					{
						if (pToAdd->desc.GetLength() > iIndex)
						{
							iLen = 1;
							if (pToAdd->desc[iIndex + 1] >= _T('0') && pToAdd->desc[iIndex + 1] <= _T('9'))
								iLen++;
							iColorIndex = _tstoi(pToAdd->desc.Mid(iIndex, iLen));
							pToAdd->desc.Delete(iIndex, iLen);
							if (iColorIndex < 16)
							{
								strTemp.Format(_T("\001%06x"), _acrColorPal[iColorIndex]);
								pToAdd->desc.Insert(iIndex, strTemp);
								iIndex += strTemp.GetLength();
							}
							if (pToAdd->desc.GetLength() > iIndex && pToAdd->desc[iIndex] == _T(',') && pToAdd->desc[iIndex + 1] >= _T('0') && pToAdd->desc[iIndex + 1] <= _T('9'))
							{
								iLen = 1;
								if (pToAdd->desc.GetLength() > iIndex + 1 && pToAdd->desc[iIndex + 2] >= _T('0') && pToAdd->desc[iIndex + 2] <= _T('9'))
									iLen++;
								iColorIndex = _tstoi(pToAdd->desc.Mid(iIndex + 1, iLen));
								pToAdd->desc.Delete(iIndex, iLen + 1);
								if (iColorIndex < 16)
								{
									strTemp.Format(_T("\002%06x"), _acrColorPal[iColorIndex]);
									pToAdd->desc.Insert(iIndex, strTemp);
									iIndex += strTemp.GetLength();
								}
							}
						}
					}
					else
					{
						pToAdd->desc.Insert(iIndex, _T('\x03'));
						iIndex++;
					}
				}
				break;
			}
			case 0x0F:
			{
				pToAdd->desc.SetAt(iIndex, _T('\x03'));
				iIndex++;
				break;
			}
			default:
			{
				if ((TBYTE)cCurChar < _T(' '))
					pToAdd->desc.Delete(iIndex);
				else
					iIndex++;
			}
		}
	}

	channelLPtrList.AddTail(pToAdd);

	int			iItems = serverChannelList.InsertItem(LVIF_PARAM, serverChannelList.GetItemCount(), 0, 0, 0, 0, (LPARAM)pToAdd);

	serverChannelList.SetItemState(iItems, LVIS_OVERLAYMASK, LVIS_OVERLAYMASK);
	serverChannelList.SetItemText(iItems, 0, pToAdd->name);
	serverChannelList.SetItemText(iItems, 1, pToAdd->users);
	serverChannelList.SetItemText(iItems, 2, pToAdd->desc);
	if (channelselect.GetCurSel() != 1)
		channelselect.SetItemState(1, TCIS_HIGHLIGHTED, TCIS_HIGHLIGHTED);
}

int CIrcWnd::SortProcChanL(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	ChannelList		*pItem1 = (ChannelList*)lParam1;
	ChannelList		*pItem2 = (ChannelList*)lParam2;

	switch (lParamSort)
	{
		case IRC2COL_NAME:
			return CString(pItem1->name).CompareNoCase(pItem2->name);
		case IRC2COL_NAME + MLC_SORTDESC:
			return CString(pItem2->name).CompareNoCase(pItem1->name);
		case IRC2COL_USERS:
			return _tstoi(pItem1->users) - _tstoi(pItem2->users);
		case IRC2COL_USERS + MLC_SORTDESC:
			return _tstoi(pItem2->users) - _tstoi(pItem1->users);
		case IRC2COL_DESCRIPTION:
			return CString(pItem1->desc).CompareNoCase(pItem2->desc);
		case IRC2COL_DESCRIPTION + MLC_SORTDESC:
			return CString(pItem2->desc).CompareNoCase(pItem1->desc);
		default:
			return 0;
	}
}

void CIrcWnd::OnColumnClickChanL(NMHDR* pNMHDR, LRESULT* pResult)
{
	NM_LISTVIEW		*pNMListView = reinterpret_cast<NM_LISTVIEW*>(pNMHDR);
	int				iSubItem = pNMListView->iSubItem;
	bool			bSortOrder = m_bSortAscendingChanList[iSubItem];

// Reverse sorting direction for the same column and keep the same if column was changed
	if ((serverChannelList.GetSortParam() & MLC_COLUMNMASK) == iSubItem)
		m_bSortAscendingChanList[iSubItem] = bSortOrder = !bSortOrder;

	serverChannelList.SetSortArrow(iSubItem, bSortOrder);
	serverChannelList.SortItems(SortProcChanL, iSubItem + ((bSortOrder) ? MLC_SORTASC : MLC_SORTDESC));
	g_App.m_pPrefs->SetColumnSortItem(CPreferences::TABLE_IRC, iSubItem);
	g_App.m_pPrefs->SetColumnSortAscending(CPreferences::TABLE_IRC, bSortOrder);
	*pResult = 0;
}

void CIrcWnd::SortInit(int iSortCode)
{
//	Get the sort column
	int		iSortColumn = (iSortCode & MLC_COLUMNMASK);
//	Get the sort order
	bool	bSortAscending = (iSortCode & MLC_SORTDESC) == 0;

	serverChannelList.SetSortArrow(iSortColumn, bSortAscending);
	serverChannelList.SortItems(SortProcChanL, iSortCode);
	m_bSortAscendingChanList[iSortColumn] = bSortAscending;
}

void CIrcWnd::OnNMRclickChanL(NMHDR *pNMHDR, LRESULT *pResult)
{
	CTitleMenu	menuChanL;
	POINT		point;
	CString		strRefresh = GetResString(IDS_CMT_REFRESH);

	::GetCursorPos(&point);

	strRefresh.Remove(_T('&'));
	menuChanL.CreatePopupMenu();
	menuChanL.AddMenuTitle(GetResString(IDS_IRC_CHANNEL));
	menuChanL.AppendMenu(MF_STRING | ((serverChannelList.GetSelectionMark() != -1) ? MF_ENABLED : MF_GRAYED), Irc_Join, GetResString(IDS_IRC_JOIN));
	menuChanL.AppendMenu(MF_SEPARATOR);
	menuChanL.AppendMenu(MF_STRING | ((this->m_bConnected) ? MF_ENABLED : MF_GRAYED), Irc_Refresh, strRefresh);
	menuChanL.SetDefaultItem(Irc_Join);
	menuChanL.TrackPopupMenuEx(TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y, this, NULL);

//	Menu objects are destroyed in their destructor
	*pResult = 0;
}

void CIrcWnd::OnNMDblclkserverChannelList(NMHDR *pNMHDR, LRESULT *pResult)
{
	JoinChannels();
	*pResult = 0;
}

/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
//	Nick List
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////

Nick* CIrcWnd::FindNickByName(CString channel, CString name)
{
	Channel		*pCurChannel = FindChannelByName(channel);
	Nick		*pCurNick;

	if (pCurChannel == NULL)
		return 0;

	for (POSITION pos = pCurChannel->nicks.GetHeadPosition(); pos != NULL;)
	{
		pCurNick = (Nick*)pCurChannel->nicks.GetNext(pos);
		if (pCurNick->nick == name)
			return pCurNick;
	}
	return 0;
}

Nick* CIrcWnd::NewNick(CString channel, CString nick)
{
	Channel		*pToAddChan = FindChannelByName(channel);
	TCHAR		cFCh;

	if (pToAddChan == NULL)
		return NULL;

	if (FindNickByName(channel, nick) != NULL)
		return NULL;

	Nick	   *pToAddNick = new Nick;

	if (nick.GetAt(0) == _T('~'))
		nick.SetAt(0, _T('!'));

	if (nick.GetAt(0) == _T('!'))
	{
		nick = nick.Mid(1);
		pToAddNick->owner = _T("!");
	}
	else
		pToAddNick->owner.Empty();

	if (nick.GetAt(0) == _T('&'))
		nick.SetAt(0, _T('*'));

	if (nick.GetAt(0) == _T('*'))
	{
		nick = nick.Mid(1);
		pToAddNick->protect = _T("*");
	}
	else
		pToAddNick->protect.Empty();

	if ((cFCh = nick.GetAt(0)) == _T('@'))
	{
		pToAddNick->op = _T("@");
		pToAddNick->hop.Empty();
		pToAddNick->voice.Empty();
		pToAddNick->uop.Empty();
		pToAddNick->nick = nick.Mid(1);
	}
	else if (cFCh == _T('%'))
	{
		pToAddNick->op.Empty();
		pToAddNick->hop = _T("%");
		pToAddNick->voice.Empty();
		pToAddNick->uop.Empty();
		pToAddNick->nick = nick.Mid(1);
	}
	else if (cFCh == _T('+'))
	{
		pToAddNick->op.Empty();
		pToAddNick->hop.Empty();
		pToAddNick->voice = _T("+");
		pToAddNick->uop.Empty();
		pToAddNick->nick = nick.Mid(1);
	}
	else if (cFCh == _T('-'))
	{
		pToAddNick->op.Empty();
		pToAddNick->hop.Empty();
		pToAddNick->voice.Empty();
		pToAddNick->uop = _T("-");
		pToAddNick->nick = nick.Mid(1);
	}
	else
	{
		pToAddNick->op.Empty();
		pToAddNick->hop.Empty();
		pToAddNick->voice.Empty();
		pToAddNick->nick = nick;
	}
	pToAddNick->nick.TrimLeft(_T("*@%+-"));
	pToAddChan->nicks.AddTail(pToAddNick);
	if (pToAddChan == m_pCurrentChannel)
	{
		CString		strMode;
		int			iItems = nickList.InsertItem(LVIF_PARAM, nickList.GetItemCount(), 0, 0, 0, 0, (LPARAM)pToAddNick);

		nickList.SetItemText(iItems, 0, (LPCTSTR)pToAddNick->nick);
		strMode.Format(_T("%s%s%s%s%s%s"), pToAddNick->owner, pToAddNick->protect, pToAddNick->op, pToAddNick->hop, pToAddNick->voice, pToAddNick->uop);
		nickList.SetItemText(iItems, 1, (LPCTSTR)strMode);
		UpdateNickCount();
	}
	return pToAddNick;
}

void CIrcWnd::RefreshNickList(CString channel)
{
	nickList.DeleteAllItems();

	Channel		*pRefresh = FindChannelByName(channel);

	if (pRefresh == NULL)
		return;

	Nick		*pCurNick;
	int			iItems;
	CString		strMode;

	for (POSITION pos = pRefresh->nicks.GetHeadPosition(); pos != NULL;)
	{
        pCurNick = (Nick*)pRefresh->nicks.GetNext(pos);
		iItems = nickList.InsertItem(LVIF_PARAM, nickList.GetItemCount(), 0, 0, 0, 0, (LPARAM)pCurNick);
		nickList.SetItemText(iItems, 0, (LPCTSTR)pCurNick->nick);
		strMode.Format(_T("%s%s%s%s%s%s"), pCurNick->owner, pCurNick->protect, pCurNick->op, pCurNick->hop, pCurNick->voice, pCurNick->uop);
		nickList.SetItemText(iItems, 1, (LPCTSTR)strMode);
	}
	UpdateNickCount();
}

bool CIrcWnd::RemoveNick(CString channel, CString nick)
{
	Channel		*pUpdate = FindChannelByName(channel);

	if (pUpdate == NULL)
		return false;

	POSITION		pos1, pos2;
	Nick			*pCurNick;
	LVFINDINFO		lvdfind;
	sint32			iResult;

	for (pos1 = pUpdate->nicks.GetHeadPosition(); (pos2 = pos1) != NULL;)
	{
		pCurNick = (Nick*)pUpdate->nicks.GetNext(pos1);
		if (pCurNick->nick == nick)
		{
			if (pUpdate == m_pCurrentChannel)
			{
				lvdfind.flags = LVFI_PARAM;
				lvdfind.lParam = (LPARAM)pCurNick;
				iResult = nickList.FindItem(&lvdfind);
				if (iResult != -1)
				{
					nickList.DeleteItem(iResult);
					UpdateNickCount();
				}
			}
			pUpdate->nicks.RemoveAt(pos2);
			delete pCurNick;
			return true;
		}
	}
	return false;
}

void CIrcWnd::DeleteAllNick(CString channel)
{
	Channel		*pCurChannel = FindChannelByName(channel);

	if (pCurChannel == NULL)
		return;

	POSITION	pos1, pos2;
	Nick		*pCurNick;

	for (pos1 = pCurChannel->nicks.GetHeadPosition(); (pos2 = pos1) != NULL;)
	{
		pCurNick = (Nick*)pCurChannel->nicks.GetNext(pos1);
		pCurChannel->nicks.RemoveAt(pos2);
		delete pCurNick;
	}
}

void CIrcWnd::DeleteNickInAll(CString nick, CString message)
{
	Channel		*pCurChannel;
	CString		strMessage = message, strLine;

	strMessage.Trim();
	for (POSITION pos = channelPtrList.GetHeadPosition(); pos != NULL;)
	{
		pCurChannel = (Channel*)channelPtrList.GetNext(pos);
		if (RemoveNick(pCurChannel->name, nick))
		{
			if (!g_App.m_pPrefs->GetIrcIgnoreInfoMessage())
			{
				strLine.Format(GetResString(IDS_IRC_HASQUIT), nick, strMessage);
				if (strMessage.IsEmpty())
					strLine.Truncate(strLine.GetLength() - 3);
				else
					strLine.Insert(strLine.GetLength() - 1, _T('\x0F'));
				AddMessage(pCurChannel->name, _T(""), RGB(51, 102, 255), strLine);
			}
		}
	}
}

bool CIrcWnd::ChangeNick(CString channel, CString oldnick, CString newnick)
{
	Channel		*pUpdate = FindChannelByName(channel);

	if (pUpdate == NULL)
		return false;

	Nick			*pCurNick;
	LVFINDINFO		lvfind;
	sint32			iResult;

	for (POSITION pos = pUpdate->nicks.GetHeadPosition(); pos != NULL;)
	{
		pCurNick = (Nick*)pUpdate->nicks.GetNext(pos);
		if (pCurNick->nick == oldnick)
		{
			if ((pUpdate = m_pCurrentChannel) != NULL)
			{
				lvfind.flags = LVFI_PARAM;
				lvfind.lParam = (LPARAM)pCurNick;
				iResult = nickList.FindItem(&lvfind);
				if (iResult != -1)
					nickList.SetItemText(iResult, 0, (LPCTSTR)newnick);
			}
			pCurNick->nick = newnick;
			return true;
		}
	}
	return false;
}

bool CIrcWnd::ChangeMode(CString channel, CString nick, CString mode)
{
	Channel		*pUpdate = FindChannelByName(channel);

	if (pUpdate == NULL)
		return false;

	Nick			*pCurNick;
	LVFINDINFO		lvfind;
	sint32			iResult;
	CString			strMode;

	for (POSITION pos = pUpdate->nicks.GetHeadPosition(); pos != NULL;)
	{
		pCurNick = (Nick*)pUpdate->nicks.GetNext(pos);
		if (pCurNick->nick == nick)
		{
			if (mode == _T("+a"))
				pCurNick->protect = _T("*");
			else if (mode == _T("-a"))
				pCurNick->protect.Empty();
			else if (mode == _T("+h"))
				pCurNick->hop = _T("%");
			else if (mode == _T("-h"))
				pCurNick->hop.Empty();
			else if (mode == _T("+o"))
				pCurNick->op = _T("@");
			else if (mode == _T("-o"))
				pCurNick->op.Empty();
			else if (mode == _T("+q"))
				pCurNick->owner = _T("!");
			else if (mode == _T("-q"))
				pCurNick->owner.Empty();
			else if (mode == _T("+u"))
				pCurNick->uop = _T("-");
			else if (mode == _T("-u"))
				pCurNick->uop.Empty();
			else if (mode == _T("+v"))
				pCurNick->voice = _T("+");
			else if (mode == _T("-v"))
				pCurNick->voice.Empty();
			if (pUpdate = m_pCurrentChannel)
			{
				lvfind.flags = LVFI_PARAM;
				lvfind.lParam = (LPARAM)pCurNick;
				iResult = nickList.FindItem(&lvfind);
				if (iResult != -1)
				{
					strMode.Format(_T("%s%s%s%s%s%s"), pCurNick->owner, pCurNick->protect, pCurNick->op, pCurNick->hop, pCurNick->voice, pCurNick->uop);
					nickList.SetItemText(iResult, 1, (LPCTSTR)strMode);
				}
			}
		}
	}
	return true;
}

void CIrcWnd::ParseChangeMode(CString channel, CString changer, CString commands, CString names)
{
	try
	{
		if (commands.GetLength() == 2)
		{
			if (ChangeMode(channel, names, commands))
				if (!g_App.m_pPrefs->GetIrcIgnoreInfoMessage())
					AddMessage(channel, _T(""), RGB(0, 147, 0), GetResString(IDS_IRC_SETSMODE), changer, commands, names);
			return;
		}
		else
		{
			CString		strDir;

			strDir = commands[0];
			if (strDir == _T("+") || strDir == _T("-"))
			{
				int		iCurMode = 1;
				int		iCurName = 0;
				int		iCurNameBack = names.Find(_T(' '), iCurName);

				while (iCurMode < commands.GetLength())
				{
					CString		strTest;

					if(iCurNameBack > iCurName)
					{
						strTest = names.Mid(iCurName, iCurNameBack - iCurName);
						iCurName = iCurNameBack + 1;
					}
					if (ChangeMode(channel, strTest, strDir + commands[iCurMode]) && !g_App.m_pPrefs->GetIrcIgnoreInfoMessage())
						AddMessage(channel, _T(""), RGB(0, 147, 0), GetResString(IDS_IRC_SETSMODE), changer, strDir + commands[iCurMode] , strTest);
					iCurNameBack = names.Find(_T(' '), iCurName + 1);
					if (iCurNameBack == -1)
						iCurNameBack = names.GetLength();
					iCurMode++;
				}
			}
		}
	}
	catch (...)
	{
		AddMessage(channel, _T(""), RGB(252, 127, 0), GetResString(IDS_IRC_NOTSUPPORTED));
	}
}

void CIrcWnd::ChangeAllNick(CString oldnick, CString newnick)
{
	Channel		*pCurChannel = FindChannelByName(oldnick);

	if (pCurChannel != NULL)
	{
		pCurChannel->name = newnick;

		TCITEM		tcitem;
		int			i;

		tcitem.mask = TCIF_PARAM;
		tcitem.lParam = -1;
		for (i = 0; i < channelselect.GetItemCount();i++)
		{
			channelselect.GetItem(i, &tcitem);
			if (((Channel*)tcitem.lParam) == pCurChannel)
				break;
		}
		if (((Channel*)tcitem.lParam) != pCurChannel)
			return;
		tcitem.mask = TCIF_TEXT;
		tcitem.pszText = newnick.GetBuffer();
		tcitem.cchTextMax = newnick.GetLength() + 1;
		channelselect.SetItem(i, &tcitem);
	}
	for (POSITION pos = channelPtrList.GetHeadPosition(); pos != NULL;)
	{
		pCurChannel = (Channel*)channelPtrList.GetNext(pos);
		if (ChangeNick(pCurChannel->name, oldnick, newnick) && !g_App.m_pPrefs->GetIrcIgnoreInfoMessage())
			AddMessage(pCurChannel->name, _T(""), RGB(0, 147, 0), GetResString(IDS_IRC_NOWKNOWNAS), oldnick, newnick);
	}
}

int CIrcWnd::SortProcNick(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	Nick	*pItem1 = (Nick*)lParam1;
	Nick	*pItem2 = (Nick*)lParam2;

	switch (lParamSort)
	{
	case IRC1COL_NICK:
		return pItem1->nick.CompareNoCase(pItem2->nick);
	case IRC1COL_NICK + MLC_SORTDESC:
		return pItem2->nick.CompareNoCase(pItem1->nick);
	case IRC1COL_STATUS:
		if (pItem1->owner == _T("!"))
		{
			if (pItem2->owner != _T("!"))
				return -1;
		}
		else if (pItem2->owner == _T("!"))
		{
			return 1;
		}
		if (pItem1->protect == _T("*"))
		{
			if (pItem2->protect != _T("*"))
				return -1;
		}
		else if (pItem2->protect == _T("*"))
		{
			return 1;
		}
		if (pItem1->op == _T("@"))
		{
			if (pItem2->op != _T("@"))
				return -1;
		}
		else if (pItem2->op == _T("@"))
		{
			return 1;
		}
		if (pItem1->hop == _T("%"))
		{
			if (pItem2->hop != _T("%"))
				return -1;
		}
		else if (pItem2->hop == _T("%"))
		{
			return 1;
		}
		if (pItem1->voice == _T("+"))
		{
			if (pItem2->voice != _T("+"))
				return -1;
		}
		else if (pItem2->voice == _T("+"))
		{
			return 1;
		}
		if (pItem1->uop == _T("-"))
		{
			if (pItem2->uop != _T("-"))
				return -1;
		}
		else if (pItem2->uop == _T("-"))
		{
			return 1;
		}
		return pItem1->nick.CompareNoCase(pItem2->nick);

	case IRC1COL_STATUS + MLC_SORTDESC:
		if (pItem2->owner == _T("!"))
		{
			if (pItem1->owner != _T("!"))
				return -1;
		}
		else if (pItem1->owner == _T("!"))
		{
			return 1;
		}
		if (pItem2->protect == _T("*"))
		{
			if (pItem1->protect != _T("*"))
				return -1;
		}
		else if (pItem1->protect == _T("*"))
		{
			return 1;
		}
		if (pItem2->op == _T("@"))
		{
			if (pItem1->op != _T("@"))
				return -1;
		}
		else if (pItem1->op == _T("@"))
		{
			return 1;
		}
		if (pItem2->hop == _T("%"))
		{
			if (pItem1->hop != _T("%"))
				return -1;
		}
		else if (pItem1->hop == _T("%"))
		{
			return 1;
		}
		if (pItem2->voice == _T("+"))
		{
			if (pItem1->voice != _T("+"))
				return -1;
		}
		else if (pItem1->voice == _T("+"))
		{
			return 1;
		}
		if (pItem2->uop == _T("-"))
		{
			if (pItem1->uop != _T("-"))
				return -1;
		}
		else if (pItem1->uop == _T("-"))
		{
			return 1;
		}
		return pItem2->nick.CompareNoCase(pItem1->nick);
	default:
		return 0;
	}
}

void CIrcWnd::OnColumnClickNick(NMHDR* pNMHDR, LRESULT* pResult)
{
	NM_LISTVIEW		*pNMListView = reinterpret_cast<NM_LISTVIEW*>(pNMHDR);
	int				iSubItem = pNMListView->iSubItem;
	bool			bSortOrder = m_bSortAscendingNickList[iSubItem];

// Reverse sorting direction for the same column and keep the same if column was changed
	if ((nickList.GetSortParam() & MLC_COLUMNMASK) == iSubItem)
		m_bSortAscendingNickList[iSubItem] = bSortOrder = !bSortOrder;

	nickList.SetSortArrow(iSubItem, bSortOrder);
	nickList.SortItems(SortProcNick, iSubItem + ((bSortOrder) ? MLC_SORTASC : MLC_SORTDESC));
	*pResult = 0;
}

void CIrcWnd::OnNMRclickNick(NMHDR *pNMHDR, LRESULT *pResult)
{
	if (nickList.GetSelectionMark() != ( -1))
	{
		CTitleMenu	menuNick;
		POINT		point;

		::GetCursorPos(&point);

		menuNick.CreatePopupMenu();
		menuNick.AddMenuTitle(GetResString(IDS_IRC_NICK));
		menuNick.AppendMenu(MF_STRING, Irc_Priv, GetResString(IDS_IRC_PRIVMESSAGE));
		menuNick.AppendMenu(MF_STRING, Irc_WhoIs, _T("WhoIs"));
	//	menuNick.AppendMenu(MF_STRING,Irc_AddFriend,GetResString(IDS_IRC_ADDTOFRIENDLIST));
	//	if( !GetSendFileString().IsEmpty() )
	//	menuNick.AppendMenu(MF_STRING,Irc_SendLink, (CString)GetResString(IDS_IRC_SENDLINK) + GetSendFileString() );
		menuNick.AppendMenu(MF_STRING, Irc_Slap, GetResString(IDS_IRC_SLAP));
		menuNick.AppendMenu(MF_STRING, Irc_Owner, _T("Owner"));
		menuNick.AppendMenu(MF_STRING, Irc_DeOwner, _T("DeOwner"));
		menuNick.AppendMenu(MF_STRING, Irc_Op, GetResString(IDS_IRC_OP));
		menuNick.AppendMenu(MF_STRING, Irc_DeOp, GetResString(IDS_IRC_DEOP));
		menuNick.AppendMenu(MF_STRING, Irc_HalfOp, GetResString(IDS_IRC_HALFOP));
		menuNick.AppendMenu(MF_STRING, Irc_DeHalfOp, GetResString(IDS_IRC_DEHALFOP));
		menuNick.AppendMenu(MF_STRING, Irc_Voice, GetResString(IDS_IRC_VOICE));
		menuNick.AppendMenu(MF_STRING, Irc_DeVoice, GetResString(IDS_IRC_DEVOICE));
		menuNick.AppendMenu(MF_STRING, Irc_Protect, _T("Protect"));
		menuNick.AppendMenu(MF_STRING, Irc_DeProtect, _T("DeProtect"));
		menuNick.AppendMenu(MF_STRING, Irc_Kick, GetResString(IDS_IRC_KICK));
		menuNick.SetDefaultItem(Irc_Priv);
		menuNick.TrackPopupMenuEx(TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y, this, NULL);

//	Menu objects are destroyed in their destructor
	}
	*pResult = 0;
}

void CIrcWnd::OnNMDblclkNickList(NMHDR *pNMHDR, LRESULT *pResult)
{
	int		iNickItem = nickList.GetSelectionMark();

	if (iNickItem != -1)
	{
		Nick	*pNick = (Nick*)nickList.GetItemData(iNickItem);

		if (pNick != NULL)
			AddMessage(pNick->nick, _T(""), RGB(0, 147, 0), GetResString(IDS_IRC_PRIVATECHANSTART));
	}
	*pResult = 0;
}

/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
//	Messages
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////

void CIrcWnd::AddMessage(CString channelName, CString targetname, COLORREF crTextColor, CString line, ...)
{
	va_list		argptr;
	CString		strTemp;
	Channel		*pUpdateChannel = NULL;

	if (!channelName.IsEmpty())
	{
		pUpdateChannel = FindChannelByName(channelName);
		if (pUpdateChannel == NULL)
			pUpdateChannel = NewChannel(channelName, (channelName[0] == _T('#')) ? 4 : 5);
	}
	else
		pUpdateChannel = (Channel*)channelPtrList.GetHead();

	if (pUpdateChannel == NULL)
		return;

	va_start(argptr, line);
	strTemp.FormatV(line, argptr);
	line = strTemp;
	va_end(argptr);

	CString		strLine = line;
	int			iLen;

	strLine.Replace(_T('\x04'), _T('%'));

	if (channelName.IsEmpty())
	{
		strTemp = m_pIrcMain->GetNick();
		iLen = strTemp.GetLength();
		if (iLen > 0 && strLine.Left(iLen) == strTemp)
		{
			if (strLine.Mid(iLen + 1, 1) == _T(":"))
				iLen ++;
			strLine = strLine.Right(strLine.GetLength() - (iLen + 1));
		}
	}

	if (strLine.IsEmpty())
		return;

	COLORREF	crForegroundColor = crTextColor, crBackgroundColor = CLR_DEFAULT;
	int			iIndex = 0, iLastIndex = 0, iColorIndex;
	DWORD		dwFlags = 0;
	TCHAR		cCurChar;

	if (g_App.m_pPrefs->GetIRCAddTimestamp())
	{
		COleDateTime		currentTime(COleDateTime::GetCurrentTime());

		strTemp.Format(_T("%s: "), currentTime.Format(VAR_TIMEVALUEONLY));
		pUpdateChannel->log.AppendText(strTemp, RGB(0, 0, 128));
	}

	if (!channelName.IsEmpty() && !targetname.IsEmpty())
		pUpdateChannel->log.AppendText(_T("<") + targetname + _T("> "), (m_pIrcMain->GetNick() == targetname) ? RGB(128, 0, 128) : RGB(51, 51, 153));

	while (strLine.GetLength() > iIndex)
	{
		cCurChar = strLine[iIndex];
		switch (cCurChar)
		{
			case 0x02:
			{
				if (iIndex > iLastIndex)
					pUpdateChannel->log.AppendText(strLine.Mid(iLastIndex, iIndex - iLastIndex), crForegroundColor, crBackgroundColor, dwFlags);
				iIndex++;
				iLastIndex = iIndex;
				dwFlags ^= HTC_BOLD;
				break;
			}
			case 0x03:
			{
				if (iIndex > iLastIndex)
					pUpdateChannel->log.AppendText(strLine.Mid(iLastIndex, iIndex - iLastIndex), crForegroundColor, crBackgroundColor, dwFlags);
				iIndex++;

				if (strLine.GetLength() > iIndex && strLine[iIndex] >= _T('0') && strLine[iIndex] <= _T('9'))
				{
					iLen = 1;

					if (strLine.GetLength() > iIndex + 1 && strLine[iIndex + 1] >= _T('0') && strLine[iIndex + 1] <= _T('9'))
						iLen++;

					iColorIndex = _tstoi(strLine.Mid(iIndex, iLen));
					crForegroundColor = (iColorIndex < 16) ? _acrColorPal[iColorIndex] : crTextColor;
					iIndex += iLen;

					if (strLine.GetLength() > iIndex + 1 && strLine[iIndex] == _T(',') && strLine[iIndex + 1] >= _T('0') && strLine[iIndex + 1] <= _T('9'))
					{
						iLen = 1;

						if (strLine.GetLength() > iIndex + 2 && strLine[iIndex + 2] >= _T('0') && strLine[iIndex + 2] <= _T('9'))
							iLen++;

						iColorIndex = _tstoi(strLine.Mid(iIndex + 1, iLen));
						crBackgroundColor = (iColorIndex < 16) ? _acrColorPal[iColorIndex]: CLR_DEFAULT;
						iIndex += 1 + iLen;
					}
				}
				else
				{
					crForegroundColor = crTextColor;
					crBackgroundColor = CLR_DEFAULT;
				}
				iLastIndex = iIndex;
				break;
			}
			case 0x0F:
			{
				if (crForegroundColor != crTextColor || crBackgroundColor != CLR_DEFAULT || dwFlags != 0)
				{
					if (iIndex > iLastIndex)
						pUpdateChannel->log.AppendText(strLine.Mid(iLastIndex, iIndex - iLastIndex), crForegroundColor, crBackgroundColor, dwFlags);
					iIndex++;
					iLastIndex = iIndex;
					crForegroundColor = crTextColor;
					crBackgroundColor = CLR_DEFAULT;
					dwFlags = 0;
				}
				else
					strLine.Delete(iIndex);
				break;
			}
			case 0x1F:
			{
				if (iIndex > iLastIndex)
					pUpdateChannel->log.AppendText(strLine.Mid(iLastIndex, iIndex - iLastIndex), crForegroundColor, crBackgroundColor, dwFlags);
				iIndex++;
				iLastIndex = iIndex;
				dwFlags ^= HTC_UNDERLINE;
				break;
			}
			default:
			{
				if ((TBYTE)cCurChar < _T(' '))
					strLine.Delete(iIndex);
				else
					iIndex++;
			}
		}
	}
	if (iLastIndex < iIndex)
		pUpdateChannel->log.AppendText(strLine.Mid(iLastIndex, iIndex - iLastIndex), crForegroundColor, crBackgroundColor, dwFlags);
	pUpdateChannel->log.AppendText(_T("\n"));

	if (m_pCurrentChannel != pUpdateChannel)
		SetActivity(pUpdateChannel->name, true);
}

void CIrcWnd::SetConnectStatus(bool flag)
{
	if (flag)
	{
		SetDlgItemText(IDC_BN_IRCCONNECT, GetResString(IDS_IRC_DISCONNECT));
		AddMessage(_T(""), _T(""), CLR_DEFAULT, GetResString(IDS_CONNECTED));
		m_bConnected = true;
	}
	else
	{
		SetDlgItemText(IDC_BN_IRCCONNECT, GetResString(IDS_IRC_CONNECT));
		AddMessage(_T(""), _T(""), CLR_DEFAULT, GetResString(IDS_DISCONNECTED));
		m_bConnected = false;
		m_bLoggedIn = false;
		ResetServerChannelList();

		Channel		*pToDel = (Channel*)channelPtrList.GetTail();

		while (channelPtrList.GetCount() > 2)
		{
			pToDel = (Channel*)channelPtrList.GetTail();
			RemoveChannel(pToDel->name);
		}
		GetDlgItem(IDC_CLOSECHAT)->EnableWindow(false);
		GetDlgItem(IDC_CHATSEND)->EnableWindow(false);
	}
}

void CIrcWnd::NoticeMessage(CString source, CString message)
{
	if (source.IsEmpty())
	{
		CString		strLine = message, strChannel;

		if (m_pCurrentChannel->type == 4 || m_pCurrentChannel->type == 5)
		{
			CString		strTemp = m_pIrcMain->GetNick();
			int			iLen = strTemp.GetLength();

			if (iLen > 0 && strLine.Left(iLen) == strTemp)
			{
				if (strLine.Mid(iLen + 1, 1) == _T(":"))
					iLen ++;
				strLine = strLine.Right(strLine.GetLength() - (iLen + 1));
			}
			strChannel = m_pCurrentChannel->name;
		}
		AddMessage(strChannel, _T(""), RGB(255, 80, 80), strLine);
		return;
	}

	bool		bFlag = false;
	Channel		*pCurChannel = FindChannelByName(source);
	Nick		*pCurNick;

	if (pCurChannel != NULL)
	{
		AddMessage(source, _T(""), RGB(153, 51, 102), _T("-%s- %s"), source, message);
		bFlag = true;
	}

	for (POSITION pos = channelPtrList.GetHeadPosition(); pos != NULL;)
	{
		pCurChannel = (Channel*)channelPtrList.GetNext(pos);
		pCurNick = FindNickByName(pCurChannel->name, source);
		if (pCurNick != NULL)
		{
			AddMessage(pCurChannel->name, _T(""), RGB(153, 51, 102), _T("-%s- %s"), source, message);
			bFlag = true;
		}
	}

	if (!bFlag)
		AddMessage((m_pCurrentChannel->type == 4 || m_pCurrentChannel->type == 5) ? m_pCurrentChannel->name : _T(""), _T(""), RGB(153, 51, 102), _T("-%s- %s"), source, message);
}

void CIrcWnd::SetTitle(CString channel, CString title)
{
	Channel		*pCurChannel = FindChannelByName(channel);

	if (pCurChannel == NULL)
		return;

	pCurChannel->title = title;

	if (pCurChannel != m_pCurrentChannel || m_pCurrentChannel->type != 4)
		return;

	titleWindow.Reset();

	CString		strLine = pCurChannel->title;

	if (strLine.IsEmpty())
		return;

	COLORREF	crForegroundColor = CLR_DEFAULT, crBackgroundColor = CLR_DEFAULT;
	int			iIndex = 0, iLastIndex = 0, iColorIndex, iLen;
	DWORD		dwFlags = 0;
	TCHAR		cCurChar;

	while (strLine.GetLength() > iIndex)
	{
		cCurChar = strLine[iIndex];
		switch (cCurChar)
		{
			case 0x02:
			{
				if (iIndex > iLastIndex)
					titleWindow.AppendText(strLine.Mid(iLastIndex, iIndex - iLastIndex), crForegroundColor, crBackgroundColor, dwFlags);
				iIndex++;
				iLastIndex = iIndex;
				dwFlags ^= HTC_BOLD;
				break;
			}
			case 0x03:
			{
				if (iIndex > iLastIndex)
					titleWindow.AppendText(strLine.Mid(iLastIndex, iIndex - iLastIndex), crForegroundColor, crBackgroundColor, dwFlags);
				iIndex++;

				if (strLine.GetLength() > iIndex && strLine[iIndex] >= _T('0') && strLine[iIndex] <= _T('9'))
				{
					iLen = 1;

					if (strLine.GetLength() > iIndex + 1 && strLine[iIndex + 1] >= _T('0') && strLine[iIndex + 1] <= _T('9'))
						iLen++;

					iColorIndex = _tstoi(strLine.Mid(iIndex, iLen));
					crForegroundColor = (iColorIndex < 16) ? _acrColorPal[iColorIndex] : CLR_DEFAULT;
					iIndex += iLen;

					if (strLine.GetLength() > iIndex + 1 && strLine[iIndex] == _T(',') && strLine[iIndex + 1] >= _T('0') && strLine[iIndex + 1] <= _T('9'))
					{
						iLen = 1;

						if (strLine.GetLength() > iIndex + 2 && strLine[iIndex + 2] >= _T('0') && strLine[iIndex + 2] <= _T('9'))
							iLen++;

						iColorIndex = _tstoi(strLine.Mid(iIndex + 1, iLen));
						crBackgroundColor = (iColorIndex < 16) ? _acrColorPal[iColorIndex]: CLR_DEFAULT;
						iIndex += 1 + iLen;
					}
				}
				else
				{
					crForegroundColor = CLR_DEFAULT;
					crBackgroundColor = CLR_DEFAULT;
				}
				iLastIndex = iIndex;
				break;
			}
			case 0x0F:
			{
				if (iIndex > iLastIndex)
					titleWindow.AppendText(strLine.Mid(iLastIndex, iIndex - iLastIndex), crForegroundColor, crBackgroundColor, dwFlags);
				iIndex++;
				iLastIndex = iIndex;
				crForegroundColor = CLR_DEFAULT;
				crBackgroundColor = CLR_DEFAULT;
				dwFlags = 0;
				break;
			}
			case 0x1F:
			{
				if (iIndex > iLastIndex)
					titleWindow.AppendText(strLine.Mid(iLastIndex, iIndex - iLastIndex), crForegroundColor, crBackgroundColor, dwFlags);
				iIndex++;
				iLastIndex = iIndex;
				dwFlags ^= HTC_UNDERLINE;
				break;
			}
			default:
			{
				if ((TBYTE)cCurChar < _T(' '))
					strLine.Delete(iIndex);
				else
					iIndex++;
			}
		}
	}
	if (iLastIndex < iIndex)
		titleWindow.AppendText(strLine.Mid(iLastIndex, iIndex - iLastIndex), crForegroundColor, crBackgroundColor, dwFlags);
}

void CIrcWnd::SetActivity(CString channel, bool flag)
{
	Channel		*pRefresh = FindChannelByName(channel);

	if (pRefresh == NULL)
	{
		pRefresh = (Channel*)channelPtrList.GetHead();
		if (pRefresh == NULL)
			return;
	}
	TCITEM		tcitem;
	int			i;

	tcitem.mask = TCIF_PARAM;
	tcitem.lParam = -1;
	for (i = 0; i < channelselect.GetItemCount();i++)
	{
		channelselect.GetItem(i, &tcitem);
		if (((Channel*)tcitem.lParam) == pRefresh)
			break;
	}
	if (((Channel*)tcitem.lParam) != pRefresh)
		return;
	if (channelselect.GetCurSel() != i)
		channelselect.SetItemState(i, TCIS_HIGHLIGHTED, TCIS_HIGHLIGHTED);
}

void CIrcWnd::OnBnClickedTextBold()
{
	CString		strSend;
	int			iStartChar, iEndChar;

	inputWindow.GetSel(iStartChar, iEndChar);
	inputWindow.GetWindowText(strSend);

	if (iEndChar > iStartChar && iEndChar <= strSend.GetLength())
		strSend.Insert(iEndChar, _T('\x02'));

	if (iStartChar <= strSend.GetLength())
	{
		strSend.Insert(iStartChar, _T('\x02'));
		iStartChar++;
		iEndChar++;
	}

	inputWindow.SetWindowText(strSend);
	inputWindow.SetSel(iStartChar, iEndChar);
	inputWindow.SetFocus();
}

void CIrcWnd::OnBnClickedTextUnderline()
{
	CString		strSend;
	int			iStartChar, iEndChar;

	inputWindow.GetSel(iStartChar, iEndChar);
	inputWindow.GetWindowText(strSend);

	if (iEndChar > iStartChar && iEndChar <= strSend.GetLength())
		strSend.Insert(iEndChar, _T('\x1F'));

	if (iStartChar <= strSend.GetLength())
	{
		strSend.Insert(iStartChar, _T('\x1F'));
		iStartChar++;
		iEndChar++;
	}

	inputWindow.SetWindowText(strSend);
	inputWindow.SetSel(iStartChar, iEndChar);
	inputWindow.SetFocus();
}

void CIrcWnd::OnBnClickedTextColor()
{
    CRect		rColorButton;

	m_ctrlTextColorBtn.GetWindowRect(&rColorButton);

	new CColourPopup(CPoint(rColorButton.right - 149, rColorButton.top - 68), CLR_DEFAULT, this,
		GetResString(IDS_DEFAULT), NULL, true);

	inputWindow.SetFocus();
}

LONG CIrcWnd::OnColorSelEndOK(WPARAM wParam, LPARAM lParam)
{
    CWnd			*pParent = GetParent();

    if (pParent != NULL)
	{
        pParent->SendMessage(CPN_CLOSEUP, lParam, (WPARAM)GetDlgCtrlID());
        pParent->SendMessage(CPN_SELENDOK, lParam, (WPARAM)GetDlgCtrlID());
    }

	CString		strSend, strColor = _T("\003");
	int			iStartChar, iEndChar;

	inputWindow.GetSel(iStartChar, iEndChar);
	inputWindow.GetWindowText(strSend);

	if (iEndChar > iStartChar && iEndChar <= strSend.GetLength())
		strSend.Insert(iEndChar, strColor);

	if (iStartChar <= strSend.GetLength())
	{
		if ((int)lParam >= 0 && (int)lParam < 16)
			strColor.AppendFormat(_T("%i"), (int)lParam);
		strSend.Insert(iStartChar, strColor);
		iStartChar += strColor.GetLength();
		iEndChar += strColor.GetLength();
	}

	inputWindow.SetWindowText(strSend);
	inputWindow.SetSel(iStartChar, iEndChar);

    return true;
}

LONG CIrcWnd::OnColorSelEndCancel(WPARAM wParam, LPARAM /*lParam*/)
{
	CWnd		*pParent = GetParent();

    if (pParent != NULL)
	{
        pParent->SendMessage(CPN_CLOSEUP, wParam, (WPARAM)GetDlgCtrlID());
        pParent->SendMessage(CPN_SELENDCANCEL, wParam, (WPARAM)GetDlgCtrlID());
    }

    return true;
}

void CIrcWnd::OnBnClickedChatsend()
{
	inputWindow.SetFocus();

	if (!this->m_bConnected || m_pCurrentChannel == NULL || m_pCurrentChannel->log.m_hWnd == NULL)
		return;

	CString		strSend;

	inputWindow.GetWindowText(strSend);
	inputWindow.SetWindowText(_T(""));

	if (strSend.IsEmpty())
		return;

	if (m_pCurrentChannel->history.GetCount() == g_App.m_pPrefs->GetMaxChatHistoryLines())
		m_pCurrentChannel->history.RemoveAt(0);
	m_pCurrentChannel->history.Add(strSend);
	m_pCurrentChannel->history_pos = m_pCurrentChannel->history.GetCount();

	CString		strTemp;
	int			iIndex;

	if ((GetKeyState(VK_CONTROL) & 0x8000) == 0 && strSend[0] == _T('/'))
	{
		strSend.TrimLeft(_T("/ "));
		iIndex = strSend.Find(_T(" "));

		if (iIndex <= 0)
		{
			m_pIrcMain->SendString(strSend);
			return;
		}

		strTemp = strSend.Mid(iIndex);
		strTemp.Trim();

		if (strTemp.IsEmpty())
			return;

		strSend.Truncate(iIndex);
		strSend.MakeLower();

		if (strSend == _T("hop"))
		{
			if (m_pCurrentChannel->type != 4)
				return;

			m_pIrcMain->SendString(_T("PART ") + m_pCurrentChannel->name);
			m_pIrcMain->SendString(_T("JOIN ") + strTemp);
			return;
		}
		else if (strSend == _T("msg") || strSend == _T("privmsg"))
		{
			iIndex = strTemp.Find(_T(" "));

			if (iIndex <= 0)
				return;

			strSend = strTemp.Mid(iIndex);
			strSend.Trim();
			strTemp.Truncate(iIndex);
			strTemp.Trim();

			if (strSend.IsEmpty() || strTemp.IsEmpty())
				return;

			if (strTemp.CompareNoCase(_T("nickserv")) == 0)
				strTemp = _T("ns ");
			else if (strTemp.CompareNoCase(_T("chanserv")) == 0)
				strTemp = _T("cs ");
			else
			{
				strTemp.Insert(0, _T("PRIVMSG "));
				strTemp += _T(" :");
			}
			strTemp += strSend;
			strSend.Replace(_T('%'), _T('\x04'));
			strSend.Insert(0, _T("* >> "));
			AddMessage((m_pCurrentChannel->type == 4 || m_pCurrentChannel->type == 5) ? m_pCurrentChannel->name : _T(""), _T(""), RGB(153, 51, 102), strSend);
			m_pIrcMain->SendString(strTemp);
			return;
		}
		else if (strSend == _T("topic"))
		{
			iIndex = strTemp.Find(_T(" "));

			if ((iIndex <= 0 || strTemp[iIndex] != _T('#')) && m_pCurrentChannel->type == 4)
			{
				strSend = strTemp;
				strTemp = m_pCurrentChannel->name;
			}
			else if (iIndex > 0)
			{
				strSend = strTemp.Mid(iIndex);
				strSend.Trim();
				strTemp.Truncate(iIndex);
				strTemp.Trim();
			}
			else
				return;

			if (strSend.IsEmpty() || strTemp.IsEmpty() || strTemp[0] != _T('#'))
				return;

			strTemp.Insert(0, _T("TOPIC "));
			strTemp += _T(" :");
			strTemp += strSend;
			m_pIrcMain->SendString(strTemp);
			return;
		}
		else if (strSend == _T("me"))
		{
			if (m_pCurrentChannel->type != 4 && m_pCurrentChannel->type != 5)
				return;

			strSend.Format(_T("PRIVMSG %s :\001ACTION %s\001"), m_pCurrentChannel->name, strTemp);
			strTemp.Replace(_T('%'), _T('\x04'));
			AddMessage(m_pCurrentChannel->name, _T(""), RGB(0, 147, 0), _T("* %s %s"), m_pIrcMain->GetNick(), strTemp);
			m_pIrcMain->SendString(strSend);
			return;
		}
		else
		{
			strSend += _T(" ");
			strSend += strTemp;
			m_pIrcMain->SendString(strSend);
			return;
		}
	}

	if (m_pCurrentChannel->type < 4)
	{
		m_pIrcMain->SendString(strSend);
		return;
	}

	strTemp.Format(_T("PRIVMSG %s :%s"), m_pCurrentChannel->name, strSend);
	m_pIrcMain->SendString(strTemp);
	strSend.Replace(_T('%'), _T('\x04'));
	AddMessage(m_pCurrentChannel->name, m_pIrcMain->GetNick(), CLR_DEFAULT, strSend);
}

/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
//	Channels
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////

Channel* CIrcWnd::FindChannelByName(CString name)
{
	Channel		*pCurChannel;

	for (POSITION pos = channelPtrList.GetHeadPosition(); pos != NULL;)
	{
		pCurChannel = (Channel*)channelPtrList.GetNext(pos);
		if (pCurChannel->name.CompareNoCase(name.Trim()) == 0 && (pCurChannel->type == 4 || pCurChannel->type == 5))
			return pCurChannel;
	}
	return 0;
}

Channel* CIrcWnd::NewChannel(CString name, byte type)
{
	Channel		*pToAdd = new Channel;
	TCITEM		tcitem;

	pToAdd->name = name;
	pToAdd->title = name;
	pToAdd->type = type;
	pToAdd->history_pos = 0;
	if (type != 2)
	{
		CRect		rcChannel;

		serverChannelList.GetWindowRect(&rcChannel);
		ScreenToClient(&rcChannel);
		if (pToAdd->type == 4)
			rcChannel.top += 18;
		pToAdd->log.Create(WS_CHILD | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_HSCROLL | WS_VSCROLL | ES_MULTILINE | ES_READONLY | ES_NOHIDESEL, rcChannel, this, (UINT)-1);
		pToAdd->log.ModifyStyleEx(0, WS_EX_STATICEDGE, SWP_FRAMECHANGED);
		pToAdd->log.SendMessage(EM_SETMARGINS, EC_LEFTMARGIN | EC_RIGHTMARGIN, MAKELONG(3, 3));
		pToAdd->log.SetEventMask(pToAdd->log.GetEventMask() | ENM_MOUSEEVENTS);
		pToAdd->log.SetFont(&g_App.m_pMDlg->m_fontDefault);
		pToAdd->log.SetTitle(name);
		pToAdd->log.m_dwFlags = (HTC_ISAUTOSCROLL | HTC_ISWORDWRAP);
		pToAdd->log.SetTargetDevice(NULL, 0);
	}
	channelPtrList.AddTail(pToAdd);

	tcitem.mask = TCIF_PARAM | TCIF_TEXT;
	if (type >= 4)
		tcitem.mask |= TCIF_IMAGE;
	tcitem.lParam = (LPARAM)pToAdd;
	tcitem.pszText = name.GetBuffer();
	tcitem.iImage = 0;
	uint32 pos = channelselect.GetItemCount();

	channelselect.InsertItem(pos, &tcitem);
	if (type >= 4)
	{
		channelselect.SetCurSel(pos);
		channelselect.SetCurFocus(pos);
		OnTcnSelchangeTab2(NULL, NULL);
		GetDlgItem(IDC_CLOSECHAT)->EnableWindow(true);
		GetDlgItem(IDC_CHATSEND)->EnableWindow(true);
	}

	return pToAdd;
}

void CIrcWnd::RemoveChannel(CString channel)
{
	Channel		*pToDel = FindChannelByName(channel);

	if (pToDel == NULL)
		return;

	TCITEM		tcitem;
	int			i;

	tcitem.mask = TCIF_PARAM;
	tcitem.lParam = -1;

	for (i = 0; i < channelselect.GetItemCount();i++)
	{
		channelselect.GetItem(i, &tcitem);
		if (((Channel*)tcitem.lParam) == pToDel)
			break;
	}

	if (((Channel*)tcitem.lParam) != pToDel)
		return;

	channelselect.DeleteItem(i);

	if (pToDel == m_pCurrentChannel)
	{
		nickList.DeleteAllItems();
		if (channelselect.GetItemCount() > 2 && i > 1)
		{
			if (i == 2)
				i++;
			channelselect.SetCurSel(i - 1);
			channelselect.SetCurFocus(i - 1);
			OnTcnSelchangeTab2(NULL, NULL);
		}
		else
		{
			channelselect.SetCurSel(0);
			channelselect.SetCurFocus(0);
			OnTcnSelchangeTab2(NULL, NULL);
			GetDlgItem(IDC_CLOSECHAT)->EnableWindow(false);
			GetDlgItem(IDC_CHATSEND)->EnableWindow(false);
		}
	}
	DeleteAllNick(pToDel->name);
	channelPtrList.RemoveAt(channelPtrList.Find(pToDel));
	delete pToDel;
}

void CIrcWnd::DeleteAllChannel()
{
	POSITION	pos1, pos2;
	Channel		*pChannel;

	for (pos1 = channelPtrList.GetHeadPosition();(pos2 = pos1) != NULL;)
	{
		pChannel = (Channel*)channelPtrList.GetNext(pos1);
		DeleteAllNick(pChannel->name);
		channelPtrList.RemoveAt(pos2);
		delete pChannel->log;
		delete pChannel;
	}
}

void CIrcWnd::JoinChannels()
{
	if (!this->m_bConnected)
		return;

	int			iIndex = -1;
	POSITION	pos = serverChannelList.GetFirstSelectedItemPosition();

	while (pos != NULL)
	{
		iIndex = serverChannelList.GetNextSelectedItem(pos);
		if (iIndex > -1)
			m_pIrcMain->SendString(_T("JOIN ") + serverChannelList.GetItemText(iIndex, 0));
	}
}

LRESULT CIrcWnd::OnCloseTab(WPARAM wparam, LPARAM lparam)
{
	OnBnClickedClosechat((int)wparam);
	return true;
}

void CIrcWnd::SendString(CString send)
{
	if (this->m_bConnected)
		m_pIrcMain->SendString(send);
}

void CIrcWnd::ScrollHistory(bool down)
{
	if ((m_pCurrentChannel->history_pos == 0 && !down) || (m_pCurrentChannel->history_pos == m_pCurrentChannel->history.GetCount() && down))
		return;

	if (down)
		++m_pCurrentChannel->history_pos;
	else
		--m_pCurrentChannel->history_pos;

	CString		strBuffer;

	if (m_pCurrentChannel->history_pos != m_pCurrentChannel->history.GetCount())
		strBuffer = m_pCurrentChannel->history.GetAt(m_pCurrentChannel->history_pos);

	inputWindow.SetWindowText(strBuffer);
	inputWindow.SetSel(strBuffer.GetLength(), strBuffer.GetLength());
}
