//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include "PartFile.h"
#ifndef NEW_SOCKETS_ENGINE
	#include "emule.h"
	#include "updownclient.h"
#else
	#include "SafeFile.h"
	#include "Engine/Data/Prefs.h"
#endif //NEW_SOCKETS_ENGINE
#include <math.h>
#include "ED2KLink.h"
#include "Preview.h"
#include "ini2.h"
#include "ArchiveRecovery.h"
#include <sys/stat.h>
#include <io.h>
#include "Smtp.h"
#include "MMServer.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif

#define PROGRESS_HEIGHT 4

CPartFile::CPartFile()
{
	Init();
}

CPartFile::CPartFile(CSearchFile* searchresult, EnumCategories eCatID)
{
	EMULE_TRY
#ifndef NEW_SOCKETS_ENGINE
	Init();

	md4cpy(m_fileHash, searchresult->GetFileHash());
	for (int i = 0; i < searchresult->m_tagArray.GetCount(); i++)
	{
		const CTag *pTag = searchresult->m_tagArray[i];

		switch (pTag->GetTagID())
		{
			case FT_FILENAME:
				if (pTag->IsStr())
				{
#ifdef _UNICODE
					if (IsFileNameEmpty())
#endif
						SetFileName(pTag->GetStringValue());
				}
				break;

			case FT_FILESIZE:
				if (pTag->IsInt())
					SetFileSize(pTag->GetIntValue());
				break;

			case FT_SOURCES:
			case FT_COMPLETE_SOURCES:
			case FT_LASTSEENCOMPLETE:
				break;

			default:
			{
				CTag	*pNewTag = new CTag(*pTag);
				m_tagArray.Add(pNewTag);
				break;
			}
		}
	}
//	Don't ask for a hashset, if there is none.
	if (GetFileSize() < PARTSIZE)
		m_bHashSetNeeded = false;

//	Set Category, don't save data yet
	if (eCatID == CAT_NONE)
		g_App.m_pDownloadQueue->SetAutoCat(this);
	else
		SetCatID(eCatID);

	CreatePartFile();

#endif //NEW_SOCKETS_ENGINE
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CPartFile::CPartFile(CString edonkeylink, EnumCategories eCatID)
{
	EMULE_TRY

#ifndef NEW_SOCKETS_ENGINE
	CED2KLink* pLink = NULL;
	try
	{
		pLink = CED2KLink::CreateLinkFromUrl(edonkeylink);
		_ASSERT(pLink != NULL);
		CED2KFileLink* pFileLink = pLink->GetFileLink();
		if (pFileLink == NULL)
			throw CString(_T("not an ed2k server or file link"));
		InitializeFromLink(pFileLink, eCatID);
	}
	catch (CString error)
	{
		OUTPUT_DEBUG_TRACE();
		AddLogLine(true, RGB_LOG_WARNING + GetResString(IDS_ERR_INVALIDLINK), error);
		SetStatus(PS_ERROR);
	}
	delete pLink;
#endif //NEW_SOCKETS_ENGINE

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CPartFile::CPartFile(CED2KFileLink* fileLink, EnumCategories eCatID)
{
	InitializeFromLink(fileLink, eCatID);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::InitializeFromLink(CED2KFileLink* fileLink, EnumCategories eCatID)
{
	EMULE_TRY

	Init();

	try
	{
		SetFileName(fileLink->GetName());
		SetFileSize(fileLink->GetSize());
	//	Prevent to donload an empty file
		if (!GetFileSize())
		{
#ifndef NEW_SOCKETS_ENGINE
			AddLogLine(true, RGB_LOG_WARNING + GetResString(IDS_SKIPZEROLENGTHFILE), m_strFileName);
#endif //NEW_SOCKETS_ENGINE
			SetStatus(PS_ERROR);
			return ;
		}
	//	Don't ask for a hashset, if there is none
		if (GetFileSize() < PARTSIZE)
			m_bHashSetNeeded = false;
		md4cpy(m_fileHash, fileLink->GetHashKey());
#ifndef NEW_SOCKETS_ENGINE
		if (!g_App.m_pDownloadQueue->FileExists(m_fileHash))
		{
		//	Set Category, don't save data yet
			if (eCatID == CAT_NONE)
				g_App.m_pDownloadQueue->SetAutoCat(this);
			else
				SetCatID(eCatID);

			CreatePartFile();
		}
		else
#endif //NEW_SOCKETS_ENGINE
			SetStatus(PS_ERROR);
	}
	catch (CString error)
	{
		OUTPUT_DEBUG_TRACE();
#ifndef NEW_SOCKETS_ENGINE
		AddLogLine(true, RGB_LOG_WARNING + GetResString(IDS_ERR_INVALIDLINK), error);
#endif //NEW_SOCKETS_ENGINE
		SetStatus(PS_ERROR);
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::Init()
{
	EMULE_TRY

	m_strTempDir = _T("");
	m_dwLastFileSourcesRequestTime = 0;
	m_nLastBufferFlushTime = m_dwLastPurgeTime = ::GetTickCount();
	m_bPaused = false;
	m_bStopped = false;
	SetStatus(PS_EMPTY);
	m_dwBytesTransferred = 0;
	priority = PR_NORMAL;
	m_iNumTransferringSrcs = 0;
	m_iSrcNNP = 0;
	m_iSrcOnQueue = 0;
	m_uSrcHighQR = 0;
	m_iSrcConnecting = 0;
	m_uSrcWaitForFileReq = 0;
	m_iSrcConnected = 0;
	m_iSrcConnViaServer = 0;
	m_uSrcLowToLow = 0;
	m_uSrcLowIDOnOtherServer = 0;
	m_iSrcQueueFull = 0;
	m_iSrcA4AF = 0;
	m_dwDataRate = 0;
	m_bHashSetNeeded = true;
	m_uProcessCounter = 0;
	m_dblPercentCompleted = 0.0;
	m_strPartMetFileName = _T("");
	m_dwCompletedSize = 0;
	m_bPreviewing = false;
	lastseencomplete = NULL;
	m_dwAvailablePartsCount = 0;
	m_ClientSrcAnswered = 0;
	m_LastNoNeededCheck = 0;
	m_eRating = PF_RATING_NONE;
	m_nTotalBufferData = 0;
	m_bRecoveringArchive = false;
	m_qwGainDueToCompression = 0;
	m_qwLostDueToCorruption = 0;
	m_iTotalPacketsSavedDueToICH = 0;
	m_bHasRating = false;
	m_bHasComment = false;
	m_uLastCompleteSrcCount = 0;
	m_SessionStartTime = m_timeLastDownTransfer = CTime::GetCurrentTime();
	m_SessionStartSize = 0;
	m_AvgDataRate = 0;
	m_StartTimeReset = false;
	m_eCategoryID = CAT_NONE;
	m_bIsBeingDeleted = false;
	m_bIsPreallocated = false;
//	Async data flush (i.e. only in process)
	m_bDataFlushReq = false;
//	Set to true for fakes.rar download
	m_bIsFakesDotRar = false;
//	Set intial gaps sum to 0
	m_dwGapsSum = 0;
//	Set total size of completed parts to 0
	m_dwCompletedPartsSize = 0;

#ifndef NEW_SOCKETS_ENGINE
	if (g_App.m_pPrefs->IsDAPEnabled())
		m_bAutoPriority = true;
	else
#endif //NEW_SOCKETS_ENGINE
		m_bAutoPriority = false;

	m_bIsPreviewableArchive = false;
	m_bIsMovie = false;
	m_bIsAviMovie = false;
	m_bIsMpgMovie  = false;
	m_bIsMpgAudio = false;

	::InitializeCriticalSection(&m_csSourceLists);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CPartFile::~CPartFile()
{
	EMULE_TRY

	if (!m_bIsBeingDeleted && (m_hPartFileWrite.m_hFile != INVALID_HANDLE_VALUE))
	{
	//	Ensure all buffered data is written
		FlushBuffer();

		ClosePartFile();
	//	Update met file
		SavePartFile();
	}

	m_srcPartFrequencies.RemoveAll();
	m_PartsStatusVector.clear();

//	Will be unlocked on exit
	CSingleLock lockGap(&m_csGapListAndPartStatus, TRUE);

	while (!gaplist.IsEmpty())
	{
		delete gaplist.RemoveHead();
	}

	::DeleteCriticalSection(&m_csSourceLists);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::CreatePartFile()
{
	EMULE_TRY

#ifndef NEW_SOCKETS_ENGINE
	m_strTempDir = g_App.m_pPrefs->GetTempDir();
#else
	m_strTempDir = g_stEngine.Prefs.GetTempDir();
#endif //NEW_SOCKETS_ENGINE

	if (GetCatID() != CAT_NONE)
	{
		CString strCatTempDir =	CCat::GetCatByID(GetCatID())->GetTempPath();
		if (!strCatTempDir.IsEmpty())
			m_strTempDir = strCatTempDir;
	}

	if (m_strTempDir.Right(1) == _T('\\'))
	{
		m_strTempDir.Truncate(m_strTempDir.GetLength() - 1);
	}

	SetPath(m_strTempDir);

	int iPartNumber = 0;
	CString strFullPartFileName;

	do
	{
		strFullPartFileName.Format(_T("%s\\%03u.part"), m_strTempDir, ++iPartNumber);
	}
	while (PathFileExists(strFullPartFileName));

	m_strPartMetFileName.Format(_T("%03u.part.met"), iPartNumber);
	m_strFullName.Format(_T("%s.met"), strFullPartFileName);
	SetFilePath(strFullPartFileName);

	CString	buffer = m_strPartMetFileName.Left(m_strPartMetFileName.GetLength() - 4);
	CTag	*pPartNameTag = new CTag(FT_PARTFILENAME, buffer);

	m_tagArray.Add(pPartNameTag);

	if (!m_hPartFileWrite.Open(strFullPartFileName, CFile::modeCreate | CFile::modeWrite | CFile::shareDenyWrite | CFile::osSequentialScan))
	{
		AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_CREATEPARTFILE));
		SetStatus(PS_ERROR);
	}
	if (!m_hPartFileRead.Open(strFullPartFileName, CFile::modeRead | CFile::shareDenyNone | CFile::osSequentialScan))
	{
		m_hPartFileWrite.Close();
		AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_CREATEPARTFILE));
		SetStatus(PS_ERROR);
	}

	const uint32 dwPartCnt = GetPartCount();

	m_srcPartFrequencies.SetSize(dwPartCnt);
	m_PartsStatusVector.resize(dwPartCnt);

	for (uint32 i = 0; i < dwPartCnt; i++)
	{
		m_srcPartFrequencies[i] = 0;
		m_PartsStatusVector[i] = 0;
	}

//	Note: the gap should be initialized after parts status vector was filled
	AddGap(0, GetFileSize() - 1);

	m_bPaused = false;

#ifndef NEW_SOCKETS_ENGINE
//	Before we save file we need to establish UAP
	if (g_App.m_pPrefs->IsUAPEnabled())
	{
		SetAutoULPriority(true);
		SetULPriority(PR_RELEASE);
		UpdateSharedFileDisplay();
	}
#endif //NEW_SOCKETS_ENGINE

//	Don't request hashset for file which size < PARTSIZE
//	NB! file with size = PARTSIZE has hashset consisted of two hashes
	if (GetFileSize() < PARTSIZE)
	{
		m_bHashSetNeeded = false;
	}

#ifndef NEW_SOCKETS_ENGINE
//	Filename cleanup
	if (g_App.m_pPrefs->GetAutoFilenameCleanup())
	{
		SetFileName(CleanupFilename(GetFileName()).GetBuffer());
	}
#endif //NEW_SOCKETS_ENGINE

	SavePartFile();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::LoadPartFile(LPCTSTR in_directory, LPCTSTR in_filename)
{
	EMULE_TRY

	CMap<uint16, uint16, Gap_Struct*, Gap_Struct*> gap_map;
	CSafeBufferedFile file;
	CString	strCorruptedParts;

	m_dwBytesTransferred = 0;
	m_strPartMetFileName = in_filename;
	m_strTempDir = in_directory;
	if (m_strTempDir.Right(1) == _T('\\'))
		m_strTempDir.Truncate(m_strTempDir.GetLength() - 1);
	SetPath(m_strTempDir);

	m_strFullName.Format(_T("%s\\%s"), m_strTempDir, m_strPartMetFileName);
	SetFilePath(m_strFullName.Left(m_strFullName.GetLength() - 4));

	try
	{
	//	Read file data form part.met file
		if (!file.Open(m_strFullName, CFile::modeRead | CFile::typeBinary | CFile::shareDenyWrite))
		{
#ifndef NEW_SOCKETS_ENGINE
			AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_OPENMET), m_strPartMetFileName, m_strFileName);
			g_App.m_pMDlg->DisableAutoBackup();
#endif //NEW_SOCKETS_ENGINE
			return false;
		}
		byte version;
		file.Read(&version, 1);
		if (version != PARTFILE_VERSION)
		{
			file.Close();
#ifndef NEW_SOCKETS_ENGINE
			AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_BADMETVERSION), m_strPartMetFileName, m_strFileName);
			g_App.m_pMDlg->DisableAutoBackup();
#endif //NEW_SOCKETS_ENGINE
			return false;
		}
		LoadDateFromFile(file);
		if (!LoadHashsetFromFile(file, false))
		{
#ifndef NEW_SOCKETS_ENGINE
			AddLogLine(true, RGB_LOG_WARNING + GetResString(IDS_ERR_HASHSETCORR), in_filename);
#endif //NEW_SOCKETS_ENGINE
		}
		LoadSettingsFile();

		uint32	dwTagCnt;

		file.Read(&dwTagCnt, 4);
		for (uint32 j = 0; j < dwTagCnt; j++)
		{
			CTag* newtag = new CTag();

			newtag->FillFromStream(file);
			switch (newtag->GetTagID())
			{
				case FT_FILENAME:
					if (!newtag->IsStr())
					{
#ifndef NEW_SOCKETS_ENGINE
						AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_METCORRUPT), m_strPartMetFileName, m_strFileName);
						delete newtag;
						g_App.m_pMDlg->DisableAutoBackup();
#endif //NEW_SOCKETS_ENGINE

						return false;
					}
#ifdef _UNICODE
					if (IsFileNameEmpty())
#endif
						SetFileName(newtag->GetStringValue());
					break;

				case FT_LASTSEENCOMPLETE:
					if (newtag->IsInt())
						lastseencomplete = newtag->GetIntValue();
					break;

				case FT_FILESIZE:
					if (newtag->IsInt())
						SetFileSize(newtag->GetIntValue());
					break;

				case FT_TRANSFERRED:
					if (newtag->IsInt())
						m_dwBytesTransferred = newtag->GetIntValue();
					break;

				case FT_COMPRESSION:
					if (newtag->IsInt())
						m_qwGainDueToCompression = newtag->GetIntValue();
					break;

				case FT_CORRUPTED:
					if (newtag->IsInt())
						m_qwLostDueToCorruption = newtag->GetIntValue();
					break;

				case FT_DLPRIORITY:
					if (newtag->IsInt())
					{
						if ((priority = newtag->GetIntValue()) == PR_AUTO)
						{
							SetAutoPriority(true);
							priority = PR_HIGH;
						}
						else
						{
							SetAutoPriority(false);
							if ((priority != PR_LOW) && (priority != PR_NORMAL) && (priority != PR_HIGH))
								priority = PR_NORMAL;
						}
					}
					break;

				case FT_STATUS:
					if (newtag->IsInt())
						m_bPaused = newtag->GetIntValue();
					break;

				case FT_ULPRIORITY:
					if (newtag->IsInt())
					{
						byte	byteULPrio = static_cast<byte>(newtag->GetIntValue());

						if (byteULPrio == PR_AUTO)
						{
							SetAutoULPriority(true);
							SetULPriority(PR_RELEASE);
						}
						else
						{
							SetAutoULPriority(false);
							if ( (byteULPrio != PR_VERYLOW) && (byteULPrio != PR_LOW) &&
								(byteULPrio != PR_NORMAL) && (byteULPrio != PR_HIGH) && (byteULPrio != PR_RELEASE) )
							{
								byteULPrio = PR_NORMAL;
							}
							SetULPriority(byteULPrio);
						}
					}
					break;

				case FT_CATEGORY:
					if (newtag->IsInt())
						m_eCategoryID = static_cast<_EnumCategories>(newtag->GetIntValue());
					break;

				case FT_PERMISSIONS:
					if (newtag->IsInt() && (newtag->GetIntValue() <= PERM_NOONE))
						SetPermissions((byte)newtag->GetIntValue());
					break;

				case FT_ATTRANSFERRED:						// (FT_ATTRANSFERRED:int) all time transferred (low long)
					if (newtag->IsInt())
						statistic.alltimetransferred = (statistic.alltimetransferred & (0xFFFFFFFFi64 << 32)) | newtag->GetIntValue();
					break;

				case FT_ATTRANSFERREDHI:					// (FT_ATTRANSFERREDHI:int) all time transferred (high long)
					if (newtag->IsInt())
						statistic.alltimetransferred = (statistic.alltimetransferred & 0xFFFFFFFF) |
							(static_cast<uint64>(newtag->GetIntValue()) << 32);
					break;

				case FT_ATREQUESTED:						// (FT_ATREQUESTED:int) all time requested
					if (newtag->IsInt())
						statistic.alltimerequested = newtag->GetIntValue();
					break;

				case FT_ATACCEPTED:							// (FT_ATACCEPTED:int) all time accepted
					if (newtag->IsInt())
						statistic.alltimeaccepted = newtag->GetIntValue();
					break;

				case FT_CORRUPTEDPARTS:
					if (newtag->IsStr())
						newtag->GetStringValue(&strCorruptedParts);
					break;

				default:
				{
					const char	*pcTagName = newtag->GetTagName();

					if ((newtag->GetTagID() == TAGTYPE_NONE) && (pcTagName == NULL))
						break;
					if ((pcTagName != NULL) && (pcTagName[0] == FT_GAPSTART || pcTagName[0] == FT_GAPEND))
					{
						if (newtag->IsInt())
						{
							Gap_Struct *gap;
							uint16 gapkey = atoi(&pcTagName[1]);

							if (!gap_map.Lookup(gapkey, gap))
							{
								gap = new Gap_Struct;
								gap_map.SetAt(gapkey, gap);
								gap->m_dwStartOffset = ~0u;
								gap->m_dwEndOffset = ~0u;
							}
							if (pcTagName[0] == FT_GAPSTART)
								gap->m_dwStartOffset = newtag->GetIntValue();
							if (pcTagName[0] == FT_GAPEND)
								gap->m_dwEndOffset = newtag->GetIntValue() - 1;
						}
					}
					else
					{
						m_tagArray.Add(newtag);
						continue;
					}
				}
			}
			delete newtag;
		}
		file.Close();

		CStdioFile	f;
		CString		strPath(m_strFilePath);

		strPath += _T(".dir");
		if (f.Open(strPath, CFile::modeRead | CFile::shareDenyWrite | CFile::typeText))
		{
			f.ReadString(m_strAlternativePath);
			f.Close();
		}
	}
	catch (CFileException * error)
	{
		OUTPUT_DEBUG_TRACE();
#ifndef NEW_SOCKETS_ENGINE
		if (error->m_cause == CFileException::endOfFile)
			AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_METCORRUPT), m_strPartMetFileName, m_strFileName);
		else
		{
			AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_FILEERROR), m_strPartMetFileName, m_strFileName, GetErrorMessage(error));
		}
#endif //NEW_SOCKETS_ENGINE
		error->Delete();
		for (POSITION pos = gap_map.GetStartPosition(); pos != NULL;)
		{
			Gap_Struct* gap;
			uint16 gapkey;
			gap_map.GetNextAssoc(pos, gapkey, gap);
			delete gap;
		}
#ifndef NEW_SOCKETS_ENGINE
		g_App.m_pMDlg->DisableAutoBackup();
#endif //NEW_SOCKETS_ENGINE

		return false;
	}

//	At this point, the file size was defined. initialize related variables
	const uint32 dwPartCnt = GetPartCount();

	m_srcPartFrequencies.SetSize(dwPartCnt);
	m_PartsStatusVector.resize(dwPartCnt);

	for (uint32 i = 0; i < dwPartCnt; i++)
	{
		m_srcPartFrequencies[i] = 0;
		m_PartsStatusVector[i] = PART_VERIFIED;	// AddGap will reset this where required
	}

//	Now to flush the map into the list
	for (POSITION pos = gap_map.GetStartPosition(); pos != NULL;)
	{
		Gap_Struct* gap;
		uint16 gapkey;
		gap_map.GetNextAssoc(pos, gapkey, gap);

		if (gap->m_dwStartOffset != -1 && gap->m_dwEndOffset != -1 && gap->m_dwStartOffset <= gap->m_dwEndOffset && gap->m_dwStartOffset < GetFileSize())
		{
			if (gap->m_dwEndOffset >= GetFileSize())
				gap->m_dwEndOffset = GetFileSize() - 1; 		// Clipping
			AddGap(gap->m_dwStartOffset, gap->m_dwEndOffset); 	// All tags accounted for, use safe adding
		}
		delete gap;
	}

	if (!strCorruptedParts.IsEmpty())
	{
		CString	strPart;
		uint32	dwPart;

		for (int iPos = 0;;)
		{
			strPart = strCorruptedParts.Tokenize(_T(","), iPos);
			if (strPart.IsEmpty())
				break;

			if (_stscanf(strPart, _T("%u"), &dwPart) == 1)
			{
				if (dwPart < dwPartCnt)
				{
					m_csGapListAndPartStatus.Lock();
					m_PartsStatusVector[dwPart] |= PART_CORRUPTED;
					m_csGapListAndPartStatus.Unlock();
				}
			}
		}
	}

//	Calculate the size of completed parts
	for (uint32 i = 0; i < GetPartCount(); i++)
	{
		uint32 dwStart, dwEnd;

		if (IsPartComplete(i, &dwStart, &dwEnd))
		{
			m_dwCompletedPartsSize += dwEnd - dwStart + 1;
		}
	}

	EMULE_TRY

	LoadPartFileStats();

	EMULE_CATCH

//	Open permanent handles
	if (!m_hPartFileWrite.Open(m_strFilePath, CFile::modeWrite | CFile::shareDenyWrite | CFile::osSequentialScan))
	{
#ifndef NEW_SOCKETS_ENGINE
		AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_FILEOPEN), m_strFilePath, m_strFileName);
#endif //NEW_SOCKETS_ENGINE
		return false;
	}
	if (!m_hPartFileRead.Open(m_strFilePath, CFile::modeRead | CFile::shareDenyNone | CFile::osSequentialScan))
	{
		m_hPartFileWrite.Close();
#ifndef NEW_SOCKETS_ENGINE
		AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_FILEOPEN), m_strFilePath, m_strFileName);
#endif //NEW_SOCKETS_ENGINE
		return false;
	}

	uint64	qwFileSz = m_hPartFileWrite.GetLength();

	if (qwFileSz < GetFileSize())
		AddGap(static_cast<uint32>(qwFileSz), GetFileSize() - 1);

//	Goes both ways - Partfile should never be too large
	if (qwFileSz > GetFileSize())
	{
		TRACE("Partfile \"%s\" is too large! Truncating %I64u bytes.\n", GetFileName(), qwFileSz - GetFileSize());
		m_hPartFileWrite.SetLength(GetFileSize());
	}

	SetStatus(PS_EMPTY);

//	Check hashcount, filestatus etc.
	CSingleLock Lock(&m_csHashList, TRUE);
	INT_PTR iSize = m_partHashArray.GetCount();
	Lock.Unlock();
	if (iSize != GetED2KPartHashCount())
	{
		m_bHashSetNeeded = true;
		return true;
	}
	else
	{
		m_bHashSetNeeded = false;
		for (int i = 0; i < GetPartCount(); i++)
		{
			if (IsPartComplete(i))
			{
				SetStatus(PS_READY);
				break;
			}
		}
	}

	m_csGapListAndPartStatus.Lock();

	bool	bIsGapListEmpty = gaplist.IsEmpty();

	m_csGapListAndPartStatus.Unlock();

// Is this file complete already?
	if (bIsGapListEmpty)
	{
		CompleteFile(false);
		return true;
	}

//	Check date of .part file - if it's wrong, rehash file
	CFileStatus filestatus;

//	"...returns m_attribute without high-order flags" indicates a known MFC bug, wonder how many unknown there are...
	m_hPartFileWrite.GetStatus(filestatus);
	m_timeLastDownTransfer = filestatus.m_mtime;

	uint32	dwFileDate = filestatus.m_mtime.GetTime();

	AdjustNTFSDaylightFileTime(&dwFileDate, m_strFilePath);
	if (m_timetLastWriteDate != dwFileDate)
	{
#ifndef NEW_SOCKETS_ENGINE
		AddLogLine(false, RGB_LOG_WARNING + GetResString(IDS_ERR_REHASH), m_strFileName);
#endif //NEW_SOCKETS_ENGINE
	//	Rehash
		SetStatus(PS_WAITINGFORHASH);
#ifndef NEW_SOCKETS_ENGINE
		CHashFileThread* addfilethread = (CHashFileThread*) AfxBeginThread(RUNTIME_CLASS(CHashFileThread), THREAD_PRIORITY_NORMAL + g_App.m_pPrefs->GetMainProcessPriority(), 0, CREATE_SUSPENDED);
		addfilethread->SetValues(0, GetPath(), m_hPartFileWrite.GetFileName().GetBuffer(), this);
		addfilethread->ResumeThread();
#endif //NEW_SOCKETS_ENGINE
	}

	UpdateCompletedInfos();

//	Smoothed "remaining time" calculation
	m_SessionStartSize = m_dwCompletedSize;
	if (m_bPaused)
		SetStartTimeReset(true);

//	load own file comments and ratings
	UpdateFileRatingCommentAvail();

	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::TryToRecoverPartFile(LPCTSTR in_directory, LPCTSTR in_filename)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	CString strTempName;

	m_strPartMetFileName = in_filename;
	m_strTempDir = in_directory;
	if (m_strTempDir.Right(1) == _T('\\'))
		m_strTempDir.Truncate(m_strTempDir.GetLength() - 1);
	SetPath(m_strTempDir);

	m_strFullName.Format(_T("%s\\%s"), m_strTempDir, m_strPartMetFileName);
	SetFilePath(m_strFullName.Left(m_strFullName.GetLength() - 4));

//	Rename ".met" to ".met.bad"
	strTempName.Format(_T("%s.bad"), m_strFullName);
	if (_trename(m_strFullName, strTempName) != 0)
		AddDebugLogLine(RGB_LOG_ERROR_TXT _T("Failed to rename part.met file '%s' to '%s' - %s"), m_strFullName, strTempName, _tcserror(errno));

//	Copy ".met.bak" to ".met"
	strTempName.Format(_T("%s.bak"), m_strFullName);
	if (CopyFile(strTempName, m_strFullName, false) == 0)
	{
		AddDebugLogLine(RGB_LOG_ERROR_TXT _T("Failed to copy .bak file '%s' to '%s'"), strTempName, m_strFullName);
		AddLogLine(false, RGB_LOG_ERROR_TXT + GetResString(IDS_TRYTORECOVER_FAIL), m_strFullName);
		return false;
	}

	AddLogLine(false, GetResString(IDS_TRYTORECOVER), m_strFullName);

	bool bLoadPartFileStatus = LoadPartFile(in_directory, in_filename);

	if (bLoadPartFileStatus)
		AddLogLine(false, RGB_LOG_SUCCESS + GetResString(IDS_RECOVERED_PARTMET), GetFileName());
	else
		AddLogLine(false, RGB_LOG_ERROR_TXT + GetResString(IDS_TRYTORECOVER_FAIL), m_strFullName);

	return bLoadPartFileStatus;

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE

	return false;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::SavePartFile()
{
	EMULE_TRY

	switch (m_eStatus)
	{
		case PS_WAITINGFORHASH:
		case PS_HASHING:
		case PS_COMPLETING:
		case PS_COMPLETE:
		return false;
	}

//	Protect file corruption during calling from different threads
//	will be unlocked on exit
	CSingleLock sLock(&m_csSavePartFile, TRUE);

//	Get filedate
	CFileFind ff;
	bool bEnd = !ff.FindFile(m_strFilePath, 0);
	if (!bEnd)
		ff.FindNextFile();
	if (bEnd || ff.IsDirectory())
	{
#ifndef NEW_SOCKETS_ENGINE
		AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_SAVEMET), GetResString(IDS_ERR_PART_FNF), m_strPartMetFileName, m_strFileName);
#endif //NEW_SOCKETS_ENGINE
		return false;
	}

//	Create a backup of ".met"
	::CopyFile(m_strFullName, m_strFullName + _T(".bak"), FALSE);

	CTime lwtime;
	ff.GetLastWriteTime(lwtime);
	m_timetLastWriteDate = lwtime.GetTime();
	AdjustNTFSDaylightFileTime(&m_timetLastWriteDate, ff.GetFilePath());
	ff.Close();

//	Create a temporary file
	CString strTempName(m_strFullName);

	strTempName.Append(_T(".tmp"));
	if (_tremove(strTempName) && errno != ENOENT)
#ifndef NEW_SOCKETS_ENGINE
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), strTempName);
#else
		;
#endif //NEW_SOCKETS_ENGINE

	CStdioFile file;

	try
	{
	//	Open the file
		file.Open(strTempName, CFile::modeWrite|CFile::modeCreate|CFile::typeBinary);
	//	Version
		byte uVersion = PARTFILE_VERSION;
		file.Write(&uVersion, 1);
	//	Date
		file.Write(&m_timetLastWriteDate, 4);
	//	Hash
		file.Write(&m_fileHash, 16);
		CSingleLock Lock(&m_csHashList, TRUE);
		uint16 uParts = m_partHashArray.GetCount();
		file.Write(&uParts, 2);
		for (int x = 0; x != uParts; x++)
			file.Write(m_partHashArray[x], 16);
		Lock.Unlock();

	//	Tags
		CArray<Gap_Struct, Gap_Struct> aGaps;
		CWrTag	tagWr;
		uint32	dwTmp, dwTagFilePos, dwTagCount = 0;

		dwTagFilePos = static_cast<uint32>(file.GetPosition());
		file.Write(&dwTagCount, 4);

		tagWr.WriteToFile(FT_FILENAME, GetFileName(), file);
		dwTagCount++;
		tagWr.WriteToFile(FT_FILESIZE, GetFileSize(), file);
		dwTagCount++;
		if (m_dwBytesTransferred != 0)
		{
			tagWr.WriteToFile(FT_TRANSFERRED, m_dwBytesTransferred, file);
			dwTagCount++;
		}
	//	Compression gain can be negative in case of disconnection in the middle of
	//	compressed stream, but we're not going to save it for compatibility purposes
		if (static_cast<sint64>(m_qwGainDueToCompression) > 0)
		{
			tagWr.WriteToFile(FT_COMPRESSION,
				(m_qwGainDueToCompression > 0xFFFFFFFF) ? 0xFFFFFFFF : static_cast<uint32>(m_qwGainDueToCompression), file);
			dwTagCount++;
		}
		if (m_qwLostDueToCorruption != 0)
		{
			tagWr.WriteToFile(FT_CORRUPTED,
				(m_qwLostDueToCorruption > 0xFFFFFFFF) ? 0xFFFFFFFF : static_cast<uint32>(m_qwLostDueToCorruption), file);
			dwTagCount++;
		}
		if (m_bPaused)
		{
			tagWr.WriteToFile(FT_STATUS, 1, file);
			dwTagCount++;
		}
		tagWr.WriteToFile(FT_DLPRIORITY, (IsAutoPrioritized()) ? PR_AUTO : priority, file);
		dwTagCount++;
		tagWr.WriteToFile(FT_ULPRIORITY, (IsULAutoPrioritized()) ? PR_AUTO : GetULPriority(), file);
		dwTagCount++;
		if (static_cast<uint32>(lastseencomplete.GetTime()) != 0)
		{
			tagWr.WriteToFile(FT_LASTSEENCOMPLETE, lastseencomplete.GetTime(), file);
			dwTagCount++;
		}
		if (m_eCategoryID != CAT_NONE)
		{
			tagWr.WriteToFile(FT_CATEGORY, static_cast<uint32>(m_eCategoryID), file);
			dwTagCount++;
		}
		if (statistic.GetAllTimeTransferred() != 0)
		{
			tagWr.WriteToFile(FT_ATTRANSFERRED, static_cast<uint32>(statistic.GetAllTimeTransferred()), file);
			if ((dwTmp = static_cast<uint32>(statistic.GetAllTimeTransferred() >> 32)) != 0)
			{
				tagWr.WriteToFile(FT_ATTRANSFERREDHI, dwTmp, file);
				dwTagCount++;
			}
			dwTagCount++;
		}
		if (statistic.GetAllTimeRequests() != 0)
		{
			tagWr.WriteToFile(FT_ATREQUESTED, statistic.GetAllTimeRequests(), file);
			dwTagCount++;
		}
		if (statistic.GetAllTimeAccepts() != 0)
		{
			tagWr.WriteToFile(FT_ATACCEPTED, statistic.GetAllTimeAccepts(), file);
			dwTagCount++;
		}
	//	Store corrupted part numbers
		CString	strCorruptedParts;

		for (uint32 j = 0; j < GetPartCount(); j++)
		{
			if (IsCorruptedPart(j))
				strCorruptedParts.AppendFormat((strCorruptedParts.IsEmpty()) ? _T("%u") : _T(",%u"), j);
		}
		if (!strCorruptedParts.IsEmpty())
		{
			tagWr.WriteToFile(FT_CORRUPTEDPARTS, strCorruptedParts, file);
			dwTagCount++;
		}
		if (GetPermissions() != PERM_ALL)
		{
			tagWr.WriteToFile(FT_PERMISSIONS, GetPermissions(), file);
			dwTagCount++;
		}

		dwTagCount += m_tagArray.GetCount();
		for (j = 0; j != m_tagArray.GetCount(); j++)
			m_tagArray[j]->WriteToFile(file);

	//	Make a gap list copy to avoid synchronization object allocation for a long time
		GetGapListCopy(&aGaps);

		char	acNameBuffer[16];
		uint32	dwPos = 0;

		Gap_Struct	*pGap = aGaps.GetData();
		Gap_Struct	*pGapMax = pGap + aGaps.GetCount();

		dwTagCount += 2 * aGaps.GetCount();
		for (; pGap < pGapMax; pGap++)
		{
			itoa(dwPos++, &acNameBuffer[1], 10);
			acNameBuffer[0] = FT_GAPSTART;
			tagWr.WriteToFile(acNameBuffer, pGap->m_dwStartOffset, file);
		//	gap start = first missing byte but gap ends = first non-missing byte in edonkey
		//	but I think its easier to user the real limits
			acNameBuffer[0] = FT_GAPEND;
			tagWr.WriteToFile(acNameBuffer, pGap->m_dwEndOffset + 1, file);
		}

	//	Save valid tag count
		file.Seek(dwTagFilePos, CFile::begin);
		file.Write(&dwTagCount, 4);
		file.SeekToEnd();
	}
	catch(CFileException* error)
	{
		OUTPUT_DEBUG_TRACE	();
#ifndef NEW_SOCKETS_ENGINE
		AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_SAVEMET), GetErrorMessage(error), m_strPartMetFileName, m_strFileName);
#endif //NEW_SOCKETS_ENGINE
		error->Delete();
		file.Close();
 	//	Remove the partially written or otherwise damaged temporary file
 		_tremove(strTempName);
		return false;
	}

	_commit(_fileno(file.m_pStream));
	file.Close();

//	Delete original ".met", after successfully writing the temporary part.met file...
	if (_tremove(m_strFullName) != 0 && errno != ENOENT)
		AddDebugLogLine(RGB_LOG_ERROR_TXT _T("Failed to remove '%s' - %s"), m_strFullName, _tcserror(errno));

//	Rename ".met.tmp" to ".met"
	if (_trename(strTempName, m_strFullName) != 0)
		AddDebugLogLine(RGB_LOG_ERROR_TXT _T("Failed to move temporary part.met file '%s' to '%s' - %s"), strTempName, m_strFullName, _tcserror(errno));

	CStdioFile	fDir;
	CString		strPath(m_strFilePath);

	strPath += _T(".dir");
	try
	{
	//	ChangeDir
		if (!m_strAlternativePath.IsEmpty())
		{
			if (fDir.Open(strPath, CFile::modeCreate | CFile::modeWrite | CFile::shareDenyWrite | CFile::typeText))
			{
				fDir.WriteString((LPCTSTR)m_strAlternativePath);
				fDir.Close();
			}
		}
	}
	catch(CFileException* error)
	{
		OUTPUT_DEBUG_TRACE();
		AddLogLine(false, RGB_LOG_ERROR_TXT _T("Error saving %s - %s"), strPath, GetErrorMessage(error));
		error->Delete();
		fDir.Close();
	}

	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::PartFileHashFinished(CKnownFile* pKnownFile)
{
	EMULE_TRY

	bool	bErrorFound = false;

	CSingleLock Lock(&m_csHashList, TRUE);

	uint32 dwCount = static_cast<uint32>(m_partHashArray.GetSize());

//	Number of parts to check taking into consideration files with size % PARTSIZE = 0 and size < PARTSIZE
	dwCount = (dwCount > GetPartCount()) ? GetPartCount() : dwCount;

	for (uint32 i = 0; i < dwCount; i++)
	{
		uint32 dwPartStart, dwPartEnd;

		if (IsPartComplete(i, &dwPartStart, &dwPartEnd))
		{
			if ((pKnownFile->GetPartHash(i) == NULL) || (md4cmp(pKnownFile->GetPartHash(i), GetPartHash(i)) != 0))
			{
#ifndef NEW_SOCKETS_ENGINE
				AddLogLine(false, RGB_LOG_WARNING + GetResString(IDS_ERR_FOUNDCORRUPTION), i, GetFileName());
#endif //NEW_SOCKETS_ENGINE
				AddGap(dwPartStart, dwPartEnd);
				m_csGapListAndPartStatus.Lock();
				m_PartsStatusVector[i] |= PART_CORRUPTED;
				m_csGapListAndPartStatus.Unlock();
			//	Increment corruption statistics
				AddRxCorruptedAmount(dwPartEnd - dwPartStart + 1);
				bErrorFound = true;
			}
		}
	}
	Lock.Unlock();
	delete pKnownFile;

	if (!bErrorFound)
	{
		if (m_eStatus == PS_COMPLETING)
		{
			CompleteFile(true);
			return;
		}
#ifndef NEW_SOCKETS_ENGINE
		else
			AddLogLine(false, IDS_HASHINGDONE, GetFileName());
#endif //NEW_SOCKETS_ENGINE
	}
	else
	{
		SetStatus(PS_READY);
		SavePartFile();
	//	If a file is "ready" it can't be paused
		m_bPaused = false;
		return;
	}
	SetStatus(PS_READY);
	SavePartFile();
#ifndef NEW_SOCKETS_ENGINE
	g_App.m_pSharedFilesList->SafeAddKnownFile(this);
#endif //NEW_SOCKETS_ENGINE

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::AddGap(uint32 dwStart, uint32 dwEnd)
{
	EMULE_TRY

	CSingleLock lockGap(&m_csGapListAndPartStatus, TRUE);

	POSITION pos1, pos2;

	for (pos1 = gaplist.GetHeadPosition(); (pos2 = pos1) != NULL;)
	{
		Gap_Struct* cur_gap = gaplist.GetNext(pos1);
		if (cur_gap->m_dwStartOffset >= dwStart && cur_gap->m_dwEndOffset <= dwEnd)
		{ // this gap is inside the new gap - delete
			gaplist.RemoveAt(pos2);
			m_dwGapsSum -= cur_gap->m_dwEndOffset - cur_gap->m_dwStartOffset + 1;
			FillGapInParts(cur_gap->m_dwStartOffset, cur_gap->m_dwEndOffset);
			delete cur_gap;
		}
		else if (cur_gap->m_dwStartOffset >= dwStart && cur_gap->m_dwStartOffset <= dwEnd)
		{ // a part of this gap is in the new gap - extend limit and delete
			dwEnd = cur_gap->m_dwEndOffset;
			gaplist.RemoveAt(pos2);
			m_dwGapsSum -= dwEnd - cur_gap->m_dwStartOffset + 1;
			FillGapInParts(cur_gap->m_dwStartOffset, dwEnd);
			delete cur_gap;
		}
		else if (cur_gap->m_dwEndOffset <= dwEnd && cur_gap->m_dwEndOffset >= dwStart)
		{ // a part of this gap is in the new gap - extend limit and delete
			dwStart = cur_gap->m_dwStartOffset;
			gaplist.RemoveAt(pos2);
			m_dwGapsSum -= cur_gap->m_dwEndOffset - dwStart + 1;
			FillGapInParts(dwStart, cur_gap->m_dwEndOffset);
			delete cur_gap;
		}
		else if (dwStart >= cur_gap->m_dwStartOffset && dwEnd <= cur_gap->m_dwEndOffset)
		{ // new gap is already inside this gap - return
			return;
		}
	}
	Gap_Struct* new_gap = new Gap_Struct;
	new_gap->m_dwStartOffset = dwStart;
	new_gap->m_dwEndOffset = dwEnd;
	m_dwGapsSum += dwEnd - dwStart + 1;
	AddGapToParts(dwStart, dwEnd);
	gaplist.AddTail(new_gap);

	lockGap.Unlock();

	UpdateDisplayedInfo();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	IsComplete() returns false if there are gaps in the range 'start' to 'dwEnd' (inclusive).
bool CPartFile::IsComplete(uint32 start, uint32 dwEnd)
{
	EMULE_TRY

	if (dwEnd >= GetFileSize())
		dwEnd = GetFileSize() - 1;

	bool	bRc = true;

// 	Will be unlocked on exit
	CSingleLock lockGap(&m_csGapListAndPartStatus, TRUE);

	for (POSITION pos = gaplist.GetHeadPosition(); pos != NULL;)
	{
		Gap_Struct * cur_gap = gaplist.GetNext(pos);

		if ( (cur_gap->m_dwStartOffset >= start && cur_gap->m_dwEndOffset <= dwEnd)			// gap is inside range (this check is unnecessary)
		     || (cur_gap->m_dwStartOffset >= start && cur_gap->m_dwStartOffset <= dwEnd)	// gap starts inside range
		     || (cur_gap->m_dwEndOffset <= dwEnd && cur_gap->m_dwEndOffset >= start)		// gap ends inside range
		     || (start >= cur_gap->m_dwStartOffset && dwEnd <= cur_gap->m_dwEndOffset) )	// range is inside gap
		{
			bRc = false;
			break;
		}
	}
	return bRc;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	IsPartComplete() returns true if a part is complete and verified.
bool CPartFile::IsPartComplete(uint32 dwPart) const
{
	return (m_PartsStatusVector[dwPart] == PART_VERIFIED);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	IsPartComplete() returns true if a part is complete and verified.
//		Params:
//			dwPart   - part number;
//			pdwStart - pointer to get part start offset;
//			pdwEnd   - pointer to get part end offset (inclusive).
bool CPartFile::IsPartComplete(uint32 dwPart, uint32 *pdwStart, uint32 *pdwEnd) const
{
	uint32 dwStart = dwPart * PARTSIZE;
	uint32 dwEnd = ((GetFileSize() - dwStart) > PARTSIZE) ? (dwStart + PARTSIZE - 1) : (GetFileSize() - 1);

	*pdwStart = dwStart;
	*pdwEnd = dwEnd;

	return (m_PartsStatusVector[dwPart] == PART_VERIFIED);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	IsPartFull() returns true if all data of the was received.
bool CPartFile::IsPartFull(uint32 dwPart) const
{
	return ((m_PartsStatusVector[dwPart] & 0xFFFFFF) == 0);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::IsPartDownloading(uint32 dwPart) const
{
	EMULE_TRY

	for (POSITION pos = requestedblocks_list.GetHeadPosition(); pos != NULL;)
	{
		Requested_Block_Struct *pReqBlock = requestedblocks_list.GetNext(pos);

		if (dwPart == (pReqBlock->m_dwStartOffset / PARTSIZE))
			return true;
	}

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::IsAlreadyRequested(uint32 start, uint32 dwEnd)
{
	EMULE_TRY

	for (POSITION pos = requestedblocks_list.GetHeadPosition(); pos != NULL;)
	{
		Requested_Block_Struct * cur_block = requestedblocks_list.GetNext(pos);
		if ((start <= cur_block->m_dwEndOffset) && (dwEnd >= cur_block->m_dwStartOffset))
			return true;
	}

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	GetGapsInPart() return request blocks for the corresponding part
//		Params:
//			numPart    - a part to examine;
//			newblocks  - output array for allocated request blocks (can be NULL to check for block allocation);
//			dwCount    - number of blocks to allocate.
//		Return:
//			number of allocated blocks or number of possible blocks to allocate if newblocks = NULL.
uint32 CPartFile::GetGapsInPart(uint16 uPartNum, Requested_Block_Struct** ppNewBlocks, uint32 dwCount)
{
	uint32	dwNewBlockCount = 0;

	EMULE_TRY

//	Calculate offsets of the Part
	const uint32 uPartStart = uPartNum * PARTSIZE;
	const uint32 uPartEnd = ((GetFileSize() - uPartStart) > PARTSIZE) ? (uPartStart + PARTSIZE - 1) : (GetFileSize() - 1);

	CSingleLock lockGap(&m_csGapListAndPartStatus, TRUE);	// will be unlocked on exit

//	Retrieve Gap(s) matching the part
	for (POSITION pos = gaplist.GetHeadPosition(); pos != NULL;)
	{
		Gap_Struct *CurGap = gaplist.GetNext(pos);

	//	Check if Gap is inside the limit
		if (CurGap->m_dwStartOffset > uPartEnd || CurGap->m_dwEndOffset < uPartStart)
			continue;

	//	Calculate offsets of the Gap within the part
		uint32 uGapStart = (CurGap->m_dwStartOffset < uPartStart) ? uPartStart : CurGap->m_dwStartOffset;
		const uint32 uGapEnd = (CurGap->m_dwEndOffset > uPartEnd) ? uPartEnd : CurGap->m_dwEndOffset;

	//	Create n request block(s)
		while ((uGapEnd >= uGapStart) && (dwNewBlockCount < dwCount))
		{
		//	Calculate offsets of the block
			const uint32 startOffset = uGapStart;
			const uint32 endOffset = ((uGapStart + EMBLOCKSIZE - 1) >= uGapEnd) ? uGapEnd : ((uGapStart + EMBLOCKSIZE - 1) - (uGapStart % EMBLOCKSIZE));

		//	Prepare offset of next block
			uGapStart = endOffset + 1;

		//	Check if block has already been requested
			if (!IsAlreadyRequested(startOffset, endOffset))
			{
				if (ppNewBlocks != NULL)
				{
				//	Create 1 request block
					Requested_Block_Struct * block = new Requested_Block_Struct;
					block->m_dwStartOffset = startOffset;
					block->m_dwEndOffset = endOffset;
					md4cpy(block->m_fileHash, GetFileHash());

				//	Flag the block as 'requested'
					requestedblocks_list.AddTail(block);

				//	Return the block to the source
					ppNewBlocks[dwNewBlockCount] = block;
				}
				dwNewBlockCount++;
			}
		}
	}

	EMULE_CATCH

	return dwNewBlockCount;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::FillGap(uint32 dwStart, uint32 dwEnd)
{
	EMULE_TRY

	POSITION pos1, pos2;

	CSingleLock lockGap(&m_csGapListAndPartStatus, TRUE);

	for (pos1 = gaplist.GetHeadPosition(); (pos2 = pos1) != NULL; )
	{
		Gap_Struct* cur_gap = gaplist.GetNext(pos1);
		if (cur_gap->m_dwStartOffset >= dwStart && cur_gap->m_dwEndOffset <= dwEnd)
		{ // our part fills this gap completely
			gaplist.RemoveAt(pos2);
			m_dwGapsSum -= cur_gap->m_dwEndOffset - cur_gap->m_dwStartOffset + 1;
			FillGapInParts(cur_gap->m_dwStartOffset, cur_gap->m_dwEndOffset);
			delete cur_gap;
		}
		else if (cur_gap->m_dwStartOffset >= dwStart && cur_gap->m_dwStartOffset <= dwEnd)
		{ // a part of this gap is in the part - set limit
			m_dwGapsSum -= dwEnd - cur_gap->m_dwStartOffset + 1;
			FillGapInParts(cur_gap->m_dwStartOffset, dwEnd);
			cur_gap->m_dwStartOffset = dwEnd + 1;
		}
		else if (cur_gap->m_dwEndOffset <= dwEnd && cur_gap->m_dwEndOffset >= dwStart)
		{ // a part of this gap is in the part - set limit
			m_dwGapsSum -= cur_gap->m_dwEndOffset - dwStart + 1;
			FillGapInParts(dwStart, cur_gap->m_dwEndOffset);
			cur_gap->m_dwEndOffset = dwStart - 1;
		}
		else if (dwStart >= cur_gap->m_dwStartOffset && dwEnd <= cur_gap->m_dwEndOffset)
		{
			uint32 dwTmp = cur_gap->m_dwEndOffset;

			cur_gap->m_dwEndOffset = dwStart - 1;
			cur_gap = new Gap_Struct;
			cur_gap->m_dwStartOffset = dwEnd + 1;
			cur_gap->m_dwEndOffset = dwTmp;
			m_dwGapsSum -= dwEnd - dwStart + 1;
			FillGapInParts(dwStart, dwEnd);
			gaplist.InsertAfter(pos1, cur_gap);
			break;
		}
	}
	lockGap.Unlock();

	UpdateCompletedInfos();
	UpdateDisplayedInfo();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// This function updates the part status vector based on gap information
// Must be called only under m_csGapListAndPartStatus protection
void CPartFile::FillGapInParts(uint32 dwGapStart, uint32 dwGapEnd)
{
	for (uint32 dwPart = (dwGapStart / PARTSIZE); dwPart <= (dwGapEnd / PARTSIZE); dwPart++)
	{
		uint32 dwPartStart = dwPart * PARTSIZE;
		uint32 dwPartEnd = ((GetFileSize() - dwPartStart) > PARTSIZE) ? (dwPartStart + PARTSIZE - 1) : (GetFileSize() - 1);
		uint32 dwPartFilledSize = GetPartLeftToDLSize(dwPart);

		if (dwGapStart <= dwPartStart)
		{
			if (dwGapEnd >= dwPartEnd)	//	[Gs (Ps Pe) Ge]
				dwPartFilledSize = 0;
			else						//	[Gs (Ps Ge] Pe)
				dwPartFilledSize -= dwGapEnd - dwPartStart + 1;
		}
		else
		{
			if (dwGapEnd < dwPartEnd)	//	(Ps [Gs Ge] Pe) else (Ps [Gs Pe) Ge]
				dwPartEnd = dwGapEnd;
			dwPartFilledSize -= dwPartEnd - dwGapStart + 1;
		}

		m_PartsStatusVector[dwPart] = (m_PartsStatusVector[dwPart] & 0xFF000000) | dwPartFilledSize;
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// This function updates the part status vector based on gap information
// Must be called only under m_csGapListAndPartStatus protection
void CPartFile::AddGapToParts(uint32 dwGapStart, uint32 dwGapEnd)
{
	for (uint32 dwPart = (dwGapStart / PARTSIZE); dwPart <= (dwGapEnd / PARTSIZE); dwPart++)
	{
		uint32 dwPartStart = dwPart * PARTSIZE;
		uint32 dwPartEnd = ((GetFileSize() - dwPartStart) > PARTSIZE) ? (dwPartStart + PARTSIZE - 1) : (GetFileSize() - 1);
		uint32 dwPartFilledSize = GetPartLeftToDLSize(dwPart);

		if (dwGapStart <= dwPartStart)
		{
			if (dwGapEnd >= dwPartEnd)	//	[Gs (Ps Pe) Ge]
				dwPartFilledSize = dwPartEnd - dwPartStart + 1;
			else						//	[Gs (Ps Ge] Pe)
				dwPartFilledSize += dwGapEnd - dwPartStart + 1;
		}
		else
		{
			if (dwGapEnd < dwPartEnd)	//	(Ps [Gs Ge] Pe) else (Ps [Gs Pe) Ge]
				dwPartEnd = dwGapEnd;
			dwPartFilledSize += dwPartEnd - dwGapStart + 1;
		}
	//	Reset PART_VERIFIED status as a gap was added to this part
		m_PartsStatusVector[dwPart] =
			(m_PartsStatusVector[dwPart] & (0xFF000000 & ~PART_VERIFIED)) | dwPartFilledSize;
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::UpdateCompletedInfos()
{
	EMULE_TRY

	if ((m_dwGapsSum != 0) || (requestedblocks_list.GetCount() != 0))
	{
		m_dblPercentCompleted = (1.0 - static_cast<double>(m_dwGapsSum) / GetFileSize()) * 100;
		m_dwCompletedSize = GetFileSize() - m_dwGapsSum;
	}
	else
	{
		m_dblPercentCompleted = 100.0;
		m_dwCompletedSize = GetFileSize();
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::DrawStatusBar(CDC* dc, RECT* rect, bool bFlat)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	const COLORREF crProgress = (bFlat) ? RGB(0, 150, 0) : RGB(0, 192, 0);
	COLORREF crHave, crPending, crMissing;

//	grayPause - Colors by status
	EnumPartFileStatuses	eFileStatus = GetStatus();
	bool			bNotPaused = (eFileStatus == PS_EMPTY) || (eFileStatus == PS_READY);

	if (bNotPaused)
	{
		crHave = (bFlat) ? RGB(0, 0, 0) : RGB(95, 95, 95);
		crPending = RGB(255, 208, 0);
		crMissing = RGB(255, 0, 0);
	}
	else
	{
		crHave = (bFlat) ? RGB(105, 105, 105) : RGB(142, 142, 142);
		crPending = RGB(255, 240, 142);
		crMissing = RGB(255, 142, 142);
	}

	CBarShader s_ChunkBar(rect->bottom - rect->top, rect->right - rect->left, crHave, GetFileSize());

	if (m_eStatus == PS_COMPLETE || m_eStatus == PS_COMPLETING)
	{
		s_ChunkBar.Fill(crProgress);
		s_ChunkBar.Draw(dc, rect->left, rect->top, bFlat);
		m_dblPercentCompleted = 100.0;
		m_dwCompletedSize = GetFileSize();
		return;
	}

//	Red gaps
	CArray<Gap_Struct, Gap_Struct> aGaps;

//	Make a gap list copy to avoid synchronization object allocation for a long time
	GetGapListCopy(&aGaps);

	Gap_Struct	*pGap = aGaps.GetData();
	Gap_Struct	*pGapMax = pGap + aGaps.GetCount();

	for (; pGap < pGapMax; pGap++)
	{
		uint32 dwGapBeg = pGap->m_dwStartOffset;
		uint32 dwGapEnd = pGap->m_dwEndOffset + 1;

		for (uint32 i = (dwGapBeg / PARTSIZE); i < GetPartCount(); i++)
		{
			const uint32 dwPartStart = i * PARTSIZE;
			const uint32 dwPartEnd = ((GetFileSize() - dwPartStart) > PARTSIZE) ? (dwPartStart + PARTSIZE) : GetFileSize();

			if (dwGapEnd > dwPartEnd)
				dwGapEnd = dwPartEnd;	// The rest is in the next part

			COLORREF	color;
			int			iFreq;

			if (m_srcPartFrequencies.GetCount() >= (INT_PTR)i && ((iFreq = m_srcPartFrequencies[i] - 1) >= 0))
			{
				if (bNotPaused)
				{
					if (IsPartDownloading(i))
						color = crPending;	// Downloading part in yellow
					else
						color = RGB(0, ((210 - 8 * iFreq) < 0) ? 0 : (210 - 8 * iFreq), 255);
				}
				else
					color = RGB(100, ((210 - 8 * iFreq) < 0) ? 0 : (255 - 8 * iFreq), 255);
			}
			else
				color = crMissing;

			s_ChunkBar.FillRange(dwGapBeg, dwGapEnd, color);

			dwGapBeg = dwGapEnd;
			dwGapEnd = pGap->m_dwEndOffset + 1;
			if (dwGapBeg == dwGapEnd)	// Finished?
				break;
		}
	}

	s_ChunkBar.Draw(dc, rect->left, rect->top, bFlat);

//	Green progress
	RECT gaprect;
	gaprect.top = rect->top;
	gaprect.bottom = gaprect.top + PROGRESS_HEIGHT;
	gaprect.left = rect->left;

	uint32	w = rect->right - rect->left + 1;
	uint32	wp = (uint32)(m_dblPercentCompleted / 100.0 * w + 0.5);

	if (!bFlat)
	{
		CBarShader	s_LoadBar(PROGRESS_HEIGHT, wp, crPending, m_dwCompletedSize);

		s_LoadBar.FillRange(0, m_dwCompletedPartsSize, crProgress);
		s_LoadBar.Draw(dc, gaprect.left, gaprect.top, false);
	}
	else
	{
		uint32		wc = (uint32)(static_cast<double>(m_dwCompletedPartsSize) / GetFileSize() * w + 0.5);
		int			left = rect->left;
		CBrush	   *pOldBrush;

	//	Green
		gaprect.right = left + wc;
		CBrush pr_Brush(crProgress);
		pOldBrush = dc->SelectObject(&pr_Brush);
		dc->FillRect(&gaprect, &pr_Brush);

	//	Yellow
		gaprect.left = left + wc;
		gaprect.right = left + wp;
		CBrush pe_Brush(crPending);
		dc->SelectObject(&pe_Brush);
		dc->FillRect(&gaprect, &pe_Brush);

	//	Draw gray progress only if flat
		gaprect.left = left + wp;
		gaprect.right = left + w;
		CBrush rgb_Brush(RGB(224, 224, 224));
		dc->SelectObject(&rgb_Brush);
		dc->FillRect(&gaprect, &rgb_Brush);

		dc->SelectObject(pOldBrush);
	}

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::WritePartStatus(CFile* file)
{
	EMULE_TRY

	uint16 parts = GetED2KPartCount();
	file->Write(&parts, 2);
	uint16 done = 0;
	while (done != parts)
	{
		byte towrite = 0;
		for (uint32 i = 0; i < 8; i++)
		{
		//	Report that we have "void" part for files with size % PARTSIZE = 0
			if ( ((done == (parts - 1)) && (parts != GetPartCount())) ||
				(IsPartShared(done) && IsPartComplete(done)) )
				towrite |= (1 << i);

			if (++done == parts)
				break;
		}
		file->Write(&towrite, 1);
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::WriteCompleteSourcesCount(CFile* file)
{
	uint16 completecount = GetCompleteSourcesCount();
	file->Write(&completecount, 2);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CPartFile::GetValidSourcesCount()
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	uint32	dwListsMask = SLM_VALID_SOURCES;
	uint32	dwCounter = 0;

	for (EnumDLQState eDS = DS_DOWNLOADING; eDS < DS_LAST_QUEUED_STATE; ++eDS, dwListsMask >>= 1)
	{
		if ((dwListsMask & 1) != 0)
			dwCounter += m_SourceLists[eDS].size();
	}

	return dwCounter;

	EMULE_CATCH

#endif //NEW_SOCKETS_ENGINE
	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CPartFile::GetNotCurrentSourcesCount()
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	uint32	dwCounter = 0;

	for (int iDS = 0; iDS < DS_LAST_QUEUED_STATE; iDS++)
	{
		if ((iDS != DS_ONQUEUE) && (iDS != DS_DOWNLOADING) && (iDS != DS_LOWID_ON_OTHER_SERVER))
			dwCounter += m_SourceLists[iDS].size();
	}

	return dwCounter;

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
EnumPartFileStatuses CPartFile::GetStatus(bool bIgnorePause)
{
	if (!m_bPaused || m_eStatus == PS_ERROR || m_eStatus == PS_COMPLETING || m_eStatus == PS_COMPLETE || bIgnorePause)
		return m_eStatus;
	else if (m_bStopped)
		return PS_STOPPED;
	else
		return PS_PAUSED;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Inserted by Vorlost for load rarest parts first, code by Denominator
uint16 CPartFile::GetRemainingBlocksInPart(uint16 partnumber)
{
	EMULE_TRY

	uint16		uNumRemainingBlocks = 0;
	for (uint32 i = 0; i < ((PARTSIZE + EMBLOCKSIZE - 1) / EMBLOCKSIZE); i++)
	{
		uint32 start = (PARTSIZE * partnumber) + i * EMBLOCKSIZE;
		uint32 dwEnd = (PARTSIZE * partnumber) + ((i + 1) * EMBLOCKSIZE) - 1;
		if (dwEnd >= PARTSIZE * ((uint32)partnumber + 1))
			dwEnd = (PARTSIZE * (partnumber + 1)) - 1;
		if (start >= GetFileSize())
			break;
		if (dwEnd >= GetFileSize())
			dwEnd = GetFileSize() - 1;
		if (!IsComplete(start, dwEnd) && !IsAlreadyRequested(start, dwEnd))
			uNumRemainingBlocks++;
	}
	return uNumRemainingBlocks;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Remove no needed sources from download queue
//	[In] dwNumberSources2Remove: Number of sources to be removed from list
void CPartFile::RemoveNoNeededSources(uint32 dwNumberSources2Remove)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	ClientList	sourceListCopy;
	uint32		dwIdx = 0, dwRemoved;

//	First remove clients with "NNS" status
	GetCopySourceList(DS_NONEEDEDPARTS, &sourceListCopy);
	dwRemoved = (sourceListCopy.size() > dwNumberSources2Remove) ? dwNumberSources2Remove : sourceListCopy.size();

	for (ClientList::const_iterator cIt = sourceListCopy.begin(); dwIdx < dwRemoved; dwIdx++)
		g_App.m_pDownloadQueue->RemoveSource(*cIt++);

	if (dwRemoved < dwNumberSources2Remove)
	{
	//	Remove clients with "Queue Full" status
		GetCopySourceList(DS_ONQUEUE, &sourceListCopy);
		for (ClientList::const_iterator cIt = sourceListCopy.begin(); cIt != sourceListCopy.end(); cIt++)
		{
			CUpDownClient	*pSource = *cIt;

		//	The sources will be dropped in 2 cases:
		//	1) the remote client does not have required parts
		//	2) the remote queue is still full & we waited at least 3 reasks to enter into the queue
			if (pSource->IsRemoteQueueFull() && (pSource->GetAskedCountDown() > 3))
			{
				g_App.m_pDownloadQueue->RemoveSource(pSource);
				if (++dwRemoved >= dwNumberSources2Remove)
					break;
			}
		}
	}

	EMULE_CATCH

	m_dwLastPurgeTime = ::GetTickCount();
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CPartFile::Process(uint32 dwReduceDownload /*in percent*/, uint32 dwIteration)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	ClientList		sourceListCopy;
	uint16			nOldTransSourceCount = GetTransferringSrcCount();
	uint16			iOldSrcOnQueueCount = GetOnQueueSrcCount();
	DWORD			dwCurTick = ::GetTickCount();
	CUpDownClient  *pSource = NULL;
	POSITION		pos1;
	uint32			dwCurrentDataRate = 0;
	uint32			dwLastNoNeededCheckTime = 0;
	uint32			dwDataRate = 0;

//	flush the data to disk if:
//	1. it was requested
//	2. buffer size exceeds limit
//	3. data was not written within time limit
	if (m_bDataFlushReq
		|| (m_nTotalBufferData > (static_cast<uint32>(g_App.m_pPrefs->GetFileBufferSize()) * 30000))
		|| (dwCurTick > (m_nLastBufferFlushTime + BUFFER_TIME_LIMIT)) )
	{
	//	Avoid flushing while copying preview file
		if (!m_bPreviewing)
		{
			FlushBuffer();
		}
	}

//	Always process the transferring sources
//	Note: in case of disconnection sources will be put back to queue after download timeout
	if (!m_SourceLists[DS_DOWNLOADING].empty())
	{
		GetCopySourceList(DS_DOWNLOADING, &sourceListCopy);
		for (ClientList::const_iterator cIt = sourceListCopy.begin(); cIt != sourceListCopy.end(); cIt++)
		{
			pSource = *cIt;

		// Check if new client has non DL state, then it was removed after creation of the list copy, so ignore it
			if (pSource->GetDownloadState() != DS_DOWNLOADING)
				continue;

			dwCurrentDataRate = pSource->DownloadProcess(dwReduceDownload);

		//	Since client can be dropped due timeout we need to check the state again
			if (pSource->GetDownloadState() != DS_DOWNLOADING)
				continue;
			dwDataRate += dwCurrentDataRate;
		}
	}
	m_dwDataRate = dwDataRate;
	m_iNumTransferringSrcs = m_SourceLists[DS_DOWNLOADING].size();

//	Count connecting souces
	m_iSrcConnecting = m_SourceLists[DS_CONNECTING].size();
	m_iSrcConnViaServer = m_SourceLists[DS_WAITCALLBACK].size();

	if ((g_App.m_pServerConnect == NULL) || !g_App.m_pServerConnect->IsConnected())
		return m_dwDataRate;

//	Process every 10th time the rest of the sources
	if (dwIteration == 1)
	{
		if (!m_SourceLists[DS_CONNECTED].empty())
		{
			GetCopySourceList(DS_CONNECTED, &sourceListCopy);
			for (ClientList::const_iterator cIt = sourceListCopy.begin(); cIt != sourceListCopy.end(); cIt++)
			{
				pSource = *cIt;

#ifdef OLD_SOCKETS_ENABLED
			//	If the request socket doesn't exist or isn't connected
				if ((pSource->m_pRequestSocket == NULL) || !pSource->m_pRequestSocket->IsConnected())
				{
					pSource->SetDownloadState(DS_ERROR);
					pSource->Disconnected();
				}
				else if (dwCurTick - pSource->GetEnteredConnectedState() > CONNECTION_TIMEOUT + 20000)
				{
					if (pSource->HasClientNeededParts())
						pSource->SetDownloadState(DS_ONQUEUE);
					else
						pSource->SetDownloadState(DS_NONEEDEDPARTS);
				}
#endif //OLD_SOCKETS_ENABLED
			}
		}
		m_iSrcConnected = m_SourceLists[DS_CONNECTED].size();
	}
	else if (dwIteration == 2)	// process NNS sources
	{
		if (!m_SourceLists[DS_NONEEDEDPARTS].empty()
			&& !g_App.m_pListenSocket->TooManySockets())
		{
			uint32	dwLocalServerIP = g_App.m_pServerConnect->GetCurrentServer()->GetIP();
			bool	bIsLowId = g_App.m_pServerConnect->IsLowID();

			GetCopySourceList(DS_NONEEDEDPARTS, &sourceListCopy);
			for (ClientList::const_iterator cIt = sourceListCopy.begin(); cIt != sourceListCopy.end(); cIt++)
			{
			//	Process waiting sources unless it is allowed to open a new connection
				if (g_App.m_pListenSocket->TooManySockets())
					break;

				pSource = *cIt;

				if (pSource->HasLowID())
				{
					if (bIsLowId)
						pSource->SetDownloadState(DS_LOWTOLOWIP);
					else if (pSource->GetServerIP() != dwLocalServerIP)
						pSource->SetDownloadState(DS_LOWID_ON_OTHER_SERVER);
					continue;
				}

				if (dwCurTick > pSource->GetNextFileReaskTime())
					pSource->SendReask();
			}
		}
		m_iSrcNNP = m_SourceLists[DS_NONEEDEDPARTS].size();
	}
	else if (dwIteration == 3)	// process the sources on the queue
	{
		if (!m_SourceLists[DS_ONQUEUE].empty())
		{
			if (!g_App.m_pListenSocket->TooManySockets())
			{
				uint32	dwLocalServerIP = g_App.m_pServerConnect->GetCurrentServer()->GetIP();
				bool	bIsLowId = g_App.m_pServerConnect->IsLowID();
				uint16	uSrcQueueFull = 0;
				uint16	uSrcHighQR = 0;

				GetCopySourceList(DS_ONQUEUE, &sourceListCopy);
				for (ClientList::const_iterator cIt = sourceListCopy.begin(); cIt != sourceListCopy.end(); cIt++)
				{
				//	Process waiting sources unless it is allowed to open a new connection
					if (g_App.m_pListenSocket->TooManySockets())
					{
						uSrcQueueFull = 0xFFFF;
						break;
					}

					pSource = *cIt;

					if (pSource->HasLowID())
					{
						if (bIsLowId)
							pSource->SetDownloadState(DS_LOWTOLOWIP);
						else if (pSource->GetServerIP() != dwLocalServerIP)
							pSource->SetDownloadState(DS_LOWID_ON_OTHER_SERVER);
						continue;
					}

					if (pSource->IsRemoteQueueFull())
						uSrcQueueFull++;
					else if (pSource->GetRemoteQueueRank() > QUEUERANK_HIGH)
						uSrcHighQR++;

					if (dwCurTick > pSource->GetNextFileReaskTime())
						pSource->SendReask();
				}
				if (uSrcQueueFull != 0xFFFF)
				{
					m_iSrcQueueFull = uSrcQueueFull;
					m_uSrcHighQR = uSrcHighQR;
				}
			}
			m_iSrcOnQueue = m_SourceLists[DS_ONQUEUE].size();
		}
		else
		{
			m_iSrcOnQueue = m_iSrcQueueFull = m_uSrcHighQR = 0;
		}
	}
	else if (dwIteration == 4)	// process LowID clients on other servers
	{
		if (!m_SourceLists[DS_LOWID_ON_OTHER_SERVER].empty())
		{
			uint32	dwLocalServerIP = g_App.m_pServerConnect->GetCurrentServer()->GetIP();
			bool	bIsLowId = g_App.m_pServerConnect->IsLowID();

			GetCopySourceList(DS_LOWID_ON_OTHER_SERVER, &sourceListCopy);
			for (ClientList::const_iterator cIt = sourceListCopy.begin(); cIt != sourceListCopy.end(); cIt++)
			{
				pSource = *cIt;

			//	Since we can't connect to the client we can keep it in the list only if it keeps to request the file(s) from us
				if ((dwCurTick - pSource->GetLastUpRequest()) > MAX_PURGEQUEUETIME)
				{
					g_App.m_pDownloadQueue->RemoveSource(pSource);
					continue;
				}

				if (bIsLowId)
				{
					pSource->SetDownloadState(DS_LOWTOLOWIP);
					continue;
				}

			//	Check if a local or remote client changes a server
				if (pSource->GetServerIP() == dwLocalServerIP)
				{
					if (pSource->HasClientNeededParts())
						pSource->SetDownloadState(DS_ONQUEUE);
					else
						pSource->SetDownloadState(DS_NONEEDEDPARTS);
#if 0
					AddDebugLogLine( RGB_BLUE_VIOLET_TXT _T("LowID client %s is on again the same server (%s)"),
									pSource->GetClientNameWithSoftware(), this->GetFileName() );
#endif
				}
			}
		}
		m_uSrcLowIDOnOtherServer = m_SourceLists[DS_LOWID_ON_OTHER_SERVER].size();
	}
	else if (dwIteration == 5)	// process sources which haven't requested file yet
	{
		if (!m_SourceLists[DS_WAIT_FOR_FILE_REQUEST].empty()
			&& !g_App.m_pListenSocket->TooManySockets())
		{
			GetCopySourceList(DS_WAIT_FOR_FILE_REQUEST, &sourceListCopy);
			for (ClientList::const_iterator cIt = sourceListCopy.begin(); cIt != sourceListCopy.end(); cIt++)
			{
			//	Process waiting sources unless it is allowed to open a new connection
				if (g_App.m_pListenSocket->TooManySockets())
					break;

			//	The first request is always TCP request
				(*cIt)->AskForDownload();
			}
		}
		m_uSrcWaitForFileReq = m_SourceLists[DS_WAIT_FOR_FILE_REQUEST].size();
	}
	else if (dwIteration == 6)	// process LowID to LowID clients
	{
		if (!m_SourceLists[DS_LOWTOLOWIP].empty())
		{
			uint32	dwLocalServerIP = g_App.m_pServerConnect->GetCurrentServer()->GetIP();
			bool	bIsLowId = g_App.m_pServerConnect->IsLowID();

			GetCopySourceList(DS_LOWTOLOWIP, &sourceListCopy);
			for (ClientList::const_iterator cIt = sourceListCopy.begin(); cIt != sourceListCopy.end(); cIt++)
			{
				pSource = *cIt;

			//	Since we can't connect to the client we can keep it in the list only if it keeps to request the file(s) from us
				if ((dwCurTick - pSource->GetLastUpRequest()) > MAX_PURGEQUEUETIME)
				{
					g_App.m_pDownloadQueue->RemoveSource(pSource);
					continue;
				}

			//	Check local LowID status
				if (!bIsLowId)
				{
					if (pSource->GetServerIP() != dwLocalServerIP)
						pSource->SetDownloadState(DS_LOWID_ON_OTHER_SERVER);
					else if (pSource->HasClientNeededParts())
						pSource->SetDownloadState(DS_ONQUEUE);
					else
						pSource->SetDownloadState(DS_NONEEDEDPARTS);
				}
			}
		}
		m_uSrcLowToLow = m_SourceLists[DS_LOWTOLOWIP].size();
	}
	else if (dwIteration == 7)	// manage other sources (A4AF, request from the server)
	{
		uint16	uSourceNumber = GetSourceCount();

	//	A4AF management
	//	Swap sources with NNS status if possible
		bool bIsNNS;

		if (((!m_LastNoNeededCheck) || (dwCurTick - m_LastNoNeededCheck) > 30000))
		{
			m_LastNoNeededCheck = dwCurTick;
			GetCopySourceList(DS_NONEEDEDPARTS, &sourceListCopy);
			for (ClientList::const_iterator cIt = sourceListCopy.begin(); cIt != sourceListCopy.end(); cIt++)
			{
				pSource = *cIt;
			//	Check if A4AF Auto file is enable
				if (g_App.m_pDownloadQueue->GetA4AFAutoFile() == NULL)
				//	A4AF auto is disable => switch to any file
					pSource->SwapToAnotherFile(NULL);
				else if (g_App.m_pDownloadQueue->GetA4AFAutoFile() != this)
				{
				//	A4AF auto is enable & isn't current file => switch to A4AF auto file
					if (pSource->m_otherNoNeededMap.IsEmpty())
					{
						pSource->SwapToAnotherFile(g_App.m_pDownloadQueue->GetA4AFAutoFile());
					}
					else
					{
						bIsNNS = pSource->m_otherNoNeededMap.Lookup(g_App.m_pDownloadQueue->GetA4AFAutoFile(), dwLastNoNeededCheckTime);
						if (!bIsNNS || (dwCurTick - dwLastNoNeededCheckTime) > FILEREASKTIME)
						{
							pSource->SwapToAnotherFile(g_App.m_pDownloadQueue->GetA4AFAutoFile());
						}
					}
				}
			}
		}

	//	If this "File" is A4AF auto & A4AF list isn't empty,
	//	then check A4AF list and swap source to "this", if source isn't NNS
		if (g_App.m_pDownloadQueue->GetA4AFAutoFile() == this && !m_A4AFsrcList.IsEmpty())
		{
		//	Use a two-pointer loop to prevent exception due to swaping (delete)
			for (pos1 = m_A4AFsrcList.GetHeadPosition(); pos1 != NULL;)
			{
				pSource = m_A4AFsrcList.GetNext(pos1);
			// allow A4AF switch only for clients with static states 
			// with one exception "DS_WAIT_FOR_FILE_REQUEST" for paused files
				if ( (pSource != NULL) 
					&& (pSource->GetDownloadState() == DS_ONQUEUE 
						|| pSource->GetDownloadState() == DS_NONEEDEDPARTS
						|| pSource->GetDownloadState() == DS_LOWID_ON_OTHER_SERVER
						|| (pSource->GetDownloadState() == DS_WAIT_FOR_FILE_REQUEST
							&& GetStatus(false) == PS_PAUSED)) )
				{
					if (pSource->m_otherNoNeededMap.IsEmpty())
					{
						pSource->SwapToAnotherFile(this);
					}
					else
					{
						bIsNNS = pSource->m_otherNoNeededMap.Lookup(this,dwLastNoNeededCheckTime);
						if (!bIsNNS || (dwCurTick - dwLastNoNeededCheckTime) > FILEREASKTIME)
						{
							pSource->SwapToAnotherFile(this);
						}
					}
				}
			}
		}
		m_iSrcA4AF = m_A4AFsrcList.GetCount();

	//	Check if we want new sources from server
	//	1. connection state will be checked in DL-queue manager, so we don't need to do it here
	//	2. we can call SetRequiredSourcesRefresh() so many time as we want, cause this function
	//	   checks if the file exists.
		if ( ( (!m_dwLastFileSourcesRequestTime) || (dwCurTick - m_dwLastFileSourcesRequestTime) > SERVERREASKTIME)
			&& g_App.m_pPrefs->GetMaxSourcePerFileSoft() > uSourceNumber )
		{
		//	Local server
			m_dwLastFileSourcesRequestTime = dwCurTick;
			g_App.m_pDownloadQueue->SetRequiredSourcesRefresh(this);
		}

	//	Save sources in case there isn't DL (no need to do it so precise, so we can easily skip some ticks)
#ifdef OLD_SOCKETS_ENABLED
		if (g_App.m_pPrefs->SLSEnable() && m_SourceLists[DS_DOWNLOADING].empty() && g_App.m_pServerConnect->IsConnected())
		{
			m_sourcesaver.Process(this, g_App.m_pPrefs->SLSMaxSourcesPerFile());
		}
#endif //OLD_SOCKETS_ENABLED

	//	Calculate data rate, set limit etc. (every 5 sec)
		m_uProcessCounter++;
		if (m_uProcessCounter == 5)
		{
			m_uProcessCounter = 0;
			UpdateCompletedInfos();
			UpdateDisplayedInfo();
		}
	}

//	For status category filters check the status change
//	1. is transfering <-> is not transfering
//	2. is on the queue <-> is not on the queue
	if ( (nOldTransSourceCount == 0 && GetTransferringSrcCount() != 0)
		|| (nOldTransSourceCount != 0 && GetTransferringSrcCount() == 0)
	     || (iOldSrcOnQueueCount == 0 && GetOnQueueSrcCount() != 0)
	     || (iOldSrcOnQueueCount != 0 && GetOnQueueSrcCount() == 0) )
	{
		UpdateGUIAfterStateChange();
	}

	return m_dwDataRate;

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::AddSources(CMemFile& pSources, uint32 iServerIP, uint16 iServerPort)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

//	If this part file is paused or complete/completing or erroneous, no sources needed any more
	EnumPartFileStatuses	eFileStatus = GetStatus();

	if (eFileStatus == PS_STOPPED || eFileStatus == PS_COMPLETE
		|| eFileStatus == PS_COMPLETING || eFileStatus == PS_ERROR)
	{
		return;
	}

	byte byteCount;
	byte byteLowIDDropped = 0;
	byte bytePossibleSources = 0;

	pSources.Read(&byteCount, 1);

//	If there is a need for more sources just remove NNS
	int iSourceCountPlusNew = GetSourceCount() + byteCount;

	if ( iSourceCountPlusNew > g_App.m_pPrefs->GetMaxSourcePerFile()
		&& (m_dwLastPurgeTime + PURGE_TIME) < ::GetTickCount() )
	{
		RemoveNoNeededSources(iSourceCountPlusNew - g_App.m_pPrefs->GetMaxSourcePerFile());
	}

	uint32 dwUserID;
	uint16 iUserPort;

	for (int i = 0; i < static_cast<int>(byteCount); i++)
	{
		pSources.Read(&dwUserID,4);
		pSources.Read(&iUserPort,2);

#ifdef OLD_SOCKETS_ENABLED

	//	Check first if we are this source
		if (g_App.m_pServerConnect->GetClientID() < 0x1000000 && g_App.m_pServerConnect->IsConnected())
		{
			if ( g_App.m_pServerConnect->GetClientID() == dwUserID &&
				 g_App.m_pPrefs->GetPort() == iUserPort &&
				 inet_addr(g_App.m_pServerConnect->GetCurrentServer()->GetFullIP()) == iServerIP )
			{
				continue;
			}
		}
		else if ( g_App.m_pServerConnect->GetClientID() == dwUserID &&
				  g_App.m_pPrefs->GetPort() == iUserPort )
		{
			continue;
		}
		else if ( dwUserID < 0x1000000 &&
				  !g_App.m_pServerConnect->IsLocalServer(iServerIP, iServerPort) )
		{
			byteLowIDDropped++;
			continue;
		}
#endif //OLD_SOCKETS_ENABLED

		if (g_App.m_pPrefs->GetMaxSourcePerFile() > GetSourceCount())
		{
			bytePossibleSources++;
			CUpDownClient* pNewSource = g_App.m_pDownloadQueue->CheckAndAddSource(this, dwUserID, iUserPort, iServerIP, iServerPort, NULL);
			if (pNewSource)
				pNewSource->SetUserName(GetResString(IDS_SERVER_SOURCE));
		}
		else
			break;
	}

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	NewSrcPartsInfo & UpdateAvailablePartsCount & CompleteSourcesCount merged (CPU load)
void CPartFile::NewSrcPartsInfo()
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

//	Update the sources only if it's required
	if ((m_eStatus != PS_READY) && (m_eStatus != PS_EMPTY))
		return;

//	Cache part count
	uint32		dwPartCount = GetPartCount();

	CUpDownClient			*pSource;
	uint16					uNewCompleteSrcCount = 0;
	CArray<uint16,uint16>	PartFrequencyArray;
	ClientList				clientListCopy;

	PartFrequencyArray.SetSize(dwPartCount);	//	array is zero initialized inside

	GetCopySourceLists(SLM_ALL, &clientListCopy);
	for (ClientList::const_iterator cIt = clientListCopy.begin(); cIt != clientListCopy.end(); cIt++)
	{
		pSource = *cIt;

		uint32		dwSourcePartsCounter = 0;
	//	Count available parts from source
		for (uint32 i = 0; i < dwPartCount; i++)
		{
			if (pSource->IsPartAvailable(i))
			{
				PartFrequencyArray[i]++;
				dwSourcePartsCounter++;
			}
		}
	//	Count full sources
		if (dwSourcePartsCounter == dwPartCount)
			uNewCompleteSrcCount++;
	}

	m_uLastCompleteSrcCount = uNewCompleteSrcCount;

//	Increase size if necessary
	if (static_cast<uint32>(m_srcPartFrequencies.GetSize()) < dwPartCount)
	{
		m_srcPartFrequencies.SetSize(dwPartCount);
	}

	uint32	dwTmp, dwNewMaxSrcPartFrequency = 0;
	uint32	dwNewAvailablePartsCount = 0;
//	Count available parts over all sources
	for (uint32 i = 0; i < dwPartCount; i++)
	{
		dwTmp = static_cast<uint32>(PartFrequencyArray[i]);
		m_srcPartFrequencies[i] = static_cast<uint16>(dwTmp);

		if (dwTmp != 0)
		{
			dwNewAvailablePartsCount++;
			if (dwTmp > dwNewMaxSrcPartFrequency)
				dwNewMaxSrcPartFrequency = dwTmp;
		}
	}

	m_uMaxSrcPartFrequency = static_cast<uint16>(dwNewMaxSrcPartFrequency);
	m_dwAvailablePartsCount = dwNewAvailablePartsCount;

	if (dwNewAvailablePartsCount == dwPartCount)
	{
		lastseencomplete = CTime::GetCurrentTime();
	}

	UpdateDisplayedInfo();

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Intelligent Chunk Request(ICR)
uint16 CPartFile::GetNextRequiredPart(CUpDownClient *pSource)
{
#ifdef OLD_SOCKETS_ENABLED
	EMULE_TRY

	if (pSource == NULL)
		return 0xFFFF;

	uint32	dwGoodPart = pSource->GetLastDownPartAsked();

	if ((dwGoodPart != 0xFFFF) && (GetGapsInPart(dwGoodPart, NULL, 1) != 0))
	{
//		AddDebugLogLine(_T("Priority 0 (always choose currently downloaded part until it will be finished). File: %s, Part: %u"), this->GetFileName(), dwGoodPart);
		return dwGoodPart;
	}

	const uint32		dwPartCnt = GetPartCount();
	std::vector<uint32> PartsFilledSizeVector(dwPartCnt, NOT_AVAILABLE);	// Mark parts as not available to save CPU load
	uint32				dwNumAvailableNeededParts = 0;
	uint32				dwNumPartiallyBlockedParts = 0;

//	Scan for requested chunks in order to find out how much they're filled
	for (uint32 i = 0; i < dwPartCnt; i++)
	{
		if (pSource->IsPartAvailable(i))
		{
			if ((PartsFilledSizeVector[i] = GetPartLeftToDLSize(i)) != 0)
				dwNumAvailableNeededParts++;
		}
	}
	if (dwNumAvailableNeededParts == 0)
		return 0xFFFF;

//	Scan DownloadList for Blocked parts
	std::vector<byte>	abyteBlockedParts(dwPartCnt, FREE_TO_DL);
	uint32				dwTmp;

	for (POSITION pos = requestedblocks_list.GetHeadPosition(); pos != NULL;)
	{
		Requested_Block_Struct *pReqBlock = requestedblocks_list.GetNext(pos);

		dwTmp = pReqBlock->m_dwStartOffset / PARTSIZE;
		if ((PartsFilledSizeVector[dwTmp] == NOT_AVAILABLE)
			|| (abyteBlockedParts[dwTmp] == PARTIALLY_BLOCKED) || (abyteBlockedParts[dwTmp] == FULLY_BLOCKED))
			continue;
	//	Let's check if all gaps in part were already requested.
		if (GetGapsInPart(dwTmp, NULL, 1) == 0)
		{
		//	If we can't get this part from this client, reduce the count
			if (PartsFilledSizeVector[dwTmp] != 0)
				dwNumAvailableNeededParts--;
		//	No gaps found => fully blocked
			abyteBlockedParts[dwTmp] = FULLY_BLOCKED;
		}
		else
		{
		//	Count the blocked parts only in case if they are available & was not counted before
			dwNumPartiallyBlockedParts++;
			abyteBlockedParts[dwTmp] = PARTIALLY_BLOCKED;
		}
	}

//	All arrays/lists self-destruct in their destructors on return
	std::vector<uint32>	adwRareParts(dwPartCnt);	// Priority 2
	uint32				adwMovieParts[3];			// Priority 1 => only first+last parts
	std::vector<uint32>	adwRandomParts(dwPartCnt);	// Priority 4
	const uint32		dwPartCnt1 = dwPartCnt - 1, dwLastPartSize = GetLastPartSize();
	const uint32		dwPreviewMode = GetMovieMode();
//	Mode 0=NoPreview, 1=AVI(1F,1L), 4=MPG(1F), 5=MPG(2F), 7=AVI(1F,2L) F=First,L=Last

//	Define condition of rareness
//	Reduce this threshold for low number of sources to avoid downloading of the same part from several sources
	uint32		dwRarestPart = (m_uMaxSrcPartFrequency > 11) ? 4 : (m_uMaxSrcPartFrequency / 4 + 1);

//	Define a shortest-to-complete condition, cause it will be inverted if all needed parts are downloaded
	const bool	bShortestToComplete = (dwNumAvailableNeededParts > dwNumPartiallyBlockedParts);
	byte		byteMostCompletedRareState = PARTIALLY_BLOCKED;

	uint32		dwMostCompletedRarePartIdx, dwMostCompletedPartIdx;
	uint32		dwMostCompletedRarePartSz = PARTSIZE, dwMoviePartsCnt = 0;
	uint32		dwMostCompletedPartSz = (bShortestToComplete) ? PARTSIZE : 0;
	uint32		dwRareRank = ~0u, dwRandomRank = ~0u, dwRareVirginCnt = 0;
	uint32		dwRareFilledRank = 0, dwRarePartCnt = 0, dwRandomPartCnt = 0;

//	For each part in the part file...
	for (uint32 i = 0; i < dwPartCnt; i++)
	{
	//	If the part is available and it hasn't been completed...
		if (PartsFilledSizeVector[i] != NOT_AVAILABLE && abyteBlockedParts[i] != FULLY_BLOCKED && PartsFilledSizeVector[i] != 0)
		{
		//	------------------------ Rare Parts -------------------------
			if (dwRarestPart >= (dwTmp = m_srcPartFrequencies[i]))
			{
				if (PartsFilledSizeVector[i] == PARTSIZE || ((i == dwPartCnt1) && PartsFilledSizeVector[i] == dwLastPartSize))
				{
					if (dwTmp <= dwRareRank)	// Select the rarest parts
					{
						if (dwTmp < dwRareRank)
						{
							dwRareRank = dwTmp;
							dwRarePartCnt = 0;
							dwRareVirginCnt = 0;
						}
						adwRareParts[dwRarePartCnt++] = i;
					//	Calculate a number of untouched rare parts (which aren't currently requested)
						if (abyteBlockedParts[i] != PARTIALLY_BLOCKED)
							dwRareVirginCnt++;
					}
				}
			//	There can be situations when several rare sources are available for downloading,
			//	they can have the same rare parts, better to start downloading of different rare parts
			//	than download one from several sources. Besides selecting the shortest part to complete,
			//	we also take into account is this part already requested from another rare source or not
				else if ( ((byteMostCompletedRareState == abyteBlockedParts[i]) && (PartsFilledSizeVector[i] < dwMostCompletedRarePartSz)) ||
					((byteMostCompletedRareState != abyteBlockedParts[i]) && (byteMostCompletedRareState == PARTIALLY_BLOCKED)) )
				{
					byteMostCompletedRareState = abyteBlockedParts[i];
					dwMostCompletedRarePartSz = PartsFilledSizeVector[i];
					dwMostCompletedRarePartIdx = i;
					dwRareFilledRank = dwTmp;
				}
			}
		//	--------------------------- Movie or Archive ---------------------------
			switch (dwPreviewMode)
			{
				case 7:	// AVI(1F,2L) downloading sequence: chunk 0, last, next to last
					if ((i == 0) || ((dwPartCnt > 2) && (i == dwPartCnt - 2)))
						adwMovieParts[dwMoviePartsCnt++] = i;
					if (dwPartCnt > 2 && (i == dwPartCnt - 1))
					{
						if ((dwMoviePartsCnt != 0) && (adwMovieParts[dwMoviePartsCnt - 1] != 0))
						{
							adwMovieParts[dwMoviePartsCnt] = adwMovieParts[dwMoviePartsCnt - 1];
							adwMovieParts[dwMoviePartsCnt - 1] = i;
							dwMoviePartsCnt++;
						}
						else
							adwMovieParts[dwMoviePartsCnt++] = i;
					}
					break;

				case 5:	// MPG(2F) downloading sequence: chunk 0, chunk 1
					if ((i == 0) || ((i == 1) && (dwPartCnt > 1)))
						adwMovieParts[dwMoviePartsCnt++] = i;
					break;

				case 1:	// AVI(1F,1L) downloading sequence: chunk 0, last chunk
					if (dwPartCnt > 1 && (i == dwPartCnt1))
						adwMovieParts[dwMoviePartsCnt++] = i;
				case 4:	// MPG(1F)
					if (i == 0)
						adwMovieParts[dwMoviePartsCnt++] = 0;
			}
		//	--------------------- widespreaded parts ---------------------
		//	Check blocked parts - don't download already downloading parts unless there's something else
			if ((abyteBlockedParts[i] == PARTIALLY_BLOCKED) && bShortestToComplete)
				continue;

			if (PartsFilledSizeVector[i] == PARTSIZE || ((i == dwPartCnt1) && PartsFilledSizeVector[i] == dwLastPartSize))
			{
			//	Fresh part
				if ((dwTmp = m_srcPartFrequencies[i]) <= dwRandomRank)
				{
					if (dwTmp < dwRandomRank)	// Select the rarest parts among them
					{
						dwRandomRank = dwTmp;
						dwRandomPartCnt = 0;
					}
					adwRandomParts[dwRandomPartCnt++] = i;
				}
			}
			else
			{
			//	Invert condition shortest-to-complete condition if all parts are partially blocked:
			//	1) select the most completed part if there're chunks not in downloading state
			//	2) select the most incomplete part if all available parts from this source're in downloading state
				if (!((PartsFilledSizeVector[i] < dwMostCompletedPartSz) ^ bShortestToComplete))
				{
					dwMostCompletedPartIdx = i;
					dwMostCompletedPartSz = PartsFilledSizeVector[i];
				}
			}
		}
	}

//	Select final result
//	---------------------------------- Movies Part  -----------------------------------------------------
//	The preview parts have highest priority in order to provide the user information about the file as soon as possible
//	This will lead to the faster decision about futher download
	if (dwMoviePartsCnt != 0)
	{
	//	Select the most appropriate part (for the case when all preview parts are already in downloading)
		dwGoodPart = adwMovieParts[0];
	//	Avoid several downloads for the same part if there's a preview part which isn't in downloading state yet
		i = 0;
		do
		{
			dwTmp = adwMovieParts[i];
			if (abyteBlockedParts[dwTmp] != PARTIALLY_BLOCKED)
			{
				dwGoodPart = dwTmp;
				break;
			}
		} while (++i < dwMoviePartsCnt);
//		AddDebugLogLine(_T("Priority 1 (Preview part). File: %s, Part: %u"), this->GetFileName(), dwGoodPart);
	}
//	---------------------------------- Rare Parts ------------------------------------------------------
	else if ((dwRarePartCnt != 0) || (dwRareFilledRank != 0))
	{
	//	Download an incomplete part first, but if there're more than one source for an incomplete part
	//	and there's a fresh part with only one source, start downloading of the unique fresh part;
	//	Also if an incomplete part is currently being downloading and there is a chance
	//	to start download of fresh rare part, choose the fresh one
		if ((dwRareFilledRank != 0) && ( (dwRareFilledRank == 1) || (dwRarePartCnt == 0) ||
			((dwRareRank != 1) && (abyteBlockedParts[dwMostCompletedRarePartIdx] != PARTIALLY_BLOCKED)) ) )
		{
			dwGoodPart = dwMostCompletedRarePartIdx;
//			AddDebugLogLine(_T("Priority 2 (Rarest unfinished part). File: %s, Part: %u"), this->GetFileName(), dwGoodPart);
		}
		else
		{
		//	If there is only one rare part, then take it, otherwise take random one
			if (dwRarePartCnt == 1)
				dwTmp = 0;
			else
			{
			//	There can be situations when several rare sources are available for downloading,
			//	they can have the same rare parts, better to start downloading of different rare parts
			//	than download one from several sources. Also due to reply delay and data buffering,
			//	a part can be seen as untouched (PartsFilledSizeVector[i] == PARTSIZE), but as
			//	the download request was already sent, this part also should be treated as occupied
				if ((dwRarePartCnt == dwRareVirginCnt) || (dwRareVirginCnt == 0))
					dwTmp = static_cast<uint32>(rand() * (dwRarePartCnt - 1) + RAND_MAX / 2) / RAND_MAX;
				else
				{
					dwTmp = static_cast<uint32>(rand() * (dwRareVirginCnt - 1) + RAND_MAX / 2) / RAND_MAX;
					for (i = 0;; i++)
					{
						if ((abyteBlockedParts[adwRareParts[i]] != PARTIALLY_BLOCKED) && (static_cast<int>(--dwTmp) < 0))
						{
							dwTmp = i;
							break;
						}
					}
				}
			}
			dwGoodPart = adwRareParts[dwTmp];
//			AddDebugLogLine(_T("Priority 3 (Rarest unrequested part). File: %s, Part: %u"), this->GetFileName(), dwGoodPart);
		}
	}
//	--------------------------------------- Requested ----------------------------------------------------
//	Try to complete an UnfinishedPart and MostDownloaded as soon as possible to become a new source
//	it doesn't matter which one to download, if there are same filled parts.
	else if ((dwMostCompletedPartSz != PARTSIZE) && (dwMostCompletedPartSz != 0))
	{
		dwGoodPart = dwMostCompletedPartIdx;
//		AddDebugLogLine(_T("Priority 4 (Widespread unfinished part). File: %s, Part: %u"), this->GetFileName(), dwGoodPart);
	}
//	-------------------------------- Unrequested & Random -------------------------------------------------
	else if (dwRandomPartCnt != 0)
	{
		if (dwRandomPartCnt == 1)
			dwTmp = 0;
		else
			dwTmp = static_cast<uint32>(rand() * (dwRandomPartCnt - 1) + RAND_MAX / 2) / RAND_MAX;
		dwGoodPart = adwRandomParts[dwTmp];
//		AddDebugLogLine(_T("Priority 5 (Widespread unrequested part). File: %s, Part: %u"), this->GetFileName(), dwGoodPart);
	}
	else
	{
		dwGoodPart = 0xFFFF;
	}

/*
	CString strDebug;

	strDebug.Format(_T("ICR-Test: File: %s, A-parts : %u, BP-Parts %u. Choosen Part: %u"),
						GetFileName(), dwNumAvailableNeededParts, dwNumPartiallyBlockedParts, dwGoodPart);
	if (dwGoodPart < dwPartCnt)
	{
		strDebug.AppendFormat(_T(", P-Freq: %u, blocked %s"), m_srcPartFrequencies[dwGoodPart], ((abyteBlockedParts[dwGoodPart] == PARTIALLY_BLOCKED)? _T("yes"):_T("no")));
	}
	AddDebugLogLine(strDebug);
*/

	return dwGoodPart;

	EMULE_CATCH
#endif //OLD_SOCKETS_ENABLED
	return 0xFFFF;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CPartFile::GetPartFilledSize(uint32 dwPartNum) const
{
	uint32 dwStart = dwPartNum * PARTSIZE;
	uint32 dwEnd = ((GetFileSize() - dwStart) > PARTSIZE) ? (dwStart + PARTSIZE - 1) : (GetFileSize() - 1);

	uint32 dwLeftSize = (dwEnd - dwStart + 1) - GetPartLeftToDLSize(dwPartNum);

	ASSERT(dwLeftSize <= PARTSIZE);
	return dwLeftSize;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::RemoveBlockFromList(uint32 start, uint32 dwEnd)
{
	EMULE_TRY

	POSITION pos1, pos2;
	for (pos1 = requestedblocks_list.GetHeadPosition(); (pos2 = pos1) != NULL;)
	{
		if (requestedblocks_list.GetNext(pos1)->m_dwStartOffset <= start && requestedblocks_list.GetAt(pos2)->m_dwEndOffset >= dwEnd)
			requestedblocks_list.RemoveAt(pos2);
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::RemoveAllRequestedBlocks(void)
{
	EMULE_TRY

	requestedblocks_list.RemoveAll();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::CompleteFile(bool bIsHashingDone)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	StopFile(false);

//	Show completing instead of paused
	SetStatus(PS_COMPLETING);

	g_App.m_pMDlg->m_wndTransfer.UpdateCatTabTitles();

	UpdateDisplayedInfo();

//	If the part file isn't already hashed...
	if (!bIsHashingDone)
	{
	//	Create a thread to hash the part file in the background
		m_dwDataRate = 0;

		CString	partFileNameBase = m_strPartMetFileName.Left(m_strPartMetFileName.GetLength() - 4);
		CHashFileThread	*addFileThread;

		addFileThread = (CHashFileThread*)AfxBeginThread( RUNTIME_CLASS(CHashFileThread),
		                THREAD_PRIORITY_BELOW_NORMAL + g_App.m_pPrefs->GetMainProcessPriority(),
		                0,
		                CREATE_SUSPENDED );
		addFileThread->SetValues(0, m_strTempDir, partFileNameBase, this);
		addFileThread->ResumeThread();
		return;
	}
	else
	{
		CWinThread *pThread = AfxBeginThread((AFX_THREADPROC)CompleteThreadProc, this, THREAD_PRIORITY_BELOW_NORMAL + g_App.m_pPrefs->GetMainProcessPriority(), 0, 0);

		if (pThread == NULL)
		{
			throw CString(_T("error creating file complete thread"));
		}
	}
	GetAvgDataRate(true);

	g_App.m_pMDlg->m_wndTransfer.UpdateCatTabTitles();

	UpdateDisplayedInfo();
	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
static void GenerateNewFileName(const CString& outputDir, CString& newname)
{
	ASSERT(!newname.IsEmpty());

	int namecount = 0;

	TCHAR p_fname[_MAX_FNAME];
	TCHAR p_fext[_MAX_EXT];
	_tsplitpath(newname, NULL, NULL, p_fname, p_fext);

	CString fname(p_fname);
	CString fext(p_fext);

//	Search for matching ()s and check if it contains a number inside
	if (fname.Right(1) == ')')
	{
		int ll = fname.GetLength();
		int ob = fname.ReverseFind( '(' );
	//	Check for ) existence or () situation
		if (ob != -1 && ob + 2 != ll)
		{
			bool found = true;
			for (int i = ob + 1; i < ll - 1; i++)
			{
			//	Check for digits inside brackets
				if (!_istdigit(fname[i]))
				{
					found = false;
					break;
				}
			}
			if (found)
			{
				namecount = _ttoi(fname.Mid(ob + 1, ll - ob - 2));
				fname.Truncate(ob);
			}
		}
	}

	do
	{
		namecount++;
		newname.Format(_T("%s\\%s(%d)%s"), outputDir, fname, namecount, fext);
	}
	while (PathFileExists(newname));
}

//	Use threads for file completion
BOOL CPartFile::PerformFileComplete(void)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

//	Removes any old fakes.rar that may exist in shared directories before completing the newest fakes.rar
	if (IsFakesDotRar())
	{
		CStringList	strSharedDirsList;
		CString		strFakesFile;

		strFakesFile.Format(_T("%s\\fakes.rar"), g_App.m_pPrefs->GetIncomingDir());
		if (PathFileExists(strFakesFile))
			if (!::DeleteFile(strFakesFile))
				AddDebugLogLine(RGB_LOG_ERROR_TXT _T("Couldn't delete fakes.rar (%s)"), strFakesFile);


		g_App.m_pPrefs->SharedDirListCopy(&strSharedDirsList);

		for (POSITION pos = strSharedDirsList.GetHeadPosition(); pos != NULL; )
		{
			strFakesFile.Format(_T("%s\\fakes.rar"), strSharedDirsList.GetNext(pos));
			if (PathFileExists(strFakesFile))
				if (!::DeleteFile(strFakesFile))
					AddDebugLogLine(RGB_LOG_ERROR_TXT _T("Couldn't delete fakes.rar (%s)"), strFakesFile);
		}
	}

//	Get current category
	EnumCategories eCatID = GetCatID();

	CString strFullPartFileName(m_strFilePath);	// save .part file name for deletion

	CString newname, strOutputDir = GetOutputDir();
	newname.Format(_T("%s\\%s"), strOutputDir, g_App.StripInvalidFilenameChars(GetFileName()));

//	Create all required directories
	CreateAllDirectories(&strOutputDir);

	bool bRenamed = false;
	if (PathFileExists(newname))
	{
		bRenamed = true;
		GenerateNewFileName(strOutputDir, newname);
	}

//	Save full .met file name for temp. file deletions, because a name is altered before it
	CString	strTempFullName = m_strFullName;

//	Make copy if files are on different partitions and a file isn't small (> 1 Mb)
//	Small files will be copied by system routine which blocks uploading start of the same file
	if ( (_toupper(m_strFilePath[0]) != _toupper(newname[0])) &&
		(m_strFilePath[1] == _T(':') && newname[1] == _T(':')) && (GetFileSize() > 1024*1024) )
	{
		bool	bException = false;
		CFile	fInputFile, fOutputFile;
		CString	strTempNew = newname;

		strTempNew += _T(".$$$");
		try
		{
		//	We going to write everything into the temporary file at the same place the newname located
		//	and after we finished - rename, this will take care of synchronization problem with SharedFiles
			if ( !fInputFile.Open( m_strFilePath, CFile::modeRead | CFile::shareDenyNone | CFile::typeBinary
			                       | CFile::osSequentialScan ) )
			{
			//	Let's catch to care about closing everything
				throw CString(_T("can't open file for reading (") + m_strFilePath + _T(')'));
			}
			if ( !fOutputFile.Open( strTempNew, CFile::modeWrite | CFile::shareExclusive | CFile::osSequentialScan
								   | CFile::modeCreate ) )
			{
			//	Let's catch to care about closing everything
				throw CString(_T("can't open file for writing (") + strTempNew + _T(')'));
			}

			int		iBlockSize = g_App.m_pPrefs->SlowCompleteBlockSize() * 1024;
			char	*pcBuffer = new char[iBlockSize];
			DWORD	dwRead;

		//	Only one extensive hard drive access at a time to reduce I/O load
			CSingleLock Lock(&g_App.m_csPreventExtensiveHDAccess, TRUE);

		//	Most of the work done here
		//	just copy the file data from one to another
			do
			{
				dwRead = fInputFile.Read(pcBuffer, iBlockSize);
				fOutputFile.Write(pcBuffer, dwRead);
			//	Release the CPU for other processes
				Sleep(0);
			}
			while (dwRead > 0);

			Lock.Unlock();
			delete []pcBuffer;

			fInputFile.Close();
			fOutputFile.Close();

		//	Rename temporary file back to normal name
			if (_trename(strTempNew, newname))
			{
			//	Lets catch to care about closing everything
				throw CString(_T("renaming of completed file failed"));
			}

		//	Synchronization with reading upload thread
			CSingleLock lockComplete(&m_csFileCompletion, TRUE);

			SetPath(strOutputDir);
			SetFilePath(newname);
			m_bPaused = false;
			m_eStatus = PS_COMPLETE;
			m_iSrcA4AF = 0;

			lockComplete.Unlock();

		//	Update GUI after synchronization is released
			UpdateGUIAfterStateChange();

		//	Close permanent handles
			try
			{
				ClosePartFile();
			}
			catch (CFileException *error)
			{
				OUTPUT_DEBUG_TRACE();
				AddDebugLogLine(_T("Failed to close permanent handle %s - %s"), GetFileName(), GetErrorMessage(error));
				error->Delete();
			}
			::DeleteFile(strFullPartFileName);
		}
	//	Check for exception that will be thrown by "new"
		catch (CMemoryException * error)
		{
			error->Delete();
			AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_OUTMEM));
			bException = true;
		}
		catch (CString str)
		{
			AddLogLine(true, RGB_LOG_ERROR + str);
			bException = true;
		}
		catch (CException *error)
		{
		//	Sending message for erroneous file
			CString MessageText;
			MessageText.Format(GetResString(IDS_ERR_COMPLETIONFAILED), GetFileName(), GetErrorMessage(error));
			g_App.m_pSMTPConnection->SendMail(MessageText, g_App.m_pPrefs->GetNotifierPopOnImportantError(), g_App.m_pPrefs->IsSMTPWarningEnabled());
			g_App.m_pMDlg->ShowNotifier(MessageText, TBN_IMPORTANTEVENT, false, g_App.m_pPrefs->GetNotifierPopOnImportantError());
			AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_COMPLETIONFAILED), GetFileName(), GetErrorMessage(error));
			error->Delete();
			bException = true;
		}
		if (bException)
		{
		//	File should be closed to make Initialization possible
			try
			{
				ClosePartFile();
			}
			catch (CFileException *error)
			{
				OUTPUT_DEBUG_TRACE();
				AddDebugLogLine(_T("Failed to close permanent handle %s - %s"), GetFileName(), GetErrorMessage(error));
				error->Delete();
			}

			if (fOutputFile.m_hFile != INVALID_HANDLE_VALUE)
				fOutputFile.Close();
		//	Delete temporary file if it still exists
			::DeleteFile(strTempNew);

			m_bPaused = true;
			SetStatus(PS_ERROR);
		//	If fInputFile is still opened, it will be closed here in its destructor
			return FALSE;
		}
	}
	else
	{
	//	Synchronization with reading upload thread
		CSingleLock lockComplete(&m_csFileCompletion, TRUE);

	//	Close permanent handles
		try
		{
			ClosePartFile();
		}
		catch (CFileException *error)
		{
			OUTPUT_DEBUG_TRACE();
			AddDebugLogLine(_T("Failed to close permanent handle %s - %s"), GetFileName(), GetErrorMessage(error));
			error->Delete();
		}

		if (_trename(m_strFilePath, newname))
		{
			lockComplete.Unlock();

		//	Sending message for erroneous file
			CString MessageText;
			MessageText.Format(GetResString(IDS_ERR_COMPLETIONFAILED), GetFileName(), _tcserror(errno));
			g_App.m_pSMTPConnection->SendMail(MessageText, g_App.m_pPrefs->GetNotifierPopOnImportantError(), g_App.m_pPrefs->IsSMTPWarningEnabled());
			g_App.m_pMDlg->ShowNotifier(MessageText, TBN_IMPORTANTEVENT, false, g_App.m_pPrefs->GetNotifierPopOnImportantError());
			AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_COMPLETIONFAILED), GetFileName(), _tcserror(errno));
			AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_COMPLETIONFAILED2));
			m_bPaused = true;
			SetStatus(PS_ERROR);
			return FALSE;
		}

		SetPath(strOutputDir);
		SetFilePath(newname);
		m_bPaused = false;
		m_eStatus = PS_COMPLETE;
		m_iSrcA4AF = 0;

		lockComplete.Unlock();

	//	Update GUI after synchronization is released
		UpdateGUIAfterStateChange();
	}

//	To have the accurate date stored in known.met we have to update the 'date' of a just completed file.
//	if we don't update the file date here (after commiting the file and before adding the record to known.met),
//	that file will be rehashed at next startup and there would also be a duplicate entry (hash+size) in known.met because of different file date!
	struct _stat st;
	if (_tstat(newname, &st) == 0)
	{
		m_timetLastWriteDate = st.st_mtime;
		AdjustNTFSDaylightFileTime(&m_timetLastWriteDate, newname);
	}

	if (_tremove(strTempFullName))
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), strTempFullName);

	CString	strFileToRemove;

//	Remove .dir file
	strFileToRemove = strFullPartFileName + _T(".dir");
	if (_tremove(strFileToRemove) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), strFileToRemove);

//	Remove .bak file
	strFileToRemove = strTempFullName + _T(".bak");
	if (_tremove(strFileToRemove) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), strFileToRemove);

//	Remove .settings file
	strFileToRemove = strFullPartFileName + _T(".settings");
	if (_tremove(strFileToRemove) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), strFileToRemove);

//	Remove .txtsrc file
	strFileToRemove = strFullPartFileName + _T(".txtsrc");
	if (_tremove(strFileToRemove) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), strFileToRemove);

//	Remove the stats file
	strFileToRemove = strFullPartFileName + _T(".stats");
	if (_tremove(strFileToRemove) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), strFileToRemove);

//	Change AutoPriority to NORMAL when UAP is disabled
	if (IsULAutoPrioritized() && !g_App.m_pPrefs->IsUAPEnabled())
	{
		SetAutoULPriority(false);
		SetULPriority(PR_NORMAL);
	}

	UpdateDisplayedInfo();

	CCat		*pCat = CCat::GetCatByID(GetCatID());
	CString		strCategory = (pCat != NULL) ? pCat->GetTitle() : GetResString(IDS_CAT_UNCATEGORIZED);

	AddLogLine(true, RGB_LOG_SUCCESS + GetResString(IDS_DOWNLOADDONE), GetFileName(), strCategory);

//	Sending message for completed file
	CString MessageText;
	MessageText.Format(GetResString(IDS_TBN_DOWNLOADDONE), GetFileName(), strCategory);
	g_App.m_pSMTPConnection->SendMail(MessageText, g_App.m_pPrefs->GetUseDownloadNotifier(), g_App.m_pPrefs->IsSMTPInfoEnabled());
	g_App.m_pMDlg->ShowNotifier(MessageText, TBN_DLOAD, false, g_App.m_pPrefs->GetUseDownloadNotifier());
	if (bRenamed)
		AddLogLine(true, IDS_DOWNLOADRENAMED, _tcsrchr(newname, '\\') + 1);
	g_App.m_pKnownFilesList->SafeAddKnownFile(this);
	g_App.m_pDownloadQueue->RemoveFile(this);
	UpdateDisplayedInfo();

#ifdef OLD_SOCKETS_ENABLED
	g_App.m_pMMServer->AddFinishedFile(this);
#endif //OLD_SOCKETS_ENABLED

//	Starts next paused file
	if (g_App.m_pPrefs->DownloadPausedOnComplete() && !IsFakesDotRar())
	{
		g_App.m_pDownloadQueue->StartNextFile(eCatID);
	}

	g_App.m_pPrefs->Add2DownCompletedFiles();			// Increments cumDownCompletedFiles in prefs struct
	g_App.m_pPrefs->Add2DownSessionCompletedFiles(); 	// Increments sesDownCompletedFiles in prefs struct
	g_App.m_pPrefs->SaveCompletedDownloadsStat();		// Saves cumDownCompletedFiles to INI

//	Republish a file on the server to update server complete sources counter
	g_App.m_pSharedFilesList->RepublishFile(this);

	if (GetPath() != g_App.m_pPrefs->GetIncomingDir())
	{
		if (g_App.m_pPrefs->SharedDirListCheckAndAdd(GetPath(), true))
		{
		//	New path was added to the list, scan for files
			g_App.m_pSharedFilesList->Reload();
		}
	}

	if (g_App.m_pPrefs->IsAutoClearCompleted())
	{
		g_App.m_pMDlg->m_wndTransfer.m_ctlDownloadList.SetRedraw(false);
		g_App.m_pDownloadList->ClearCompleted(CAT_ALL);
		g_App.m_pMDlg->m_wndTransfer.m_ctlDownloadList.SetRedraw(true);
	}

	g_App.m_pMDlg->m_wndTransfer.UpdateCatTabTitles();

//	Scan completed files for viruses
	if (g_App.m_pPrefs->IsAVEnabled() &&
		g_App.m_pPrefs->IsAVScanCompleted() && !g_App.m_pPrefs->GetAVPath().IsEmpty())
	{
		strTempFullName.Format(_T("%s \"%s\""), g_App.m_pPrefs->GetAVParams(), m_strFilePath);
		ShellExecute(NULL, _T("open"), g_App.m_pPrefs->GetAVPath(), strTempFullName, NULL, SW_SHOW);
	}

//	FakeCheck
	if (IsFakesDotRar())
	{
		CFakeCheck	*pFakeCheck = g_App.m_pFakeCheck;
		CString		strFakesDotDatPath = g_App.m_pPrefs->GetConfigDir() + _T("fakes.dat");
		CString		strBackupFilePath(strFakesDotDatPath);

		strBackupFilePath += _T(".bak");

	//	Backups fakes.dat
		::MoveFile(strFakesDotDatPath, strBackupFilePath);

		if (pFakeCheck->ExtractRARArchive(m_strFilePath, g_App.m_pPrefs->GetConfigDir()))
		{
			g_App.AddLogLine(true, RGB_LOG_SUCCESS + GetResString(IDS_FAKE_SUCCESS_UPDATE), g_App.m_pPrefs->GetFakesDatVersion(), g_App.m_pPrefs->GetDLingFakeListVersion());
			g_App.m_pPrefs->SetFakesDatVersion(g_App.m_pPrefs->GetDLingFakeListVersion());
			g_App.m_pPrefs->SetDLingFakeListVersion(0);
			g_App.m_pPrefs->SetDLingFakeListLink(_T(""));
			pFakeCheck->LoadFromDatFile();
			::DeleteFile(strBackupFilePath);
		}
		else
		{
		//	The extraction was unsuccessful, so we restore fakes.dat
			::DeleteFile(strFakesDotDatPath);
			::MoveFile(strBackupFilePath, strFakesDotDatPath);
		}

		CKnownFile	*pThisFile = g_App.m_pSharedFilesList->GetFileByID(GetFileHash());

		if (pThisFile != NULL)
		{
			pThisFile->SetAutoULPriority(false);
			pThisFile->SetULPriority(PR_RELEASE);
			g_App.m_pSharedFilesList->Reload();
		}
	}

	return TRUE;

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE

	return FALSE;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::RemoveAllSources(bool bTryToSwap)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

//	TODO transfer sources to other downloading files if possible
	CUpDownClient	*pRemoveSource;
	ClientList		clientListCopy;

	GetCopySourceLists(SLM_ALL, &clientListCopy);

	ClearSourceLists();

	for (ClientList::const_iterator cIt = clientListCopy.begin(); cIt != clientListCopy.end(); cIt++)
	{
		pRemoveSource = *cIt;

		if (bTryToSwap)
		{
			if (!pRemoveSource->SwapToAnotherFile(NULL, A4AF_REMOVE))
			{
			//	If the switching wasn't successful we remove the source from the download queue
				g_App.m_pDownloadQueue->RemoveSource(pRemoveSource);
			}
			else
			{
			//	If the switching was successful we need to remove the entry from graphical list
				g_App.m_pDownloadList->RemoveSource(pRemoveSource, this);
			}
		}
		else
		{
		//	We weren't asked to swap so we remove it
			g_App.m_pDownloadQueue->RemoveSource(pRemoveSource);
		}
	}

//	Remove A4AF entries from graphical list & anotherRequestList
	if (!m_A4AFsrcList.IsEmpty())
	{
		for (POSITION pos = m_A4AFsrcList.GetHeadPosition(); pos != NULL;)
		{
			pRemoveSource = m_A4AFsrcList.GetNext(pos);
			g_App.m_pDownloadList->RemoveSource(pRemoveSource, this);
			pRemoveSource->RemoveRequestForAnotherFile(this);
		}
	}

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::DeleteFile()
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	ASSERT (!m_bPreviewing);

	CString	strFileToRemove;

	GetResString(&strFileToRemove, IDS_CANCELLED);
	strFileToRemove += _T(": %s");
	AddLogLine(false, strFileToRemove, GetFileName());

	m_bIsBeingDeleted = true;

	StopFile();

	ClosePartFile();

	if (_tremove(m_strFullName))
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), m_strFullName);

	if (_tremove(m_strFilePath))
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), m_strFilePath);

	CString	strFilePath = RemoveFileExtension(m_strFullName);

//	Remove .dir file
	strFileToRemove = strFilePath + _T(".dir");
	if (_tremove(strFileToRemove) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), strFileToRemove);

//	Remove .bak file
	strFileToRemove = m_strFullName + _T(".bak");
	if (_tremove(strFileToRemove) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), strFileToRemove);

//	Remove .settings file
	strFileToRemove = strFilePath + _T(".settings");
	if (_tremove(strFileToRemove) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), strFileToRemove);

//	Remove .txtsrc file
	strFileToRemove = strFilePath + _T(".txtsrc");
	if (_tremove(strFileToRemove) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), strFileToRemove);

//	Remove the stats file
	strFileToRemove = strFilePath + _T(".stats");
	if (_tremove(strFileToRemove) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), strFileToRemove);

//	File properties get deleted inside one of these calls now...
	g_App.m_pSharedFilesList->RemoveFile(this);
	g_App.m_pDownloadQueue->RemoveFile(this);
	g_App.m_pDownloadList->RemoveFile(this);

	delete this;

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::HashSinglePart(uint32 dwPartNum)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	UINT dwResErrStrID;
	uint32 dwPartCnt = static_cast<uint32>(GetPartCount());

	if ((GetHashCount() <= dwPartNum) && (dwPartCnt > 1))
	{
		dwResErrStrID = IDS_ERR_HASHERRORWARNING;
	}
	else if (!GetPartHash(dwPartNum) && (dwPartCnt != 1))
	{
		dwResErrStrID = IDS_ERR_INCOMPLETEHASH;
	}
	else
	{
		uchar	hashresult[16];
	//	Duplicate part file read handle to eliminate read collisions
		CFile	*pPartFile = m_hPartFileRead.Duplicate();

		pPartFile->Seek((LONGLONG)PARTSIZE * dwPartNum, CFile::begin);

		uint32 length = PARTSIZE;

		if ((ULONGLONG)PARTSIZE * (dwPartNum + 1) > pPartFile->GetLength())
		{
			length = (pPartFile->GetLength() - ((ULONGLONG)PARTSIZE * dwPartNum));
		}
		ASSERT(length <= PARTSIZE);
		CreateHashFromFile(pPartFile, length, hashresult);

	//	File handle is closed in destructor
		delete pPartFile;

		return (md4cmp(hashresult, (GetFileSize() >= PARTSIZE) ? GetPartHash(dwPartNum) : m_fileHash) == 0) ? true : false;
	}

	AddLogLine(true, RGB_LOG_WARNING + GetResString(dwResErrStrID), GetFileName());
	this->m_bHashSetNeeded = true;
	return true;

	EMULE_CATCH

#endif //NEW_SOCKETS_ENGINE
	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Allow to preview zip/rar files
bool CPartFile::IsArchive(bool bOnlyPreviewable)
{
	if (bOnlyPreviewable)
		return m_bIsPreviewableArchive;

	return (ED2KFT_ARCHIVE == GetFileType());
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::SetPriority(byte byteNewPriority, bool bSaveSettings /*= true*/)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	priority = byteNewPriority;
	g_App.m_pDownloadQueue->SortByPriority();
	UpdateDisplayedInfo();
	if (bSaveSettings)
		SavePartFile();

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::StopFile(bool bUpdateDisplay /* TRUE */)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	if (m_eStatus == PS_COMPLETE || m_eStatus == PS_COMPLETING)
		return;

	ClientList	DownloadingListCopy;

//	Change status as soon as possible in order to prevent
//	unpredictable behaviour	during the source removal
	m_bPaused = true;
	m_bStopped = true;

//	Need to tell any connected clients to stop sending the file
//	and send downloading clients a cancel (NOT if completing)
	GetCopySourceList(DS_DOWNLOADING, &DownloadingListCopy);
	for (ClientList::const_iterator cIt = DownloadingListCopy.begin(); cIt != DownloadingListCopy.end(); cIt++)
	{
		CUpDownClient	*pSource = *cIt;

#ifdef OLD_SOCKETS_ENABLED
		pSource->SendCancelTransfer();
		pSource->SetDownloadState(DS_ONQUEUE);
#endif //OLD_SOCKETS_ENABLED
	}

//	Reset A4AF auto file
	if (g_App.m_pDownloadQueue->GetA4AFAutoFile() == this)
	{
		g_App.m_pDownloadQueue->SetA4AFAutoFile(NULL);
	}

	RemoveAllSources(true);

	m_dwDataRate = 0;
	m_iNumTransferringSrcs = 0;
	m_iSrcNNP = 0;
	m_iSrcOnQueue = 0;
	m_uSrcHighQR = 0;
	m_iSrcConnecting = 0;
	m_uSrcWaitForFileReq = 0;
	m_iSrcConnected = 0;
	m_iSrcConnViaServer = 0;
	m_uSrcLowToLow = 0;
	m_uSrcLowIDOnOtherServer = 0;
	m_iSrcQueueFull = 0;
	m_iSrcA4AF = 0;
	m_A4AFsrcList.RemoveAll();

//	Save info to disk
	if (!m_bIsBeingDeleted)
	{
		FlushBuffer();
		SavePartFile();
		SaveSettingsFile();
	}

//	Update GUI information (if the file was stopped for completing GUI will be updated in CompleteFile)
	if (bUpdateDisplay)
	{
		UpdateDisplayedInfo();
		g_App.m_pMDlg->m_wndTransfer.UpdateCatTabTitles();
	}

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::PauseFile()
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	if ((m_eStatus == PS_COMPLETE) || (m_eStatus == PS_COMPLETING))
		return;

	CUpDownClient	*pSource;
	ClientList		clientListCopy;

//	Reset A4AF auto file
	if (g_App.m_pDownloadQueue->GetA4AFAutoFile() == this)
	{
		g_App.m_pDownloadQueue->SetA4AFAutoFile(NULL);
	}

//	Send CANCEL to downloading clients
	GetCopySourceList(DS_DOWNLOADING, &clientListCopy);
	for (ClientList::const_iterator cIt = clientListCopy.begin(); cIt != clientListCopy.end(); cIt++)
	{
		pSource = *cIt;
		pSource->SendCancelTransfer();
		pSource->SetDownloadState(DS_ONQUEUE);
	}
//	At this point nobody is downloading so we can try to swap sources
	GetCopySourceLists(SLM_ALL, &clientListCopy);
	for (ClientList::const_iterator cIt = clientListCopy.begin(); cIt != clientListCopy.end(); cIt++)
	{
		pSource = *cIt;
		pSource->SwapToAnotherFile(NULL, A4AF_IGNORE_TIME_LIMIT);
	}

	m_bPaused = true;
	m_bStopped = false;

	UpdateGUIAfterStateChange();

	m_dwDataRate = 0;
	m_iNumTransferringSrcs = 0;
	UpdateDisplayedInfo();

	if (!m_bIsBeingDeleted)
	{
		SavePartFile();
		SaveSettingsFile();
	}

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::ResumeFile()
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	if (m_eStatus == PS_COMPLETE || m_eStatus == PS_COMPLETING)
		return;

//	Check if chunks are availible & set status
	EnumPartFileStatuses	eNewStatus = PS_EMPTY;

	for (uint32 ui = 0; ui < GetPartCount(); ui++)
	{
		if (IsPartComplete(ui))
		{
			eNewStatus = PS_READY;
			break;
		}
	}
	SetStatus(eNewStatus);

	if (GetStartTimeReset())
	{
	//	Needed for files which are paused after adding or loading
		m_SessionStartTime = CTime::GetCurrentTime();
		SetStartTimeReset(false);
	}

	m_bPaused = false;
	m_bStopped = false;
	m_dwLastFileSourcesRequestTime = 0;
	SavePartFile();
	SaveSettingsFile();
	UpdateDisplayedInfo();

	g_App.m_pMDlg->m_wndTransfer.UpdateCatTabTitles();

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Used to load initial sources from Server right after connect
void CPartFile::GetSourcesAfterServerConnect()
{
	EMULE_TRY

	if (m_dwLastFileSourcesRequestTime > 0)
	{
		m_dwLastFileSourcesRequestTime = 0;
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CString CPartFile::GetPartFileStatus()
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	UINT		dwResStrId;

	if (GetTransferringSrcCount() > 0)
	{
		dwResStrId = IDS_DOWNLOADING;
	}
	else
	{
		dwResStrId = (!IsStalled()) ? IDS_WAITING : IDS_STALLED;
	}

	switch (GetStatus())
	{
		case PS_HASHING:
		case PS_WAITINGFORHASH:
			dwResStrId = IDS_HASHING;
			break;
		case PS_COMPLETING:
			dwResStrId = IDS_COMPLETING;
			break;
		case PS_COMPLETE:
			dwResStrId = IDS_COMPLETE;
			break;
		case PS_PAUSED:
			dwResStrId = IDS_PAUSED;
			break;
		case PS_ERROR:
			dwResStrId = IDS_ERRORLIKE;
			break;
		case PS_STOPPED:
			dwResStrId = IDS_STOPPED;
			break;
	}
	return GetResString(dwResStrId);

	EMULE_CATCH

#endif //NEW_SOCKETS_ENGINE
	return _T("");
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int CPartFile::GetPartFileStatusID()
{
	EMULE_TRY

	int	ipfStatus = 0;

	if (GetTransferringSrcCount() > 0)
	{
		ipfStatus = PS_DOWNLOADING;
	}
	else
	{
		if (!IsStalled())
			ipfStatus = PS_WAITINGFORSOURCE;
		else
			ipfStatus = PS_STALLED;
	}

	switch (GetStatus())
	{
		case PS_HASHING:
		case PS_WAITINGFORHASH:
			ipfStatus = PS_HASHING;
			break;
		case PS_COMPLETING:
		case PS_COMPLETE:
		case PS_PAUSED:
		case PS_ERROR:
		case PS_STOPPED:
			ipfStatus = GetStatus();
			break;
	}
	return ipfStatus;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int CPartFile::GetPartFileStatusRang()
{
	EMULE_TRY

//	Sort order is Complete - Completing - Downloading - Waiting - Stalled - Paused - Errors
	int status = 2;
	if (GetTransferringSrcCount() == 0)
	{
		if (!IsStalled())
			status = 3;
		else
			status = 4;
	}
	switch (GetStatus())
	{
		case PS_HASHING:
		case PS_WAITINGFORHASH:
			status = 6;
			break;
		case PS_COMPLETING:
			status = 1;
			break;
		case PS_COMPLETE:
			status = 0;
			break;
		case PS_PAUSED:
			status = 5;
			break;
		case PS_STOPPED:
			status = 7;
			break;
		case PS_ERROR:
			status = 8;
			break;
	}
	return status;

	EMULE_CATCH

	return 1;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
sint32 CPartFile::GetTimeRemaining(bool bAverage /*=false*/)
{
	EMULE_TRY

	uint32	iDataRate = 0;
	sint32	iTimeRemaining = -1;

	if (bAverage)
		iDataRate = GetAvgDataRate();
	else
		iDataRate = GetDataRate();
	iTimeRemaining = (iDataRate > 0 ? (GetFileSize() - GetCompletedSize()) / iDataRate : -1);
	return ((iTimeRemaining > 0 && iTimeRemaining <= 8640000 /*100 days*/) ? iTimeRemaining : -1);

	EMULE_CATCH

	return -1;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	GetAvgDataRate() recalculates the average bytes/second for the session. It stores the
//	average in 'm_AvgDataRate' as well as returning it.
uint32 CPartFile::GetAvgDataRate(bool bUpdated /*=false*/)
{
	EMULE_TRY

	if (bUpdated)
	{
		CTimeSpan	timeSpan;

	//	Calc on last FlushBuffer() to keep sessionstats for completed files
		if (m_eStatus == PS_COMPLETING || m_eStatus == PS_COMPLETE)
		{
			timeSpan = GetFlushTimeSpan();
		}
		else
		{
		//	TODO: Modify to use time since first data
			timeSpan = GetSessionTimeSpan();
		}

	//	Delay 1st calculation a bit & dont use datarate for balancing (peak-stopper?)
		uint32	iSeconds = timeSpan.GetTotalSeconds();

		if (iSeconds > 10)
			m_AvgDataRate = ((GetSessionTransferred() /*+ GetDataRate()*/) / (iSeconds /*+ 1*/));
		else
			m_AvgDataRate = 0;
	}

	EMULE_CATCH
	return m_AvgDataRate;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::PreviewFile()
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	if (IsArchive(true))
	{
		if ((!m_bRecoveringArchive) && (!m_bPreviewing))
			CArchiveRecovery::recover(this, true);
		return;
	}

	if (!PreviewAvailable())
		return;

	if (g_App.m_pPrefs->BackupPreview())
	{
		m_bPreviewing = true;
		CPreviewThread* pThread = (CPreviewThread*) AfxBeginThread(RUNTIME_CLASS(CPreviewThread), THREAD_PRIORITY_NORMAL + g_App.m_pPrefs->GetMainProcessPriority(), 0, CREATE_SUSPENDED);
		pThread->SetValues(this, g_App.m_pPrefs->GetVideoPlayer());
		pThread->ResumeThread();
	}
	else
	{
		CString strLine(m_strFilePath);
		CString strPlayer = g_App.m_pPrefs->GetVideoPlayer();

	//	If path has spaces use quotes
		if (strLine.Find(_T(' ')) >= 0)
		{
			strLine = _T('\"') + strLine + _T('\"');
		}

		if (strPlayer.IsEmpty())
		{
			ShellExecute(NULL, _T("open"), strLine, NULL, NULL, SW_SHOWNORMAL);
		}
		else
		{
			CString strPlayerPath;
			int i = strPlayer.ReverseFind('\\');

			if (i >= 0)
				strPlayerPath = strPlayer.Left(i + 1);
			ShellExecute(NULL, _T("open"), strPlayer, strLine, strPlayerPath, SW_SHOWNORMAL);
		}
	}

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::PreviewAvailable()
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	EnumPartFileStatuses	eFileStatus = GetStatus();

// 	No preview while completing or hashing or when file is already complete
	if (eFileStatus == PS_COMPLETING || eFileStatus == PS_COMPLETE || eFileStatus == PS_HASHING)
		return false;

//	Allow preview of archives of any length > 1k
	if (IsArchive(true))
	{
		if ( (GetFileSize() > 1024) && (GetCompletedSize() > 1024) && !m_bRecoveringArchive
			&& ((GetFreeDiskSpaceX(m_strTempDir) + 100000000) > (2 * GetFileSize())) )
			return true;
		else
			return false;
	}

	byte	byteMovieMode = GetMovieMode();

	if (byteMovieMode == 0)
	{
	//	Check only for AVI style movie, as below first chunk check is enough for all MPG-style ones
		if (IsAviMovie())
			byteMovieMode = (GetLastPartSize() < (GetFileSize() / 100)) ? 7 : 1;
	}

	if (g_App.m_pPrefs->BackupPreview())
	{
		return ( ((eFileStatus == PS_READY) || IsPaused())
				&& !m_bPreviewing && (GetPartCount() > 2) && IsMovie() && IsPartComplete(0)
				&& (((byteMovieMode != 1) && (byteMovieMode != 7)) || IsPartComplete(GetPartCount() - 1))
				&& ((byteMovieMode != 7) || IsPartComplete(GetPartCount() - 2))
				&& ((GetFreeDiskSpaceX(m_strTempDir) + 100000000) >= GetFileSize()) );
	}
	else
	{
		TCHAR szVideoPlayerFileName[_MAX_FNAME];
		_tsplitpath(g_App.m_pPrefs->GetVideoPlayer(), NULL, NULL, szVideoPlayerFileName, NULL);

	//	Enable the preview command if the according option is specified 'PreviewSmallBlocks'
	//	or if VideoLAN client is specified
		if (g_App.m_pPrefs->GetPreviewSmallBlocks() || !_tcsicmp(szVideoPlayerFileName, _T("vlc")))
		{
			if (m_bPreviewing)
				return false;

			if (!(eFileStatus == PS_READY || eFileStatus == PS_EMPTY || IsPaused()))
				return false;

			EED2KFileType	eFileType = GetFileType();

			if (eFileType != ED2KFT_VIDEO && eFileType != ED2KFT_AUDIO && eFileType != ED2KFT_CDIMAGE)
				return false;

		//	If it's an MPEG file, VLC is even capable of showing parts of the file if the beginning of the file is missing!
			if (IsMpgMovie() || IsMpgAudio())
			{
			//	TODO: search a block which is at least 16K (Audio) or 256K (Video)
				if (GetCompletedSize() < 16 * 1024)
					return false;
			}
			else
			{
			//	For AVI files it depends on the used codec..
				if (!IsComplete(0, 256 * 1024))
					return false;
			}

			return true;
		}
		else
		{
			return ( ((eFileStatus == PS_READY) || IsPaused())
					&& !m_bPreviewing && (GetPartCount() > 2) && IsMovie() && IsPartComplete(0)
					&& (((byteMovieMode != 1) && (byteMovieMode != 7)) || IsPartComplete(GetPartCount() - 1))
					&& ((byteMovieMode != 7) || IsPartComplete(GetPartCount() - 2)) );
		}
	}

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::AllowGet1stLast()
{
	if (!IsMovie() || (GetPartCount() < 3))
		return false;

	byte	byteMovieMode = GetMovieMode();

	if (byteMovieMode == 0)
	{
	//	Check only for AVI style movie, as below first chunk check is enough for all MPG-style ones
		if (IsAviMovie())
			byteMovieMode = (GetLastPartSize() < (GetFileSize() / 100)) ? 7 : 1;
	}
	return !( IsPartComplete(0)
				&& (((byteMovieMode != 1) && (byteMovieMode != 7)) || IsPartComplete(GetPartCount() - 1))
				&& ((byteMovieMode != 7) || IsPartComplete(GetPartCount() - 2)) );
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
Packet*	CPartFile::CreateSrcInfoPacket(CUpDownClient *pForClient)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	if (pForClient->m_pReqPartFile != this)
		return NULL;

	if (GetSourceCount() == 0)
		return NULL;

	CSafeMemFile	packetStream(1024);
	uint16			uCount = 0;

	packetStream.Write(m_fileHash, 16);
	packetStream.Write(&uCount, 2);

	CUpDownClient  *pPotentialSource;
	bool 			bNeededPart;
	byte			*pbyteForClientPartStatus, *pbytePotentialClientPartStatus;
	uint32			dwID, dwServerIP, dwRes;
	uint16			uPort, uServerPort;
	ClientList		clientListCopy;
	const uint16	uPartCount = GetPartCount();

	GetCopySourceLists(SLM_ALL, &clientListCopy);
	for (ClientList::const_iterator cIt = clientListCopy.begin(); cIt != clientListCopy.end(); cIt++)
	{
		pPotentialSource = *cIt;

		if (md4cmp(pPotentialSource->m_reqFileHash, pForClient->m_reqFileHash) != 0)
			continue;

	//	Don't send a client to itself as a source (part 1)
		if (pForClient == pPotentialSource)
			continue;
	//	Don't send a client to itself as a source (part 2)
		if ((dwRes = pForClient->Compare(pPotentialSource)) != 0)
		{
		//	Actually if function 'CClientList::AttachToAlreadyKnown()' is working properly, then we should never see this message
			AddDebugLogLine( RGB_BLUE_VIOLET_TXT _T("Found same client during SX c1=(%s,%s), c2=(%s,%s)"), 
							pPotentialSource->GetClientNameWithSoftware(),
							HashToString(pPotentialSource->GetUserHash()),
							pForClient->GetClientNameWithSoftware(),
							HashToString(pForClient->GetUserHash()) );
			continue;
		}
	//	Check LowID status
		if (pPotentialSource->HasLowID())
		{
		//	Don't send a LowID client to LowID as they will not able to connect
			if (pForClient->HasLowID())
				continue;
		//	Don't send a LowID client if server IP & server port are unknown
			if (pPotentialSource->GetServerIP() == 0 || pPotentialSource->GetServerPort() == 0)
				continue;
		//	Don't send a LowID client to HighID if client on different server
			if ( pPotentialSource->GetServerIP() != pForClient->GetServerIP()
				|| pPotentialSource->GetServerPort() != pForClient->GetServerPort() )
				continue;
		}
	//	Don't send clients that have no parts available
		if (pPotentialSource->GetAvailUpPartCount() == 0)
			continue;
	//	Don't share Lan clients sources as they are private ips/userids
		if (pPotentialSource->IsOnLAN())
			continue;
	//	don't send erroneous or fake sources
		if (uPartCount != pPotentialSource->GetUpPartCount())
			continue;

		bNeededPart = false;

	//	Only send source which have needed parts for this client if possible
		pbyteForClientPartStatus = pForClient->GetPartStatus();
		pbytePotentialClientPartStatus = pPotentialSource->GetPartStatus();

	//	Both clients support part statuses so we try to find needed clients
		if (pbyteForClientPartStatus != NULL && pbytePotentialClientPartStatus != NULL)
		{
			for (uint32 i = 0; i < uPartCount; i++)
			{
				if (pbytePotentialClientPartStatus[i] && !pbyteForClientPartStatus[i])
				{
					bNeededPart = true;
					break;
				}
			}
		}
	//	If not we give a sources with at least one part (for files that have more than one part)
		else
			bNeededPart = true;

		if (bNeededPart)
		{
			uCount++;
			dwID = pPotentialSource->GetUserID();
			uPort = pPotentialSource->GetUserPort();
			dwServerIP = pPotentialSource->GetServerIP();
			uServerPort = pPotentialSource->GetServerPort();

			packetStream.Write(&dwID, 4);
			packetStream.Write(&uPort, 2);
			packetStream.Write(&dwServerIP, 4);
			packetStream.Write(&uServerPort, 2);
			if (pForClient->GetSourceExchangeVersion() > 1)
				packetStream.Write(pPotentialSource->GetUserHash(), 16);
			if (uCount > 500)
				break;
		}
	}

	if (uCount == 0)
		return NULL;

	packetStream.Seek(16, 0);
	packetStream.Write(&uCount, 2);

	Packet	*pPacket = new Packet(&packetStream, OP_EMULEPROT);

	pPacket->m_eOpcode = OP_ANSWERSOURCES;
	if (pPacket->m_dwSize > 354)
		pPacket->PackPacket();

	return pPacket;

	EMULE_CATCH

#endif //NEW_SOCKETS_ENGINE
	return NULL;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::AddClientSources(CMemFile *packetStream, byte byteSourceExchangeVersion)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

//	If this part file is paused or complete/completing or erroneous no sources needed any more.
	EnumPartFileStatuses	eFileStatus = GetStatus();

	if ( (eFileStatus == PS_PAUSED) || (eFileStatus == PS_STOPPED)
		|| (eFileStatus == PS_COMPLETE) || (eFileStatus == PS_COMPLETING)
		|| (eFileStatus == PS_ERROR) )
		return;

	uint16		uNumSources;
	packetStream->Read(&uNumSources, 2);

//	Check if the data size matches the 'uNumSources' for v1 or v2 and eventually correct the source
//	exchange version while reading the packet data. Otherwise we could experience a higher
//	chance in dealing with wrong source data, userhashs and finally duplicate sources.
	UINT		uDataSize = packetStream->GetLength() - packetStream->GetPosition();

//	Checks if v1 packet has the correct size
	if (uNumSources * (4u + 2u + 4u + 2u) == uDataSize)
	{
		byteSourceExchangeVersion = 1;
	}
//	Checks if v2 packet has the correct size
	else if (uNumSources * (4u + 2u + 4u + 2u + 16u) == uDataSize)
	{
		byteSourceExchangeVersion = 2;
	}
	else
	{
	//	If v3 inserts additional data (like v2), the above code will correctly filter those packets.
	//	If v3 appends additional data after <count>(<Sources>)[count], we are in trouble with the
	//	above code. Though a client which does not understand v3+ should never receive such a packet.
		g_App.m_pMDlg->AddDebugLogLine(_T("Received invalid source exchange packet (v%u) of data size %u for %s"), byteSourceExchangeVersion, uDataSize, GetFileName());
		return;
	}

	for (int i = 0; i != uNumSources; i++)
	{
		CClientSource		source;

		packetStream->Read(&source.sourceID, 4);
		packetStream->Read(&source.sourcePort, 2);
		packetStream->Read(&source.serverIP, 4);
		packetStream->Read(&source.serverPort, 2);
		if (byteSourceExchangeVersion > 1)
			packetStream->Read(&source.achUserHash, 16);
		AddClientSource(&source, i, true, byteSourceExchangeVersion);
	}

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Updates priority of file if autopriority is activated
void CPartFile::UpdateDownloadAutoPriority(void)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	if (IsAutoPrioritized())
	{
		int iValidSources = GetSourceCount() - GetNotCurrentSourcesCount();

		if (iValidSources < g_App.m_pPrefs->PriorityHigh() && priority != PR_HIGH)
			SetPriority(PR_HIGH);
		else if (iValidSources >= g_App.m_pPrefs->PriorityHigh() && iValidSources < g_App.m_pPrefs->PriorityLow() && priority != PR_NORMAL)
			SetPriority(PR_NORMAL);
		else if (iValidSources >= g_App.m_pPrefs->PriorityLow() && priority != PR_LOW)
			SetPriority(PR_LOW);
	}

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Making this function return a higher when more sources have the extended
//	protocol will force you to ask a larger variety of people for sources
uint32 CPartFile::GetCommonFilePenalty()
{
//	TODO: implement, but never return less than MINCOMMONPENALTY!
	return MINCOMMONPENALTY;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CString CPartFile::GetStatsFullPath() const
{
	return m_strFilePath + _T(".stats");
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//--- xrmb:keepPartFileStats ---
bool CPartFile::LoadPartFileStats()
{
	EMULE_TRY

	byte	abyteTmp[4 + 4 + 8];
	FILE	*fh = _tfopen(GetStatsFullPath(), _T("rb"));

//	If we cant open this file you lose the stats
	if (fh == NULL)
		return false;

//	Get version
	byte	version = 0;
	if (fread(&version, sizeof(version), 1, fh) != 1)
	{
		fclose(fh);
		return false;
	}

//	Skip the former storage place for part file requests/accepted/tranfer statistics
//	Starting from v1.1c this data is kept in .part.met
//	'fread' is used instead of 'fseek' to increase performance of file stream buffering
	if (fread(abyteTmp, sizeof(abyteTmp), 1, fh) != 1)
	{
		fclose(fh);
		return false;
	}

//	This is for the parttraffic
//	I thought about explicit adding the traffic like for the other stats above
//	but it doesnt make sense, because you'll no transfer if you reach this
	if (LoadFromFileTraffic(fh, version) == false)
	{
		fclose(fh);
		return false;
	}

	fclose(fh);

	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::SavePartFileStats()
{
	static const uint32 s_adwDummy[4] = { 0, 0, 0, 0 };
	EMULE_TRY

	FILE *fh = _tfopen(GetStatsFullPath(), _T("wb"));

//	If we cant open this file you lose the stats
	if (fh == NULL)
		return false;

//	Set version
	byte	version = 2;
	if (fwrite(&version, sizeof(version), 1, fh) != 1)
	{
		fclose(fh);
		return false;
	}

//	Fill with zeroes the former storage place for part file requests/accepted/tranfer statistics
//	Starting from v1.1c this data is kept in .part.met
	if (fwrite(s_adwDummy, sizeof(s_adwDummy), 1, fh) != 1)
	{
		fclose(fh);
		return false;
	}

//	This is for the parttraffic
	if (SaveToFileTraffic(fh) == false)
	{
		fclose(fh);
		return false;
	}

	fclose(fh);

	return true;

	EMULE_CATCH

	return false;
}
//--- :xrmb ---
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Used as separate thread to complete file
UINT CPartFile::CompleteThreadProc(CPartFile* pFile)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	if (g_App.m_pPrefs->IsNTBased())
		SetThreadLocale(g_App.m_pPrefs->GetLanguageID());

	if (pFile == NULL)
		return ~0u;

	pFile->PerformFileComplete();

	return 0;

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE

	return ~0u;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CString CPartFile::DtLocale(CTime dt)
{
#ifndef NEW_SOCKETS_ENGINE
	static const UINT		s_adwDayResId[] =
	{
		IDS_SCH_EXCEPT_SUN, IDS_SCH_EXCEPT_MON, IDS_SCH_EXCEPT_TUE, IDS_SCH_EXCEPT_WED,
		IDS_SCH_EXCEPT_THU, IDS_SCH_EXCEPT_FRI, IDS_SCH_EXCEPT_SAT
	};
	SYSTEMTIME st;
	dt.GetAsSystemTime(st);

	int		nDateSize = GetDateFormat(LOCALE_USER_DEFAULT, 0, &st, NULL, NULL, 0);
	int		nTimeSize = GetTimeFormat(LOCALE_USER_DEFAULT, 0, &st, NULL, NULL, 0);
	CString	strDate, strTime, strBuffer;

	GetDateFormat(LOCALE_USER_DEFAULT, DATE_SHORTDATE, &st, NULL, strDate.GetBuffer(nDateSize), nDateSize);
	strDate.ReleaseBuffer();

	GetTimeFormat(LOCALE_USER_DEFAULT, TIME_NOSECONDS, &st, NULL, strTime.GetBuffer(nTimeSize), nTimeSize);
	strTime.ReleaseBuffer();

	strBuffer.Format(_T("%s, %s, %s"),
		GetResString(s_adwDayResId[dt.GetDayOfWeek() - 1]), strDate, strTime);
	return strBuffer;
#else
	return _T("");
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	LoadSettingsFile() loads settings for partfiles (Movie Preview Mode, Stopped status, Preallocation)
void CPartFile::LoadSettingsFile()
{
	EMULE_TRY

	CString strIniFile(m_strFilePath);

	strIniFile += _T(".settings");
	if (PathFileExists(strIniFile))
	{
		CIni filesettings(strIniFile);

		filesettings.SetDefaultCategory(_T("Settings"));
		byte Mode = filesettings.GetInt(_T("MoviePreviewMode"), 0);

		if ((Mode > 7) || ((Mode & 6) == 2) || (Mode == 6))	// Check for valid values 0, 1, 4, 5, 7
			Mode = 0;
		CKnownFile::SetMovieMode(Mode);
		m_bStopped = filesettings.GetBool(_T("StoppedStatus"), m_bStopped);
		m_bIsPreallocated = filesettings.GetBool(_T("Preallocated"), m_bIsPreallocated);

		filesettings.CloseWithoutSave();
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	SaveSettingsFile() save settings for partfiles (Movie Preview Mode, Stopped status, Preallocation)
void CPartFile::SaveSettingsFile()
{
	EMULE_TRY

	CString strIniFile(m_strFilePath);

	strIniFile += _T(".settings");

	CIni filesettings(strIniFile);

	filesettings.SetDefaultCategory(_T("Settings"));
	filesettings.SetInt(_T("MoviePreviewMode"), ((CKnownFile*)this)->GetMovieMode());
	filesettings.SetBool(_T("StoppedStatus"), m_bStopped);
	filesettings.SetBool(_T("Preallocated"), m_bIsPreallocated);

	filesettings.SaveAndClose();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	WriteToBuffer() buffer rx data which will be flushed later by timer or size threshold.
//		Params:
//			iCompressionGain - number of bytes saved by compression (it can be negative
//			                   after connection loss in the middle of compressed block);
//			data             - pointer to data buffer;
//			dwStart          - data start offset;
//			dwEnd            - data end offset (inclusive);
//			iFlags           - additional flags.
//				PF_WR2BUF_FL_FREEBUFFER - data buffer can be used for further operations (caller doesn't free it)
//				PF_WR2BUF_FL_ENDOFBLOCK - current data is the last portion of the requested block
//		Return: number of buffered bytes
uint32 CPartFile::WriteToBuffer(sint32 iCompressionGain, BYTE *data, uint32 dwStart, uint32 dwEnd, int iFlags)
{
	EMULE_TRY

//	This is needed a few times
	uint32 lenData = dwEnd - dwStart + 1;

	if (iCompressionGain != 0)
	{
	//	Compression gain can be negative in case of disconnection in the middle of compressed stream
		m_qwGainDueToCompression += static_cast<uint64>(iCompressionGain);
		g_App.m_pPrefs->Add2SavedFromCompression(iCompressionGain);
	}

//	Occasionally packets are duplicated, no point writing it twice
	if (IsComplete(dwStart, dwEnd))
	{
		AddDebugLogLine(_T("File '%s' has already been written from %lu to %lu size %lu"), GetFileName(), dwStart, dwEnd, lenData);
	}
	else
	{
	//	Create a new buffered queue entry
		PartFileBufferedData *item = new PartFileBufferedData;

		item->m_dwStartOffset = dwStart;
		item->m_dwEndOffset = dwEnd;

		if ((iFlags & PF_WR2BUF_FL_FREEBUFFER) == 0)
		{
		//	Create copy of data as new buffer
			BYTE *buffer = new BYTE[lenData];

			memcpy2(buffer, data, lenData);
			data = buffer;
		}
		item->m_pbyteBuffer = data;

	//	Add to the queue in the correct position (most likely the end)
		PartFileBufferedData *queueItem;
		bool added = false;
		POSITION pos = m_BufferedData_list.GetTailPosition();
		while (pos != NULL)
		{
			POSITION posNext = pos;

			queueItem = m_BufferedData_list.GetPrev(pos);
			if (item->m_dwEndOffset > queueItem->m_dwEndOffset)
			{
				added = true;
				m_BufferedData_list.InsertAfter(posNext, item);
				break;
			}
		}
		if (!added)
			m_BufferedData_list.AddHead(item);

	//	Increment buffer size marker
		m_nTotalBufferData += lenData;

	//	Mark this small section of the file as filled
		FillGap(item->m_dwStartOffset, item->m_dwEndOffset);

		uint32 dwPartNum = dwStart / PARTSIZE;

	//	Check that all data of the part is received
		if (IsPartFull(dwPartNum))
		{
		//	Prevent data flush during processesing of the data from the sockets
			m_bDataFlushReq = true;
		}
		else if (((iFlags & PF_WR2BUF_FL_ENDOFBLOCK) != 0) && IsCorruptedPart(dwPartNum))
		{
		//	Flush and check data to minimize amount of additional unrequired redownloads
		//	Do it only by end of the block basis to minimize system load as some clients send by small portions
			m_bDataFlushReq = true;
		}

		m_timeLastDownTransfer = CTime::GetCurrentTime();

	//	Return the length of data written to the buffer
		return lenData;
	}

	EMULE_CATCH

	if ((iFlags & PF_WR2BUF_FL_FREEBUFFER) != 0)
		delete [] data;

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	FlushBuffer() flushes buffered data from the source to the part file
void CPartFile::FlushBuffer(void)
{
	EMULE_TRY

	m_bDataFlushReq = false;

	m_nLastBufferFlushTime = ::GetTickCount();

//	If there's no buffered data to flush, just return
	if (m_BufferedData_list.IsEmpty())
		return;

//	Reset the Last Down Transfer timer
	m_timeLastDownTransfer = CTime::GetCurrentTime();
	GetAvgDataRate(true);

//	Calculate the number of parts in the part file
	uint32	dwPartCount = GetPartCount();
	bool *changedPart = new bool[dwPartCount];

	try
	{
	//	Remember which parts need to be checked at the end of the flush
		for (uint32 dwPartNum = 0; dwPartNum < dwPartCount; dwPartNum++)
		{
			changedPart[dwPartNum] = false;
		}

	//	Ensure file is big enough to write data to (the last item will be the furthest from the start)
		PartFileBufferedData *item = m_BufferedData_list.GetTail();

	//	If the end of the buffered data is past the current end of the part file, extend it
		if (m_hPartFileWrite.GetLength() <= item->m_dwEndOffset)
			m_hPartFileWrite.SetLength(item->m_dwEndOffset + 1);

	//	For each block of buffered data...
		for (int i = m_BufferedData_list.GetCount(); i > 0; i--)
		{
		//	Remove item from queue
			item = m_BufferedData_list.RemoveHead();

		//	For each part spanned by the buffered data, set the part's changed flag
			for (uint32 curPart = item->m_dwStartOffset / PARTSIZE; curPart <= item->m_dwEndOffset / PARTSIZE; curPart++)
				changedPart[curPart] = true;

		//	Go to the correct position in file and write block of data
			m_hPartFileWrite.Seek(item->m_dwStartOffset, CFile::begin);

			uint32 lenData = item->m_dwEndOffset - item->m_dwStartOffset + 1;

		//	Decrease buffer size
			m_nTotalBufferData -= lenData;
			m_hPartFileWrite.Write(item->m_pbyteBuffer, lenData);

		//	Release memory used by this item
			delete [] item->m_pbyteBuffer;
			delete item;
		}

	//	If the size of the part file is greater than the size of the target file,
	//	set it to the size of the target file.
		if (m_hPartFileWrite.GetLength() > GetFileSize())
			m_hPartFileWrite.SetLength(GetFileSize());

	//	Flush to disk
		m_hPartFileWrite.Flush();

	//	Check each part of the file
		uint32	partRange = GetLastPartSize() - 1;

	//	For each part in the part file (in reverse)...
		for (int iPartNum = (int)dwPartCount - 1; iPartNum >= 0; iPartNum--)
		{
		//	If we didn't just write it, skip it
			if (changedPart[iPartNum] == false)
			{
			//	Any parts other than last must be full size
				partRange = PARTSIZE - 1;
				continue;
			}

		//	Is this 9MB part complete
			if (IsComplete(PARTSIZE * iPartNum, (PARTSIZE * (iPartNum + 1)) - 1))
			{
			//	Is part corrupt
				if (!HashSinglePart(iPartNum))
				{
#ifndef NEW_SOCKETS_ENGINE
					AddLogLine(true, RGB_LOG_WARNING + GetResString(IDS_ERR_PARTCORRUPT), iPartNum, GetFileName());
#endif //NEW_SOCKETS_ENGINE
					AddGap(PARTSIZE * iPartNum, (PARTSIZE * iPartNum + partRange));
					m_csGapListAndPartStatus.Lock();
					m_PartsStatusVector[iPartNum] |= PART_CORRUPTED;
					m_csGapListAndPartStatus.Unlock();
				//	Increment corruption statistics
					AddRxCorruptedAmount(partRange + 1);
				}
				else
				{
					m_csGapListAndPartStatus.Lock();
					m_PartsStatusVector[iPartNum] = (m_PartsStatusVector[iPartNum] & 0xFFFFFF) | PART_VERIFIED;
					m_csGapListAndPartStatus.Unlock();

				//	Successfully completed part, make it available for sharing
					if (m_eStatus == PS_EMPTY)
					{
						SetStatus(PS_READY);
#ifndef NEW_SOCKETS_ENGINE
						g_App.m_pSharedFilesList->SafeAddKnownFile(this);
#endif //NEW_SOCKETS_ENGINE
					}

					m_dwCompletedPartsSize += (partRange + 1);
				}
			}
			else if (IsCorruptedPart(iPartNum))
			{
			//	Try to recover with minimal loss
				if (HashSinglePart(iPartNum))
				{
					m_csGapListAndPartStatus.Lock();
					m_PartsStatusVector[iPartNum] = (m_PartsStatusVector[iPartNum] & 0xFFFFFF) | PART_VERIFIED;
					m_csGapListAndPartStatus.Unlock();

					m_iTotalPacketsSavedDueToICH++;
#ifndef NEW_SOCKETS_ENGINE
					g_App.m_pPrefs->IncSessionPartsSavedByICH();
#endif //NEW_SOCKETS_ENGINE

					uint32 dwSaved = GetPartLeftToDLSize(iPartNum);

					FillGap(PARTSIZE * iPartNum, (PARTSIZE * iPartNum + partRange));
					RemoveBlockFromList(PARTSIZE * iPartNum, (PARTSIZE * iPartNum + partRange));
#ifndef NEW_SOCKETS_ENGINE
					AddLogLine(true, RGB_LOG_SUCCESS + GetResString(IDS_ICHWORKED), iPartNum, m_strFileName, CastItoXBytes(dwSaved));
#endif //NEW_SOCKETS_ENGINE

				//	Decrement corruption statistics by the number of saved bytes
					m_qwLostDueToCorruption -= dwSaved;
					g_App.m_pPrefs->SubLostFromCorruption(dwSaved);

				//	Successfully completed part, make it available for sharing
					if (m_eStatus == PS_EMPTY)
					{
						SetStatus(PS_READY);
#ifndef NEW_SOCKETS_ENGINE
						g_App.m_pSharedFilesList->SafeAddKnownFile(this);
#endif //NEW_SOCKETS_ENGINE
					}
					m_dwCompletedPartsSize += (partRange + 1);
				}
			}
		//	Any parts other than last must be full size
			partRange = PARTSIZE - 1;
		}

	//	Update met file
		SavePartFile();

	//	Is this file finished?
		m_csGapListAndPartStatus.Lock();

		bool	bIsFileFinished = gaplist.IsEmpty() && !m_bHashSetNeeded;	// can't complete without a hashset

		m_csGapListAndPartStatus.Unlock();

		if (bIsFileFinished)
			CompleteFile(false);
	}
	catch (CFileException * error)
	{
		OUTPUT_DEBUG_TRACE();

#ifndef NEW_SOCKETS_ENGINE
		if (g_App.m_pPrefs->IsErrorBeepEnabled())
			Beep(800, 200);
#endif //NEW_SOCKETS_ENGINE

		if (error->m_cause == CFileException::diskFull)
		{
			CString strSpace = CastItoXBytes(GetFreeDiskSpaceX(m_strTempDir));
			CString strTemp;

			strTemp.Format(_T("%s (%s)"), strSpace, m_strTempDir);
#ifndef NEW_SOCKETS_ENGINE
			AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_OUTOFSPACE), GetFileName());
			AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_DWTOT_FS), strTemp);
		//	Sending message for insufficient diskspace
			CString MessageText;
			MessageText.Format(GetResString(IDS_ERR_OUTOFSPACE), GetFileName());
			g_App.m_pSMTPConnection->SendMail(MessageText, g_App.m_pPrefs->GetNotifierPopOnImportantError(), g_App.m_pPrefs->IsSMTPWarningEnabled());
			g_App.m_pMDlg->ShowNotifier(MessageText, TBN_IMPORTANTEVENT, false, g_App.m_pPrefs->GetNotifierPopOnImportantError());
#endif //NEW_SOCKETS_ENGINE
		}
		else
		{
#ifndef NEW_SOCKETS_ENGINE
			AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_WRITEERROR), GetFileName(), GetErrorMessage(error));
#endif //NEW_SOCKETS_ENGINE
			SetStatus(PS_ERROR);
		}
		PauseFile();
		error->Delete();
	}
	catch (...)
	{
#ifndef NEW_SOCKETS_ENGINE
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_WRITEERROR), GetFileName(), GetResString(IDS_UNKNOWN));
#endif //NEW_SOCKETS_ENGINE
		SetStatus(PS_ERROR);
		PauseFile();

	//	When ERRONEOUS close PartFile to allow Initialize
		ClosePartFile();
	}
	delete[] changedPart;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// This will invert the gap list, up to caller to delete gaps when done
// 'Gaps' returned are really the filled areas, and guaranteed to be in order
void CPartFile::GetFilledList(CTypedPtrList<CPtrList, Gap_Struct*> *filled)
{
	EMULE_TRY

	Gap_Struct * gap = NULL;
	Gap_Struct *best = NULL;
	POSITION pos;
	uint32 start = 0;
	uint32 bestEnd = 0;

//	Will be unlocked on exit
	CSingleLock lockGap(&m_csGapListAndPartStatus, TRUE);

//	Loop until done
	bool finished = false;

	while (!finished)
	{
		finished = true;
	//	Find first gap after current start pos
		bestEnd = GetFileSize();
		pos = gaplist.GetHeadPosition();
		best = gaplist.GetHead();
		while (pos != NULL)
		{
			gap = gaplist.GetNext(pos);
			if ((gap->m_dwStartOffset > start) && (gap->m_dwEndOffset < bestEnd))
			{
				best = gap;
				bestEnd = best->m_dwEndOffset;
				finished = false;
			}
		}

		if (!finished)
		{
		//	Invert this gap
			gap = new Gap_Struct;
			gap->m_dwStartOffset = start;
			gap->m_dwEndOffset = best->m_dwStartOffset - 1;
			start = best->m_dwEndOffset + 1;
			filled->AddTail(gap);
		}
		else if (best->m_dwEndOffset < GetFileSize())
		{
			gap = new Gap_Struct;
			gap->m_dwStartOffset = best->m_dwEndOffset + 1;
			gap->m_dwEndOffset = GetFileSize();
			filled->AddTail(gap);
		}
	}
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::UpdateFileRatingCommentAvail()
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	if (!this)
		return;

	bool prev = (m_bHasComment || m_bHasRating);

	m_bHasComment = false;
	m_bHasRating = false;

	ClientList	clientListCopy;

	GetCopySourceLists(SLM_ALL, &clientListCopy);
	for (ClientList::const_iterator cIt = clientListCopy.begin(); cIt != clientListCopy.end(); cIt++)
	{
		CUpDownClient	*pSource = *cIt;

		if (!pSource->IsFileCommentEmpty())
			m_bHasComment = true;
		if (pSource->GetFileRating() != PF_RATING_NONE)
			m_bHasRating = true;
		if (m_bHasComment && m_bHasRating)
			break;
	}

	if ((!m_bHasComment) || (!m_bHasRating))
	{
		for (POSITION pos = m_pastCommentList.GetHeadPosition(); pos != NULL;)
		{
			CPastComment &pc = m_pastCommentList.GetNext(pos);
			if (pc.GetComment().GetLength() > 0)
				m_bHasComment = true;
			if (pc.GetRating() > 0)
				m_bHasRating = true;
			if (m_bHasComment && m_bHasRating)
				break;
		}
	}

	if ((!m_bHasComment) || (!m_bHasRating))
	{
		EnumPartFileRating	eRated = GetFileRating();

		if (!GetFileComment().IsEmpty())
			m_bHasComment = true;
		if (eRated != PF_RATING_NONE)
			m_bHasRating = true;
	}

	if (prev != (m_bHasComment || m_bHasRating))
		UpdateDisplayedInfo();

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint16 CPartFile::GetSourceCount()
{
	uint32	dwCount = 0;

	for (EnumDLQState eDS = DS_DOWNLOADING; eDS < DS_LAST_QUEUED_STATE; ++eDS)
		dwCount += m_SourceLists[eDS].size();
	return static_cast<uint16>(dwCount);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
EnumPartFileRating CPartFile::GetRating()
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	if (!m_bHasRating)
		return PF_RATING_NONE;

	int num, tot, iRate;

	num = tot = 0;

	ClientList	clientListCopy;

	GetCopySourceLists(SLM_ALL, &clientListCopy);
	for (ClientList::const_iterator cIt = clientListCopy.begin(); cIt != clientListCopy.end(); cIt++)
	{
		CUpDownClient	*pSource = *cIt;

		iRate = pSource->GetFileRating();
		if (iRate > 0)
		{
			num++;
		//	Bugfix: for some $%#ing reason fair=4 & good=3, breaking the progression from fake(1) to excellent(5)
			if (iRate == PF_RATING_GOOD || iRate == PF_RATING_FAIR)
				iRate = (iRate == PF_RATING_GOOD) ? 4 : 3;

			tot += iRate;
		}
	}

	for (POSITION pos = m_pastCommentList.GetHeadPosition(); pos != NULL;)
	{
		iRate = m_pastCommentList.GetNext(pos).GetRating();

		if (iRate > 0)
		{
			num++;
		//	Bugfix: for some $%#ing reason  fair=4 & good=3, breaking the progression from fake(1) to excellent(5)
			if (iRate == PF_RATING_GOOD || iRate == PF_RATING_FAIR)
				iRate = (iRate == PF_RATING_GOOD) ? 4 : 3;

			tot += iRate;
		}
	}

	iRate = GetFileRating();
	if (iRate > 0)
	{
			num++;
		//	Bugfix: for some $%#ing reason  fair=4 & good=3, breaking the progression from fake(1) to excellent(5)
			if (iRate == PF_RATING_GOOD || iRate == PF_RATING_FAIR)
				iRate = (iRate == PF_RATING_GOOD) ? 4 : 3;

			tot += iRate;
	}

	if (num > 0)
	{
		num = static_cast<double>(tot) / num + .5;	//Cax2 - get the average of all the ratings
	//	Bugfix: for some $%#ing reason good=3 & fair=4, breaking the progression from fake(1) to excellent(5)
		if (num == 3 || num == 4)
			num = (num == 3) ? PF_RATING_FAIR : PF_RATING_GOOD;
	}
	return static_cast<_EnumPartFileRating>(num);	//Cax2 - if no ratings found, will return 0!

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE

	return PF_RATING_NONE;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	UpdateDisplayedInfo() causes the DownloadList ctrl to redisplay the item for this file if either
void CPartFile::UpdateDisplayedInfo()
{
	EMULE_TRY

#ifndef NEW_SOCKETS_ENGINE
	g_App.m_pDownloadList->UpdateFile(this);
#endif //NEW_SOCKETS_ENGINE

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Hostname sources in ED2K superLink supporting Source Exchange v2
void CPartFile::AddClientSource(CClientSource* pSource, int iSource, bool bExchanged, byte byteSourceExchangeVersion)
{
	EMULE_TRY

//	Check first if we are this source
#ifdef OLD_SOCKETS_ENABLED
	if (g_App.m_pServerConnect->GetClientID() < 0x1000000 && g_App.m_pServerConnect->IsConnected())
	{
		if ( g_App.m_pServerConnect->GetClientID() == pSource->sourceID &&
			 g_App.m_pPrefs->GetPort() == pSource->sourcePort )
		{
			return;
		}

	//	Although we are currently having a LowID, we could have had a HighID before, which
	//	that client is sending us now! seems unlikely ... it happend!
		if ( g_App.m_pServerConnect->GetLocalIP() == pSource->sourceID &&
			 g_App.m_pPrefs->GetPort() == pSource->sourcePort )
		{
			return;
		}
	}
	else if ( g_App.m_pServerConnect->GetClientID() == pSource->sourceID &&
			  g_App.m_pPrefs->GetPort() == pSource->sourcePort )
	{
		return;
	}
//	Use lowID source only if it's on the same server as us
	else if ( pSource->sourceID < 0x1000000 &&
			  !g_App.m_pServerConnect->IsLocalServer(pSource->serverIP, pSource->serverPort) )
	{
		return;
	}
#endif //OLD_SOCKETS_ENABLED

#ifndef NEW_SOCKETS_ENGINE
	if ( g_App.m_pPrefs->GetMaxSourcePerFile() > GetSourceCount()
	     && iSource < static_cast<int>(g_App.m_pPrefs->GetMaxSourcePerFileSoft()) )
	{
		CUpDownClient* pNewSource = g_App.m_pDownloadQueue->CheckAndAddSource(this,pSource->sourceID,pSource->sourcePort, pSource->serverIP, pSource->serverPort, pSource->achUserHash);
		if (bExchanged && pNewSource)
		{
		//	Name the exchanged source
			pNewSource->SetUserName(GetResString(IDS_EXCHANGEDSOURCE));
		//	Check exchange version
			if (byteSourceExchangeVersion > 1)
				pNewSource->SetUserHash(pSource->achUserHash);
		}
	}
#endif //NEW_SOCKETS_ENGINE

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::AddClientSources(CTypedPtrList<CPtrList, CClientSource*>* sources)
{
	EMULE_TRY

	POSITION pos;
	pos = sources->GetHeadPosition();
	while (pos != NULL)
	{
		CClientSource * source = (CClientSource*)sources->GetAt(pos);

		if (source->sourceType != ED2KLINK_SOURCE_IP)
		{
			sources->GetNext(pos);
			continue;
		}

		POSITION removePos = pos;
		sources->GetNext(pos);
		sources->RemoveAt(removePos);

		AddClientSource(source);
		delete source;
	}
#ifndef NEW_SOCKETS_ENGINE
	g_App.m_pDownloadQueue->AddClientHostnameToResolve(sources);
#endif //NEW_SOCKETS_ENGINE

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Create source ed2k from downloads
CString CPartFile::CreateED2KSourceLink(uint32 expirein, int sourcenum, bool bIsSourceSaver /*=false*/)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	CTypedPtrList<CPtrList, CUpDownClient*> srcstoadd;
	POSITION	pos2;
	ClientList	clientListCopy;

	GetCopySourceLists(SLM_ALL, &clientListCopy);
	for (ClientList::const_iterator cIt = clientListCopy.begin(); cIt != clientListCopy.end(); cIt++)
	{
		CUpDownClient	*pSource = *cIt;

		if (pSource->HasLowID())
			continue;
		if (srcstoadd.IsEmpty())
		{
			srcstoadd.AddHead(pSource);
			continue;
		}
		if ((sourcenum == 0) || (srcstoadd.GetCount() < sourcenum) || (pSource->GetAvailablePartCount() > srcstoadd.GetTail()->GetAvailablePartCount()))
		{
			if ((sourcenum != 0) && (srcstoadd.GetCount() == sourcenum))
				srcstoadd.RemoveTail();
			ASSERT(srcstoadd.GetCount() < sourcenum);
			bool bInserted = false;
			for (pos2 = srcstoadd.GetTailPosition(); pos2 != NULL;)
			{
				CUpDownClient * cur_srctoadd = srcstoadd.GetPrev(pos2);
				if (cur_srctoadd->GetAvailablePartCount() > pSource->GetAvailablePartCount())
				{
					srcstoadd.InsertAfter(pos2, pSource);
					bInserted = true;
					break;
				}
			}
			if (!bInserted)
				srcstoadd.AddHead(pSource);
		}
	}

	uint32 dwID = 0;
#ifdef OLD_SOCKETS_ENABLED
	dwID = g_App.m_pServerConnect->GetClientID();
#endif //OLD_SOCKETS_ENABLED

	CString strLink;
	strLink.Format( _T("ed2k://|file|%s|%u|%s|/"),
			g_App.StripInvalidFilenameChars(GetFileName(), false),  // Spaces to dots
			GetFileSize(),
			HashToString(GetFileHash()) );

	if (srcstoadd.IsEmpty())
	{
		if (bIsSourceSaver)
			return strLink;
#ifdef OLD_SOCKETS_ENABLED
	//	!g_App.m_pServerConnect->IsConnected() is not really needed since we give an ip but ...
		else if (!g_App.m_pServerConnect->IsConnected() || g_App.m_pServerConnect->IsLowID())
		{
			AddLogLine(true, RGB_LOG_WARNING + GetResString(IDS_SOURCELINKFAILED));
			return _T("");
		}
#endif //OLD_SOCKETS_ENABLED
	}

//	Is the expiration date is handle in other eMule mods ? a bIsSourceSaver is perhaps needed ...
	strLink += _T("|sources");
	if (expirein != 0)
	{
		COleDateTime expirationDate = COleDateTime::GetCurrentTime();
		COleDateTimeSpan timediff;
		timediff.SetDateTimeSpan(expirein, 0, 0, 0);
		expirationDate += timediff;
		strLink.AppendFormat(_T("@%02u%02u%02u"), (expirationDate.GetYear() % 100), expirationDate.GetMonth(), expirationDate.GetDay());
	}

	if (!bIsSourceSaver && dwID >= 0x1000000)
		strLink.AppendFormat(_T(",%u.%u.%u.%u:%u"), (byte)dwID, (byte)(dwID >> 8), (byte)(dwID >> 16), (byte)(dwID >> 24), g_App.m_pPrefs->GetPort());

	if (!bIsSourceSaver)
		AddDebugLogLine(_T("Generated ed2k link with %i sources for file '%s' (part count %u)"), (int)srcstoadd.GetCount(), GetFileName(), GetPartCount());

	while (!srcstoadd.IsEmpty())
	{
		CUpDownClient * pSource = srcstoadd.RemoveHead();
		if (!bIsSourceSaver)
			AddDebugLogLine(RGB_LOG_DIMMED_TXT _T("| User: %s - Available part count: %u"), pSource->GetUserName(), (int)pSource->GetAvailablePartCount());
		strLink.AppendFormat(_T(",%s:%u"), pSource->GetFullIP(), pSource->GetUserPort());
	}

	strLink.Append(_T("|/"));
	return strLink;

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE

	return _T("");
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::SetAlternativeOutputDir(CString *path)
{
	EMULE_TRY

	path->TrimRight(_T('\\'));
	m_strAlternativePath = *path;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	GetOutputDir() returns the appropriate destination directory path for this file.
CString CPartFile::GetOutputDir()
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	CString	strRetPath = g_App.m_pPrefs->GetIncomingDir();
	CString	strCatPath;

	if (GetCatID() != 0)
		strCatPath = CCat::GetCatByID(GetCatID())->GetPath();

//	If no alternative path defined
	if (m_strAlternativePath.IsEmpty())
	{
	//	... and a category dir is defined...
		if (strCatPath != _T(""))
			return strCatPath; // Return the category dir
		else
			return strRetPath;	// Return the incoming dir
	}

//	MOREVIT - Why all the heavy duty shell call nonsense? You could replace all of this
//	with a single call to PathFileExists(). Somebody just thought they were more
//	clever than they were methinks.

//	We're going to use desktop folder "static" functions , we need an object
//	of IShellFolder type and DeskTop always exists
	IShellFolder	*pDesktop = NULL;
	ITEMIDLIST	*pList = NULL;
	CString	strDir;

//	Try to create DeskTop folder object
	if (SHGetDesktopFolder(&pDesktop) == NOERROR)
	{
		ULONG	uAttrib	= SFGAO_FOLDER; //we're going to look for the "folder" attribute

		strDir = m_strAlternativePath;
	//	"ParseDisplayName" uses UNICODE so BSTR will go
	//	We have to remember to release it
		BSTR	pData	= strDir.AllocSysString();

		if (pData)
		{
		//	'm_strAlternativePath' is always created without "\" but this was the way
		//	the original code was, maybe it was for a reason :) , so I am going to leave
		//	first check without adding "\" as well ...

		//	So here we convert "display name" (path string) into ID list used by Shell
			pDesktop->ParseDisplayName(NULL, NULL, pData, NULL, &pList, &uAttrib);
		//	Release the BSTR
			::SysFreeString(pData);
		//	If list we got is null then no such object exists
			if (!pList)
			{
			//	Add "\" to get folder name
				strDir += _T('\\');
			//	Allocate new search BSTR
				pData = strDir.AllocSysString();
				if (pData)
				{
				//	We're going to look for "folder" attribute
					uAttrib	= SFGAO_FOLDER ;
				//	We convert "display name" (path string) into ID list used by Shell
					pDesktop->ParseDisplayName(NULL, NULL, pData, NULL, &pList, &uAttrib);
				//	Release the BSTR
					::SysFreeString(pData);
				}
			}
		//	If we found the object and it's "folder" type
			if (pList && (uAttrib & SFGAO_FOLDER))
			{
				LPMALLOC	ppMalloc;

				if (SHGetMalloc(&ppMalloc) == NOERROR)
				{
					ppMalloc->Free(pList);
					ppMalloc->Release();
				}

			//	Share the destination directory
				if (g_App.m_pPrefs->SharedDirListCheckAndAdd(strDir, true))
				{
				//	New path was added to the list, scan for files
					g_App.m_pSharedFilesList->Reload();
				}
				return m_strAlternativePath;
			}
		//	Just in case the object is not folder we have to release memory
			else if (pList)
			{
				LPMALLOC	ppMalloc;

				if (SHGetMalloc(&ppMalloc) == NOERROR)
				{
					ppMalloc->Free(pList);
					ppMalloc->Release();
				}
			}
		}
		pDesktop->Release();
	}

//	Default return
	strRetPath.TrimRight(_T('\\'));
	return strRetPath;

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE

	return _T("");
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CString CPartFile::GetDownloadFileInfo()
{
#ifndef NEW_SOCKETS_ENGINE
	CString strInfo, strLastSeenComplete, strLastProgress, strTimeLeft, strFileSz = CastItoXBytes(GetFileSize());
	double dblAvailability = 0.0;

	if (lastseencomplete == NULL)
		GetResString(&strLastSeenComplete, IDS_NEVER);
	else
		strLastSeenComplete = LocalizeLastSeenComplete();

	if (GetTransferred() == 0)
		GetResString(&strLastProgress, IDS_NEVER);
	else
		strLastProgress = LocalizeLastDownTransfer();

	if (GetPartCount() != 0)
		dblAvailability = GetAvailablePartCount() * 100 / GetPartCount();

	EnumPartFileStatuses	eFileStatus = GetStatus();

	if ((eFileStatus != PS_COMPLETING) && (eFileStatus != PS_COMPLETE))
		strTimeLeft = CastSecondsToHM(GetTimeRemaining());
	else
		strTimeLeft = _T("-");

	strInfo.Format( _T("%s (%s)\n\n%s: %s\n\n%s: %s\n") + GetResString(IDS_PARTINFOS),
							 GetFileName(), strFileSz,
	             GetResString(IDS_STATUS), GetPartFileStatus(),
	             GetResString(IDS_FILEHASH), HashToString(GetFileHash()),
							 GetResString(IDS_DL_FILENAME), m_strFilePath,
	             GetPartCount(), GetResString(IDS_AVAIL), GetAvailablePartCount(), dblAvailability,
	             static_cast<int>(GetPercentCompleted()),
							 CastItoXBytes(GetCompletedSize()), strFileSz, GetTransferringSrcCount(),
	             GetResString(IDS_LASTSEENCOMPL), strLastSeenComplete,
	             GetResString(IDS_FD_LASTCHANGE), strLastProgress,
				 GetResString(IDS_DLCOL_REMAININGTIME), strTimeLeft,
	             GetOnQueueSrcCount(), GetQueueFullSrcCount(), GetNoNeededPartsSrcCount(),
	             GetSrcA4AFCount(), GetConnectingSrcCount() );

	return strInfo;
#else
	return _T("");
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Improved Tooltips
CString CPartFile::GetDownloadFileInfo4Tooltips()
{
#ifndef NEW_SOCKETS_ENGINE
	CString strInfo, strLastSeenComplete, strLastProgress, strTransferSrcCount, strFileSz = CastItoXBytes(GetFileSize());
	double dblAvailability = 0.0;

	if (lastseencomplete == NULL)
		GetResString(&strLastSeenComplete, IDS_NEVER);
	else
		strLastSeenComplete = LocalizeLastSeenComplete();

	if (GetTransferred() == 0)
		GetResString(&strLastProgress, IDS_NEVER);
	else
		strLastProgress = LocalizeLastDownTransfer();

	if (GetPartCount() != 0)
		dblAvailability = GetAvailablePartCount() * 100 / GetPartCount();

	if (GetTransferringSrcCount() > 0)
		strTransferSrcCount.Format(GetResString(IDS_TT_PARTINFOS3), GetTransferringSrcCount());

	strInfo.Format(_T("<t=2><b>%s</b><br><t=2>%s (%s %s)<br><t=2>%d%s (%s / %s) %s<br><hr=100%%><br><b>%s:<t></b>%s<br><b>%s:<t></b>%s<br><b>%s:<t></b>%s<br><b>%s<t></b>%s<br><b>%s:<t></b>%d, %s: %d (%.1f%%)<br><b>%s:<t></b>%s<br><b>%s:<t></b>%s<br><b>%s:<t></b>")
	             + GetResString(IDS_TT_STATUS2) + _T("<br><b>%s:<t></b>%s"),
	             GetFileName(), strFileSz, CastItoThousands(GetFileSize()), GetResString(IDS_BYTES),
	             static_cast<int>(GetPercentCompleted()), GetResString(IDS_PROCDONE), CastItoXBytes(GetCompletedSize()), strFileSz, strTransferSrcCount,
	             GetResString(IDS_STATUS), GetPartFileStatus(),
	             GetResString(IDS_FILEHASH), HashToString(GetFileHash()),
				 GetResString(IDS_DL_FILENAME), m_strFilePath,
				 GetResString(IDS_FD_OUTPUT), GetOutputDir(),
	             GetResString(IDS_UP_PARTS), GetPartCount(), GetResString(IDS_AVAIL), GetAvailablePartCount(), dblAvailability,
	             GetResString(IDS_LASTSEENCOMPLETE), strLastSeenComplete,
				 GetResString(IDS_LASTCHANGE), strLastProgress,
	             GetResString(IDS_DL_SOURCES), GetOnQueueSrcCount(),
	             GetQueueFullSrcCount(),
	             GetNoNeededPartsSrcCount(),
	             GetSrcA4AFCount(),
	             GetConnectingSrcCount(),
                 GetResString(IDS_SIZE_ON_DISK), CastItoXBytes(GetRealFileSize()) );

	return strInfo;
#else
	return _T("");
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Ornis: Creating progressive presentation of the partfilestatuses - for webdisplay
CString CPartFile::GetProgressString(uint16 size)
{
	EMULE_TRY

	const char crProgress = '0';										// Green
	const char crHave = '1';											// Black
	const char crPending = '2';											// Yellow
	const char crMissing = '3';											// Red
	static const char crWaiting[6] = { '4', '5', '6', '7', '8', '9'};	// Differents blue

	CString	my_ChunkBar;

	for (uint16 i = 0; i <= (size + 1); i++)
		my_ChunkBar.AppendChar(crHave);	// One more for safety

	uint32 m_nFileSize = GetFileSize();
	double unit = static_cast<double>(size) / static_cast<double>(m_nFileSize);

	POSITION				pos;
	EnumPartFileStatuses	eFileStatus = GetStatus();

	if ((eFileStatus == PS_COMPLETE) || (eFileStatus == PS_COMPLETING))
	{
		CharFillRange(&my_ChunkBar, 0, static_cast<double>(m_nFileSize) * unit, crProgress);
	}
	else
	{
	//	Red gaps
	//	Will be unlocked on block exit
		CSingleLock lockGap(&m_csGapListAndPartStatus, TRUE);

		for (pos = gaplist.GetHeadPosition(); pos != NULL;)
		{
			Gap_Struct	   *pGap = gaplist.GetNext(pos);
			uint32			dwGapBeg = pGap->m_dwStartOffset;
			uint32			dwGapEnd = pGap->m_dwEndOffset + 1;

			for (uint32 i = (dwGapBeg / PARTSIZE); i < GetPartCount(); i++)
			{
				const uint32 dwPartStart = i * PARTSIZE;
				const uint32 dwPartEnd = ((GetFileSize() - dwPartStart) > PARTSIZE) ? (dwPartStart + PARTSIZE) : GetFileSize();

				if (dwGapEnd > dwPartEnd)
					dwGapEnd = dwPartEnd;	// The rest is in the next part

				byte	color;
				uint32	dwFreq;

				if (m_srcPartFrequencies.GetCount() >= (INT_PTR)i && ((dwFreq = m_srcPartFrequencies[i]) != 0))
				{
					if (IsPartDownloading(i))
						color = crPending;	// Downloading part in yellow
					else
						color = (dwFreq < 10) ? crWaiting[dwFreq / 2] : crWaiting[5];
				}
				else
					color = crMissing;

				CharFillRange(&my_ChunkBar, static_cast<double>(dwGapBeg) * unit, static_cast<double>(dwGapEnd) * unit, color);

				dwGapBeg = dwGapEnd;
				dwGapEnd = pGap->m_dwEndOffset + 1;
				if (dwGapBeg == dwGapEnd) // finished?
					break;
			}
		}
	//	m_csGapListAndPartStatus is released here
	}

	return my_ChunkBar;

	EMULE_CATCH

	return _T("");
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::CharFillRange(CString* buffer, double start, double dblEnd, char color)
{
	uint32	dwEnd = static_cast<uint32>(dblEnd);

	for (uint32 i = start; i <= dwEnd; i++)
		buffer->SetAt(i, color);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// A4AF management
void CPartFile::DownloadAllA4AF(bool bSameCat /* FALSE */)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	for (POSITION pos = m_A4AFsrcList.GetHeadPosition(); pos != NULL; )
	{
		if(bSameCat)
		{
			CUpDownClient* pClient = m_A4AFsrcList.GetNext(pos);
			if (pClient->m_pReqPartFile->GetCatID() == GetCatID())
				pClient->SwapToAnotherFile(this);
		}
		else
		{
			m_A4AFsrcList.GetNext(pos)->SwapToAnotherFile(this);
		}

	}

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
byte* CPartFile::MMCreatePartStatus()
{
	EMULE_TRY

//	Create partstatus + info in mobilemule protocol specs
//	result needs to be deleted[] | slow, but not timecritical
	byte * result = new byte[GetPartCount() + 1];
	for (int i = 0; i < GetPartCount(); i++)
	{
		result[i] = 0;
		if (IsPartComplete(i))
		{
			result[i] = 1;
		}
		else
		{
			if (IsComplete(i * PARTSIZE + (0 * (PARTSIZE / 3)), ((i * PARTSIZE) + (1 * (PARTSIZE / 3))) - 1))
				result[i] += 2;
			if (IsComplete(i * PARTSIZE + (1 * (PARTSIZE / 3)), ((i * PARTSIZE) + (2 * (PARTSIZE / 3))) - 1))
				result[i] += 4;
			if (IsComplete(i * PARTSIZE + (2 * (PARTSIZE / 3)), ((i * PARTSIZE) + (3 * (PARTSIZE / 3))) - 1))
				result[i] += 8;
			byte freq = (byte)m_srcPartFrequencies[i];
			if (freq > 44)
				freq = 44;
			freq = (freq + 2) / 3;
			freq <<= 4;
			result[i] += freq;
		}
	}
	return result;

	EMULE_CATCH

	return NULL;
};
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::CheckAndAddPastComment(CUpDownClient *pClient)
{
#ifndef NEW_SOCKETS_ENGINE
	if (pClient->IsFileCommentEmpty() && (pClient->GetFileRating() == PF_RATING_NONE))
		return;

	if (pClient->m_pReqPartFile != this)
		return;

	m_pastCommentList.Add(pClient->GetUserHash(), pClient->GetUserName(), GetFileName(), pClient->GetFileComment(), pClient->GetFileRating());
	UpdateFileRatingCommentAvail();
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::RemovePastComment(CUpDownClient *pClient, bool bRestore)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	if (!pClient)
		return;

	if (bRestore)
	{
		CString	strPastComment;
		EnumPartFileRating	ePastRating;

		if ( (m_pastCommentList.GetCommentRating(pClient->GetUserHash(), &strPastComment, &ePastRating))
		     && (pClient->IsFileCommentEmpty())
		     && (pClient->GetFileRating() == PF_RATING_NONE) )
		{
		//	Restore past comment
			pClient->SetFileComment(strPastComment);
			pClient->SetFileRating(ePastRating);
		}
	}

	m_pastCommentList.Remove(pClient->GetUserHash());

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
EnumCategories CPartFile::GetCatID() const
{
//	If catID is valid... (valid range is 1..CAT_PREDEFINED-1 but a
//	real check would have to scan the category map. This isn't the
//	best place to do that.
#ifndef VS2002 // In VS2002 the const cast operator causes an ambiguity
	if (static_cast<const _EnumCategories>(m_eCategoryID) >= CAT_PREDEFINED)
#else
	if (const_cast<CPartFile*>(this)->m_eCategoryID >= CAT_PREDEFINED)
#endif
		return CAT_NONE;
	else
		return m_eCategoryID;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::SetCatID(EnumCategories eCatID)
{
	m_eCategoryID = eCatID;

//	Set new prio
	if (IsPartFile())
	{
		int	iPriority = 0;

		if (eCatID != CAT_NONE)
		{
			iPriority = CCat::GetCatByID(GetCatID())->GetPriority();
		}
		switch (iPriority)
		{
			case 0:
			{
				break;
			}
			case 1:
			{
				SetAutoPriority(false);
				SetPriority(PR_LOW, false);
				break;
			}
			case 2:
			{
				SetAutoPriority(false);
				SetPriority(PR_NORMAL, false);
				break;
			}
			case 3:
			{
				SetAutoPriority(false);
				SetPriority(PR_HIGH, false);
				break;
			}
			case 4:
			{
				SetAutoPriority(true);
				SetPriority(PR_HIGH, false);
				break;
			}
		}

		if (IsPartFile() && !m_strTempDir.IsEmpty())
			SavePartFile();
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::SetStatus(EnumPartFileStatuses in)
{
	m_eStatus = in;

	UpdateGUIAfterStateChange();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::UpdateGUIAfterStateChange()
{
#ifndef NEW_SOCKETS_ENGINE
	EnumCategories	eCurCat = g_App.m_pMDlg->m_wndTransfer.m_ctlDownloadList.GetCurTabCat();

//	Changes in file status can affect whether the file belongs to a predefined category or not
	if (eCurCat >= CAT_PREDEFINED)
		g_App.m_pMDlg->m_wndTransfer.m_ctlDownloadList.ChangeCategoryByID(eCurCat);
	else
		UpdateDisplayedInfo();

//	Update file counts in the tabs to reflect status changes
	g_App.m_pMDlg->m_wndTransfer.UpdateCatTabTitles();
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	GetRealFileSize() returns real file size on disk with respect to NTFS compression and/or NTFS sparse files
uint64 CPartFile::GetRealFileSize()
{
	return ::GetDiskFileSize(GetFilePath());
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	IsStalled() performs obaldin's stall check
bool CPartFile::IsStalled() const
{
	bool	bIsStalled = false;

	CTimeSpan	ts = CTime::GetCurrentTime() - GetLastDownTransfer();

//	If WARN_PERIOD_OF_NO_PROGRESS days have gone by without a single download...
//	(??? or maybe sooner if there are no needed parts available?)
	if (ts.GetTotalHours() >= 24 * WARN_PERIOD_OF_NO_PROGRESS)
		bIsStalled = true;

	return bIsStalled;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	AllocateNeededSpace() is the preallocate function used to allocate the space needed by a file.
//	It uses a thread to let eMule still work and not drop sources while generating the file.
void CPartFile::AllocateNeededSpace()
{
	HANDLE	lTh;
	DWORD	lThID;
	m_bIsPreallocated = true;

	PAThreadParam *tp = new PAThreadParam;

	tp->cPartFile = this;
	tp->iPosition = GetFileSize() - 1;

	lTh = CreateThread(NULL, 0L, AllocateNeededSpaceProc, tp, 0, &lThID);
	CloseHandle(lTh);

}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	AllocateNeededSpaceProc() is the thread created by AllocateNeededSpace().
//	It receives a pointer to 'm_hPartFileWrite' and the filesize-1 (where it has to write the bogus data to allocate space.
DWORD WINAPI CPartFile::AllocateNeededSpaceProc(LPVOID lpParameter)
{
	PAThreadParam * tp = (PAThreadParam*)lpParameter;

	try
	{
		tp->cPartFile->m_hPartFileWrite.Seek(tp->iPosition, CFile::begin);
		tp->cPartFile->m_hPartFileWrite.Write("1", 1);
		tp->cPartFile->m_hPartFileWrite.Flush();

		tp->cPartFile->SaveSettingsFile();
	}
	catch (CFileException *error)
	{
		AddDebugLogLine( RGB_LOG_ERROR_TXT _T("Preallocate for file %s: error %s"),
						 tp->cPartFile->GetFileName(), GetErrorMessage(error) );

		tp->cPartFile->m_bIsPreallocated = false;
		error->Delete();
	}

	delete tp;
	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	IsLastBlockComplete() returns status of last block of file (if downloaded the file was already allocated).
bool CPartFile::IsLastBlockComplete()
{
	return IsComplete(GetFileSize() - GetLastBlockSize(), GetFileSize() - 1);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	ClosePartFile() closes all .part file handles
void CPartFile::ClosePartFile()
{
	if (m_hPartFileWrite.m_hFile != INVALID_HANDLE_VALUE)
	{
		m_hPartFileRead.Close();
		m_hPartFileWrite.Close();
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	ReadFileForUpload() opens known file and read information in circular way.
//		Params:
//			dwOffset      - file offset to start reading from;
//			dwBytesToRead - number of bytes to read;
//			pbyteBuffer   - buffer to receive data.
//		Return:
//			< 0 in case of error, > 0 - information was read from a part file, else 0.
int CPartFile::ReadFileForUpload(uint32 dwOffset, uint32 dwBytesToRead, byte *pbyteBuffer)
{
	int	iRc = -2;

	EMULE_TRY

//	Synchronization with completion thread
//	will be unlocked on exit
	CSingleLock lockComplete(&m_csFileCompletion, TRUE);

	if (m_eStatus != PS_COMPLETE)
	{
	//	This file isn't complete yet
		m_hPartFileRead.Seek(dwOffset, CFile::begin);

		if (uint32 dwDone = m_hPartFileRead.Read(pbyteBuffer, dwBytesToRead) != dwBytesToRead)
		{
			m_hPartFileRead.SeekToBegin();
			m_hPartFileRead.Read(pbyteBuffer + dwDone, dwBytesToRead - dwDone);
		}
		iRc = 1;
	}
	else
	{
	//	This file is complete and is located in destination directory
		iRc = (::CircularFileRead(m_strFilePath, dwOffset, dwBytesToRead, pbyteBuffer)) ? 0 : -1;
	}

	EMULE_CATCH

	return iRc;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	GetFirstLastChunk4Preview() sets Movie Preview Mode by checking media type and needed index size (last 1% of file)
void CPartFile::GetFirstLastChunk4Preview()
{
	EMULE_TRY

	if (GetMovieMode())
		SetMovieMode(0);
	else
	{
		if (IsAviMovie())
		{
			SetMovieMode((GetLastPartSize() < (GetFileSize() / 100)) ? 7 : 1);
		}
		else if (IsMpgMovie())
		{
			if (GetFileSize() < 209715200)
				SetMovieMode(4);
			else
				SetMovieMode(5);
		}
		else
			SetMovieMode(0);
	}
	SaveSettingsFile();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	GetGapListCopy() returns a protected copy of the gap list
void CPartFile::GetGapListCopy(CArray<Gap_Struct, Gap_Struct> *pGapArray)
{
	Gap_Struct	*pGap;

	CSingleLock lockGap(&m_csGapListAndPartStatus, TRUE);	// will be unlocked on the exit

	pGapArray->SetSize(gaplist.GetCount(), 0);
	pGap = pGapArray->GetData();

	for (POSITION pos = gaplist.GetHeadPosition(); pos != NULL;)
	{
		*pGap++ = *gaplist.GetNext(pos);
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::SetFileName(const CString& NewName, bool bClearName)
{
//	Set the name & extension
	CKnownFile::SetFileName(NewName, bClearName);

	CString strExt = GetFileExtension();
	static const TCHAR s_apcPreviewableExt[][4] =
	{
		_T("zip"), _T("rar"), _T("cbz"), _T("cbr")
	};
	static const TCHAR s_apcAviMovieExt[][5] =
	{
		_T("avi"), _T("ogm"), _T("mkv"), _T("wmv"), _T("divx"), _T("xvid")
	};
	static const TCHAR s_apcMpgMovieExt[][5] =
	{
		_T("mpg"), _T("mpeg"), _T("vob"), _T("asf"), _T("bin"), _T("mpe"), _T("dat")
	};
	static const TCHAR s_apcMpgAudioExt[][4] =
	{
		_T("mp3"), _T("mp2"), _T("mpa")
	};

//	Check previewable status
	m_bIsPreviewableArchive = false;
	for (unsigned ui = 0; ui < ELEMENT_COUNT(s_apcPreviewableExt); ui++)
	{
		if (strExt.Compare(reinterpret_cast<const TCHAR*>(&s_apcPreviewableExt[ui])) == 0)
		{
			m_bIsPreviewableArchive = true;
			break;
		}
	}

//	Check movie status
	m_bIsMovie = (ED2KFT_VIDEO == GetFileType() || strExt == _T("bin"));

//	Check AVI movie status
	m_bIsAviMovie = false;
	for (unsigned ui = 0; ui < ELEMENT_COUNT(s_apcAviMovieExt); ui++)
	{
		if (strExt.Compare(reinterpret_cast<const TCHAR*>(&s_apcAviMovieExt[ui])) == 0)
		{
			m_bIsAviMovie = true;
			break;
		}
	}

//	Check MPG movie status
	m_bIsMpgMovie= false;
	for (unsigned ui = 0; ui < ELEMENT_COUNT(s_apcMpgMovieExt); ui++)
	{
		if (strExt.Compare(reinterpret_cast<const TCHAR*>(&s_apcMpgMovieExt[ui])) == 0)
		{
			m_bIsMpgMovie = true;
			break;
		}
	}

//	Check MPG movie status
	m_bIsMpgAudio = false;
	for (unsigned ui = 0; ui < ELEMENT_COUNT(s_apcMpgAudioExt); ui++)
	{
		if (strExt.Compare(reinterpret_cast<const TCHAR*>(&s_apcMpgAudioExt[ui])) == 0)
		{
			m_bIsMpgAudio = true;
			break;
		}
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::AddRxCorruptedAmount(uint32 dwBadBytes)
{
	m_qwLostDueToCorruption += dwBadBytes;
	g_App.m_pPrefs->Add2LostFromCorruption(dwBadBytes);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::GetCopySourceList(EnumDLQState eClientDS, ClientList *pCopy)
{
	if (!pCopy->empty())
		pCopy->clear();

	if ((eClientDS < DS_LAST_QUEUED_STATE) && !m_SourceLists[eClientDS].empty())
	{
		EnterCriticalSection(&m_csSourceLists);

		pCopy->insert(pCopy->begin(), m_SourceLists[eClientDS].begin(), m_SourceLists[eClientDS].end());

		LeaveCriticalSection(&m_csSourceLists);
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::GetCopySourceLists(uint32 dwListsMask, ClientList *pCopy)
{
	if (!pCopy->empty())
		pCopy->clear();

	EnterCriticalSection(&m_csSourceLists);
	for (EnumDLQState eDS = DS_DOWNLOADING; eDS < DS_LAST_QUEUED_STATE; ++eDS, dwListsMask >>= 1)
	{
		if (((dwListsMask & 1) != 0) && !m_SourceLists[eDS].empty())
		{
			pCopy->insert(pCopy->end(), m_SourceLists[eDS].begin(), m_SourceLists[eDS].end());
		}
	}
	LeaveCriticalSection(&m_csSourceLists);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::RemoveClientFromDLSourceList(CUpDownClient *pClient)
{
	EnumDLQState	eDS = pClient->GetDownloadState();

	if (eDS < DS_LAST_QUEUED_STATE)
	{
		uint32	dwListSizeBefore, dwListSizeAfter;

		EnterCriticalSection(&m_csSourceLists);

		dwListSizeBefore = m_SourceLists[eDS].size();
		m_SourceLists[eDS].remove(pClient);
		dwListSizeAfter = m_SourceLists[eDS].size();

		LeaveCriticalSection(&m_csSourceLists);

	//	Update availability information if a client was removed
		if (dwListSizeAfter != dwListSizeBefore)
		{
			NewSrcPartsInfo();
			return true;
		}
	}
	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::AddClientToSourceList(CUpDownClient *pClient, EnumDLQState eClientDS)
{
	if (pClient == NULL)
		return;

	if (eClientDS < DS_LAST_QUEUED_STATE)
	{
		EnterCriticalSection(&m_csSourceLists);

		if (find(m_SourceLists[eClientDS].begin(), m_SourceLists[eClientDS].end(), pClient) == m_SourceLists[eClientDS].end())
			m_SourceLists[eClientDS].push_back(pClient);

		LeaveCriticalSection(&m_csSourceLists);
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::IsClientInSourceList(CUpDownClient *pClient)
{
	bool			bResult = false;
	EnumDLQState	eDS = pClient->GetDownloadState();

	if ((eDS < DS_LAST_QUEUED_STATE) && !m_SourceLists[eDS].empty())
	{
		EnterCriticalSection(&m_csSourceLists);
		ClientList::iterator itResult = find(m_SourceLists[eDS].begin(), m_SourceLists[eDS].end(), pClient);
		bResult = (itResult != m_SourceLists[eDS].end());
		LeaveCriticalSection(&m_csSourceLists);
	}
	return bResult;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::ClearSourceLists()
{
	EnterCriticalSection(&m_csSourceLists);
	for (EnumDLQState eDS = DS_DOWNLOADING; eDS < DS_LAST_QUEUED_STATE; ++eDS)
		m_SourceLists[eDS].clear();
	LeaveCriticalSection(&m_csSourceLists);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::SwapClientBetweenSourceLists(CUpDownClient *pClient, EnumDLQState eSourceDS, EnumDLQState eTargetDS)
{
//	Speed up swapping (i.e. don't call the related functions if client is outside the lists)

//	Check if we need to remove the client
	if (eSourceDS < DS_LAST_QUEUED_STATE)
	{
		EnterCriticalSection(&m_csSourceLists);
		m_SourceLists[eSourceDS].remove(pClient);
		pClient->ChangeDownloadState(eTargetDS);
	//	Check if we need to add client to another queue
		if (eTargetDS < DS_LAST_QUEUED_STATE)
		{
			m_SourceLists[eTargetDS].push_back(pClient);
			LeaveCriticalSection(&m_csSourceLists);
		}
		else
		{
			LeaveCriticalSection(&m_csSourceLists);

		//	Update file status information if client was removed from the queue
			NewSrcPartsInfo();
		}
	}
	else
	{
	//	Check if we need to add client to another queue
		if (eTargetDS < DS_LAST_QUEUED_STATE)
		{
		// don't attach the sources if file is completed
			if ((m_eStatus != PS_COMPLETE) && (m_eStatus != PS_COMPLETING))
			{
				EnterCriticalSection(&m_csSourceLists);
				pClient->ChangeDownloadState(eTargetDS);
				m_SourceLists[eTargetDS].push_back(pClient);
				LeaveCriticalSection(&m_csSourceLists);
			}
		}
		else
			pClient->ChangeDownloadState(eTargetDS);
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
